function gen_direct2_kernel_lcorr(th, tw, f, t2d)
% Generates a kernel function for the 2d direct LCC 
% for a given template size.
%
% INPUTS:
%     th    height of the template
%     tw    width of the template
%     f     filename to write to
%     t2d   size of 2d threads block (assume square)
%
% Georgios Papamakarios (gpapamak@auth.gr)
% Electrical and Computer Engineering AUTH
% April 2011


ph = t2d + th - 1;
pw = t2d + tw - 1;
p_size = ph*pw;
t_size = th*tw;

fprintf(f,'template <class flccPrecision>\n');
fprintf(f,'__global__ void lcorrelation2d_kernel_%dx%d (\n', th, tw);
fprintf(f,'\tflccPrecision *imageDev, flccPrecision *templateDev, flccPrecision *outputDev, int SW, int BW )\n');

fprintf(f,'{\n');
fprintf(f,'\tconst int u0 = IMUL(blockIdx.x, blockDim.x);\n');
fprintf(f,'\tconst int v0 = IMUL(blockIdx.y, blockDim.y);\n');

fprintf(f,'\t__shared__ flccPrecision panelSh[%d];\n', p_size);
fprintf(f,'\tconst int i_idx = u0 + IMUL(v0, BW);\n');
fprintf(f,'\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', ph, t2d);
fprintf(f,'\t\t#pragma unroll\n');
fprintf(f,'\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', pw, t2d);
fprintf(f,'\t\t\tpanelSh[IMUL(j, %d) + i] = imageDev[i_idx + IMUL(j, BW) + i]; } }\n', pw);

fprintf(f,'\t__shared__ flccPrecision templateSh[%d];\n', t_size);
if th > t2d
	if tw > t2d
		fprintf(f,'\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', th, t2d);
		fprintf(f,'\t\t#pragma unroll\n');
		fprintf(f,'\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', tw, t2d);
		fprintf(f,'\t\t\tconst int jtwi = IMUL(j, %d) + i;\n', tw);
		fprintf(f,'\t\t\ttemplateSh[jtwi] = templateDev[jtwi]; } }\n');
	else
		fprintf(f,'\tif (threadIdx.x < %d) {\n', tw);
		fprintf(f,'\t\t#pragma unroll\n');
		fprintf(f,'\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', th, t2d);
		fprintf(f,'\t\t\tconst int jtwi = IMUL(j, %d) + threadIdx.x;\n', tw);
		fprintf(f,'\t\t\ttemplateSh[jtwi] = templateDev[jtwi]; } }\n');
	end
else
	if tw > t2d
		fprintf(f,'\tif (threadIdx.y < %d) {\n', th);
		fprintf(f,'\t\t#pragma unroll\n');
		fprintf(f,'\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', tw, t2d);
		fprintf(f,'\t\t\tconst int jtwi = IMUL(threadIdx.y, %d) + i;\n', tw);
		fprintf(f,'\t\t\ttemplateSh[jtwi] = templateDev[jtwi]; } }\n');
	else
		fprintf(f,'\tif (threadIdx.x < %d && threadIdx.y < %d) {\n', tw, th);
		fprintf(f,'\t\tconst int jtwi = IMUL(threadIdx.y, %d) + threadIdx.x;\n', tw);
		fprintf(f,'\t\ttemplateSh[jtwi] = templateDev[jtwi]; }\n');
	end
end

fprintf(f,'\t__syncthreads();\n');
fprintf(f,'\tregister flccPrecision corr   = 0.0;\n');
fprintf(f,'\tregister flccPrecision sum    = 0.0;\n');
fprintf(f,'\tregister flccPrecision sum_sq = 0.0;\n');
fprintf(f,'\tconst int p_idx = threadIdx.x + IMUL(threadIdx.y, %d);\n', pw);
fprintf(f,'\tregister flccPrecision im;\n');

for i=0:th-1
    for j=0:tw-1
        fprintf(f,'\t\tim = panelSh[p_idx + %d];\n', j+i*pw);
        fprintf(f,'\t\tcorr   += im * templateSh[%d];\n', j+i*tw);
        fprintf(f,'\t\tsum    += im;\n');
        fprintf(f,'\t\tsum_sq += im * im;\n');
    end
end

fprintf(f,'\tconst flccPrecision var = sum_sq - sum*sum/%d;\n', t_size);
fprintf(f,'\tconst int out_idx = (u0+threadIdx.x) + IMUL((v0+threadIdx.y), SW);\n');
fprintf(f,'\tif (var == 0)	outputDev[out_idx] = 0.0;\n');
fprintf(f,'\telse			outputDev[out_idx] = corr / sqrtf(var);\n');
fprintf(f,'}\n');

fprintf(f,'#ifndef NO_SINGLE_PREC \n');
fprintf(f,'template __global__ void lcorrelation2d_kernel_%dx%d<float> ', th, tw);
fprintf(f,'( float *imageDev, float *templateDev, float *outputDev, int SW, int BW );\n');
fprintf(f,'#endif \n');
fprintf(f,'#ifndef NO_DOUBLE_PREC \n');
fprintf(f,'template __global__ void lcorrelation2d_kernel_%dx%d<double> ', th, tw);
fprintf(f,'( double *imageDev, double *templateDev, double *outputDev, int SW, int BW );\n');
fprintf(f,'#endif \n');
fprintf(f,'\n');
