function gen_direct3_kernel_lcorr(th, tw, td, f, t3d)
% Generates a kernel function for the 3d direct LCC 
% for a given template size.
%
% INPUTS:
%     th    height of the template
%     tw    width of the template
%     td    depth of the template
%     f     filename to write to
%     t3d   size of 3d threads block (assume cube)
%
% Georgios Papamakarios (gpapamak@auth.gr)
% Electrical and Computer Engineering AUTH
% April 2011


ph = t3d + th - 1;
pw = t3d + tw - 1;
pd = t3d + td - 1;
p_size = ph*pw*pd;
t_size = th*tw*td;

fprintf(f,'template <class flccPrecision>\n');
fprintf(f,'__global__ void lcorrelation3d_kernel_%dx%dx%d (\n', th, tw, td);
fprintf(f,'\tflccPrecision *imageDev, flccPrecision *templateDev, flccPrecision *outputDev, int SW, int SD, int BW, int BD )\n');

fprintf(f,'{\n');
fprintf(f,'\tconst int block_num_x = SD/blockDim.x;\n');
fprintf(f,'\tconst int u0 = IMUL(blockDim.x, blockIdx.x%%block_num_x);\n');
fprintf(f,'\tconst int v0 = IMUL(blockDim.y, blockIdx.x/block_num_x);\n');
fprintf(f,'\tconst int w0 = IMUL(blockDim.z, blockIdx.y);\n');

fprintf(f,'\t__shared__ flccPrecision panelSh[%d];\n', p_size);
fprintf(f,'\tconst int i_idx = u0 + IMUL(v0,BD) + IMUL(w0,IMUL(BD,BW));\n');
fprintf(f,'\tfor (int k=threadIdx.z; k<%d; k+=%d) {\n', ph, t3d);
fprintf(f,'\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', pw, t3d);
fprintf(f,'\t\t\t#pragma unroll\n');
fprintf(f,'\t\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', pd, t3d);
fprintf(f,'\t\t\t\tpanelSh[i + IMUL(j, %d) + IMUL(k, %d)] = imageDev[i_idx + i + IMUL(j,BD) + IMUL(k,IMUL(BD,BW))]; } } }\n', pd, pd*pw);

fprintf(f,'\t__shared__ flccPrecision templateSh[%d];\n', t_size);
if th > t3d
	if tw > t3d
		if td > t3d
			fprintf(f,'\tfor (int k=threadIdx.z; k<%d; k+=%d) {\n', th, t3d);
			fprintf(f,'\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', tw, t3d);
			fprintf(f,'\t\t\t#pragma unroll\n');
			fprintf(f,'\t\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', td, t3d);
			fprintf(f,'\t\t\t\tconst int ijk = i + IMUL(j,%d) + IMUL(k,%d);\n', td, td*tw);
			fprintf(f,'\t\t\t\ttemplateSh[ijk] = templateDev[ijk]; } } }\n');
		else
			fprintf(f,'\tif (threadIdx.x < %d) {\n', td);
			fprintf(f,'\t\tfor (int k=threadIdx.z; k<%d; k+=%d) {\n', th, t3d);
			fprintf(f,'\t\t\t#pragma unroll\n');
			fprintf(f,'\t\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', tw, t3d);
			fprintf(f,'\t\t\t\tconst int ijk = threadIdx.x + IMUL(j,%d) + IMUL(k,%d);\n', td, td*tw);
			fprintf(f,'\t\t\t\ttemplateSh[ijk] = templateDev[ijk]; } } }\n');
		end
	else
		if td > t3d
			fprintf(f,'\tif (threadIdx.y < %d) {\n', tw);
			fprintf(f,'\t\tfor (int k=threadIdx.z; k<%d; k+=%d) {\n', th, t3d);
			fprintf(f,'\t\t\t#pragma unroll\n');
			fprintf(f,'\t\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', td, t3d);
			fprintf(f,'\t\t\t\tconst int ijk = i + IMUL(threadIdx.y,%d) + IMUL(k,%d);\n', td, td*tw);
			fprintf(f,'\t\t\t\ttemplateSh[ijk] = templateDev[ijk]; } } }\n');
		else
			fprintf(f,'\tif (threadIdx.x < %d && threadIdx.y < %d) {\n', td, tw);
			fprintf(f,'\t\t#pragma unroll\n');
			fprintf(f,'\t\tfor (int k=threadIdx.z; k<%d; k+=%d) {\n', th, t3d);
			fprintf(f,'\t\t\tconst int ijk = threadIdx.x + IMUL(threadIdx.y,%d) + IMUL(k,%d);\n', td, td*tw);
			fprintf(f,'\t\t\ttemplateSh[ijk] = templateDev[ijk]; } }\n');
		end
	end
else
	if tw > t3d
		if td > t3d
			fprintf(f,'\tif (threadIdx.z < %d) {\n', th);
			fprintf(f,'\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', tw, t3d);
			fprintf(f,'\t\t\t#pragma unroll\n');
			fprintf(f,'\t\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', td, t3d);
			fprintf(f,'\t\t\t\tconst int ijk = i + IMUL(j,%d) + IMUL(threadIdx.z,%d);\n', td, td*tw);
			fprintf(f,'\t\t\t\ttemplateSh[ijk] = templateDev[ijk]; } } }\n');
		else
			fprintf(f,'\tif (threadIdx.x < %d && threadIdx.z < %d) {\n', td, th);
			fprintf(f,'\t\t#pragma unroll\n');
			fprintf(f,'\t\tfor (int j=threadIdx.y; j<%d; j+=%d) {\n', tw, t3d);
			fprintf(f,'\t\t\tconst int ijk = threadIdx.x + IMUL(j,%d) + IMUL(threadIdx.z,%d);\n', td, td*tw);
			fprintf(f,'\t\t\ttemplateSh[ijk] = templateDev[ijk]; } }\n');
		end
	else
		if td > t3d
			fprintf(f,'\tif (threadIdx.y < %d && threadIdx.z < %d) {\n', tw, th);
			fprintf(f,'\t\t#pragma unroll\n');
			fprintf(f,'\t\tfor (int i=threadIdx.x; i<%d; i+=%d) {\n', td, t3d);
			fprintf(f,'\t\t\tconst int ijk = i + IMUL(threadIdx.y,%d) + IMUL(threadIdx.z,%d);\n', td, td*tw);
			fprintf(f,'\t\t\ttemplateSh[ijk] = templateDev[ijk]; } }\n');
		else
			fprintf(f,'\tif (threadIdx.x < %d && threadIdx.y < %d && threadIdx.z < %d) {\n', td, tw, th);
			fprintf(f,'\t\tconst int ijk = threadIdx.x + IMUL(threadIdx.y,%d) + IMUL(threadIdx.z,%d);\n', td, td*tw);
			fprintf(f,'\t\ttemplateSh[ijk] = templateDev[ijk]; }\n');
		end
	end
end

fprintf(f,'\t__syncthreads();\n');
fprintf(f,'\tregister flccPrecision corr   = 0.0;\n');
fprintf(f,'\tregister flccPrecision sum    = 0.0;\n');
fprintf(f,'\tregister flccPrecision sum_sq = 0.0;\n');
fprintf(f,'\tconst int p_idx = threadIdx.x + IMUL(threadIdx.y,%d) + IMUL(threadIdx.z,%d);\n', pd, pd*pw);
fprintf(f,'\tregister flccPrecision im;\n');

for i=0:th-1
    for j=0:tw-1
		for k=0:td-1
			fprintf(f,'\t\tim = panelSh[p_idx + %d];\n', k + j*pd + i*pd*pw);
			fprintf(f,'\t\tcorr   += im * templateSh[%d];\n', k + j*td + i*td*tw);
			fprintf(f,'\t\tsum    += im;\n');
			fprintf(f,'\t\tsum_sq += im * im;\n');
		end
    end
end

fprintf(f,'\tconst flccPrecision var = sum_sq - sum*sum/%d;\n', t_size);
fprintf(f,'\tconst int out_idx = (u0+threadIdx.x) + IMUL((v0+threadIdx.y),SD) + IMUL((w0+threadIdx.z),IMUL(SD,SW));\n');
fprintf(f,'\tif (var == 0)	outputDev[out_idx] = 0.0;\n');
fprintf(f,'\telse			outputDev[out_idx] = corr / sqrtf(var);\n');
fprintf(f,'}\n');

fprintf(f,'#ifndef NO_SINGLE_PREC \n');
fprintf(f,'template __global__ void lcorrelation3d_kernel_%dx%dx%d<float> ', th, tw, td);
fprintf(f,'( float *imageDev, float *templateDev, float *outputDev, int SW, int SD, int BW, int BD );\n');
fprintf(f,'#endif \n');
fprintf(f,'#ifndef NO_DOUBLE_PREC \n');
fprintf(f,'template __global__ void lcorrelation3d_kernel_%dx%dx%d<double> ', th, tw, td);
fprintf(f,'( double *imageDev, double *templateDev, double *outputDev, int SW, int SD, int BW, int BD );\n');
fprintf(f,'#endif \n');
fprintf(f,'\n');
