/* 
 * FILE:
 * FLCC_cuda_defs.h 
 * version 1.1
 *
 * DESCRIPTION:
 * These are the global CUDA definitions for the FLCC library source files.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 
 
#ifndef FLCC_CUDA_DEFS_H
#define FLCC_CUDA_DEFS_H
 
 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// System includes
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <cuda_runtime_api.h>
#include <cufft.h>
#include <gen_direct2_kernel.h>
#include <gen_direct3_kernel.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Macro functions and definitions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define CUDA_SAFE_CALL(call)     if ((call) != cudaSuccess)	            return FLCC_EXEC_FAILED                                         
#define CUFFT_SAFE_CALL(call)    if ((call) != CUFFT_SUCCESS)           return FLCC_EXEC_FAILED                                              
#define CHECK_ERROR              if (cudaGetLastError() != cudaSuccess) return FLCC_EXEC_FAILED                                             

#define IMUL(a, b) __mul24(a, b)

#define KERNEL_MAX_SIZE_2D 32
#define KERNEL_MAX_SIZE_3D 8


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Functions prototypes
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template <class flccPrecision>
flccResult  conv2_dev               (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int);
template <class flccPrecision>
flccResult  conv3_dev               (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  fconv2_dev              (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  cufftHandle,  cufftHandle);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  fconv3_dev              (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  cufftHandle,  cufftHandle);
template <class flccPrecision>
flccResult  conv2_testStream_dev    (flccPrecision *,  flccSize,  flccPrecision *,  flccSize);
template <class flccPrecision>
flccResult  conv3_testStream_dev    (flccPrecision *,  flccSize,  flccPrecision *,  flccSize);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  fconv2_testStream_dev   (flccPrecision *,  flccSize,  cufftComplexPrecision *,  flccSize,  cufftHandle,  cufftHandle);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  fconv3_testStream_dev   (flccPrecision *,  flccSize,  cufftComplexPrecision *,  flccSize,  cufftHandle,  cufftHandle);

template <class flccPrecision>
flccResult  lcorr2_dev              (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int);
template <class flccPrecision>
flccResult  lcorr3_dev              (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  flcorr2_dev             (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  cufftHandle,  cufftHandle);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  flcorr3_dev             (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  cufftHandle,  cufftHandle);
template <class flccPrecision>
flccResult  lcorr2_testStream_dev   (flccPrecision *,  flccSize,  flccPrecision *,  flccSize);
template <class flccPrecision>
flccResult  lcorr3_testStream_dev   (flccPrecision *,  flccSize,  flccPrecision *,  flccSize);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  flcorr2_testStream_dev  (flccPrecision *,  flccSize,  cufftComplexPrecision *,  cufftComplexPrecision *,  flccSize,  cufftHandle,  cufftHandle);
template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult  flcorr3_testStream_dev  (flccPrecision *,  flccSize,  cufftComplexPrecision *,  cufftComplexPrecision *,  flccSize,  cufftHandle,  cufftHandle);


GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   2,  2 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   3,  3 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   4,  4 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   5,  5 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   6,  6 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   7,  7 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   8,  8 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,   9,  9 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  10, 10 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  11, 11 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  12, 12 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  13, 13 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  14, 14 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  15, 15 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  16, 16 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  17, 17 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  18, 18 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  19, 19 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  20, 20 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  21, 21 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  22, 22 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  23, 23 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  24, 24 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  25, 25 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  26, 26 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  27, 27 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  28, 28 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  29, 29 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  30, 30 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  31, 31 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( convolution2d_kernel,  32, 32 );

GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  2,  2 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  3,  3 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  4,  4 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  5,  5 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  6,  6 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  7,  7 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  8,  8 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel,  9,  9 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 10, 10 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 11, 11 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 12, 12 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 13, 13 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 14, 14 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 15, 15 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 16, 16 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 17, 17 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 18, 18 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 19, 19 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 20, 20 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 21, 21 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 22, 22 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 23, 23 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 24, 24 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 25, 25 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 26, 26 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 27, 27 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 28, 28 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 29, 29 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 30, 30 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 31, 31 );
GEN_DIRECT2_KERNEL_PROTOTYPE ( lcorrelation2d_kernel, 32, 32 );

GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  2, 2, 2 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  3, 3, 3 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  4, 4, 4 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  5, 5, 5 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  6, 6, 6 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  7, 7, 7 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( convolution3d_kernel,  8, 8, 8 ); 

GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 2, 2, 2 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 3, 3, 3 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 4, 4, 4 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 5, 5, 5 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 6, 6, 6 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 7, 7, 7 );
GEN_DIRECT3_KERNEL_PROTOTYPE ( lcorrelation3d_kernel, 8, 8, 8 );

template <class flccPrecision>
__global__ void  lcorrelation2d_kernel_generic (
		flccPrecision *, 
		flccPrecision *,
		flccPrecision *,
		int, int,
		int, int);

template <class flccPrecision>
__global__ void  convolution2d_kernel_generic (
		flccPrecision *, 
		flccPrecision *,
		flccPrecision *,
		int, int,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makePermanent2d (
		cufftRealPrecision *, 
		flccPrecision *, 
		cufftRealPrecision *, 
		flccPrecision, flccPrecision,
		int, int,
		int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makeTransient2d (
		cufftRealPrecision *, 
		cufftRealPrecision *, 
		flccPrecision *,
		int, int,
		int);

template <class cufftComplexPrecision>
__global__ void  complexProduct2d (
		cufftComplexPrecision *, 
		cufftComplexPrecision *, 
		cufftComplexPrecision *,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makeLccArray2d (
		flccPrecision *, 
		cufftRealPrecision *,
		cufftRealPrecision *, 
		cufftRealPrecision *,
		int, int,
		int, int);

template <class flccPrecision>
__global__ void  lcorrelation3d_kernel_generic (
		flccPrecision *, 
		flccPrecision *,
		flccPrecision *,
		int, int, int,
		int, int,
		int, int);

template <class flccPrecision>
__global__ void  convolution3d_kernel_generic (
		flccPrecision *, 
		flccPrecision *,
		flccPrecision *,
		int, int, int,
		int, int,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  padSignal3d (
		cufftRealPrecision *, 
		flccPrecision *, 
		int, int, int,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makeConvArray3d (
		flccPrecision *, 
		cufftRealPrecision *, 
		int, int, int,	
		int, int);

template <class flccPrecision>
__global__ void  padSignal3d_both (
		flccPrecision *, 
		flccPrecision *,
		int, int, int, 
		int, int, int,
		int, int, int);

template <class flccPrecision>
__global__ void  depadSignal3d (
		flccPrecision *, 
		flccPrecision *, 
		int, int, int,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makePermanent3d (
		cufftRealPrecision *, 
		flccPrecision *, 
		cufftRealPrecision *, 
		flccPrecision, flccPrecision,
		int, int, int,
		int, int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makeTransient3d (
		cufftRealPrecision *, 
		cufftRealPrecision *, 
		flccPrecision *,
		int, int, int,
		int, int );

template <class cufftComplexPrecision>
__global__ void  complexProduct3d (
		cufftComplexPrecision *, 
		cufftComplexPrecision *, 
		cufftComplexPrecision *,
		int, int,
		int);

template <class flccPrecision, class cufftRealPrecision>
__global__ void  makeLccArray3d (
		flccPrecision *, 
		cufftRealPrecision *,
		cufftRealPrecision *,
		cufftRealPrecision *, 
		int, int, int,	
		int, int, int);

		
template <class flccPrecision>
cufftResult cufftPlan2d_R2C_t (cufftHandle *plan, int nx, int ny);

template <class flccPrecision>
cufftResult cufftPlan2d_C2R_t (cufftHandle *plan, int nx, int ny);

template <class flccPrecision>
cufftResult cufftPlan3d_R2C_t (cufftHandle *plan, int nx, int ny, int nz);

template <class flccPrecision>
cufftResult cufftPlan3d_C2R_t (cufftHandle *plan, int nx, int ny, int nz);

template <class cufftRealPrecision, class cufftComplexPrecision>
cufftResult cufftExecR2C_t (cufftHandle plan, cufftRealPrecision *idata, cufftComplexPrecision *odata);

template <class cufftRealPrecision, class cufftComplexPrecision>
cufftResult cufftExecC2R_t (cufftHandle plan, cufftComplexPrecision *idata, cufftRealPrecision *odata);


#endif // FLCC_CUDA_DEFS_H
