/* 
 * FILE:
 * FLCC_general_defs.h 
 * version 1.0
 *
 * DESCRIPTION:
 * These are the global general definitions for the FLCC library source files.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * April 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 
 
#ifndef FLCC_GENERAL_DEFS_H
#define FLCC_GENERAL_DEFS_H
 
 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// System includes
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <sys/time.h>
#include <cycle.h>
#include <pthread.h>
#include <gen_direct2_thread.h>
#include <gen_direct3_thread.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Data types definitions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

enum flccResult {
	FLCC_SUCCESS,
	FLCC_INVALID_PLAN,
	FLCC_INVALID_DIMENSION,
	FLCC_INVALID_SIZE,
	FLCC_INVALID_TYPE,
	FLCC_INVALID_PLATFORM,
	FLCC_INVALID_VALUE,
	FLCC_ALLOC_FAILED,
	FLCC_EXEC_FAILED
};

struct flccSize {
	int h;
	int w;
	int d;
};		

enum flccMethod {
	FLCC_DIRECT2_DEV,
	FLCC_DIRECT3_DEV,
	FLCC_FAST2_DEV,
	FLCC_FAST3_DEV,
	FLCC_DIRECT2_HOST,
	FLCC_DIRECT3_HOST,
	FLCC_FAST2_HOST,
	FLCC_FAST3_HOST,
	FLCC_NULL_METHOD
};

enum flccType {
	FLCC_SINGLE,
	FLCC_STREAM
};

enum flccPlatform {
	FLCC_HOST,
	FLCC_DEVICE,
	FLCC_ANY
}; 		

struct flccPlan {
	flccSize        imageSize;
	flccSize        templateSize;
	int             dimension;			
	flccMethod      method;
	flccType        type;
	flccPlatform    platform; 	
	int             numCpuCores;
	int             numGpuDevices;		
	void *          c2r_plan_dev;
	void *          r2c_plan_dev;
	void *          c2r_plan_host_diff;
	void *          r2c_plan_host_diff;
};

template <class flccPrecision>
struct padParm {
	int            initialPosition;
	int            offset;
	flccPrecision  *padded;
	flccPrecision  *unpadded;
	int            th, tw, td;
	int            sh, sw, sd;
	int            iw, id;
	int            bw, bd;
};

template <class flccPrecision>
struct corrParm {
	int            initialPosition;
	int            offset;
	flccPrecision  *image;
	flccPrecision  *templat;
	flccPrecision  *lcc;
	int            th, tw, td;
	int            sw, sd;
	int            bw, bd;
};


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Macro functions and definitions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define MALLOC_SAFE_CALL(call)   if ((call) == NULL)                    return FLCC_ALLOC_FAILED   
#define PTHREAD_SAFE_CALL(call)  if ((call) != 0)                       return FLCC_EXEC_FAILED          

#define MAX_TIME DBL_MAX

#define THREAD_MAX_SIZE_2D 32
#define THREAD_MAX_SIZE_3D 10       


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Functions prototypes
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef NO_SINGLE_PREC
flccResult    lcorr_plan        (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    lcorr_plan_f      (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    lcorr_exec        (flccPlan,   float *,   float *,   float *,  int);
flccResult    lcorr_exec_f      (flccPlan,   float *,   float *,   float *,  int);
flccResult    lcorr_destroy     (flccPlan *);
flccResult    lcorr_destroy_f   (flccPlan *);

flccResult    conv_plan         (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    conv_plan_f       (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    conv_exec         (flccPlan,   float *,   float *,   float *,  int);
flccResult    conv_exec_f       (flccPlan,   float *,   float *,   float *,  int);
flccResult    conv_destroy      (flccPlan *);
flccResult    conv_destroy_f    (flccPlan *);
#endif

#ifndef NO_DOUBLE_PREC
flccResult    lcorr_plan_d      (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    lcorr_exec_d      (flccPlan,  double *,  double *,  double *,  int);
flccResult    lcorr_destroy_d   (flccPlan *);

flccResult    conv_plan_d       (flccPlan *,  int,  flccSize,  flccSize,  flccType,  flccPlatform);
flccResult    conv_exec_d       (flccPlan,  double *,  double *,  double *,  int);
flccResult    conv_destroy_d    (flccPlan *);
#endif

void *        flcc_malloc       (size_t);
void          flcc_free         (void *);


template <class flccPrecision>
flccResult    conv2_host                      (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  int);
template <class flccPrecision>
flccResult    conv3_host                      (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  int);
template <class flccPrecision>
flccResult    lcorr2_host                     (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  int);
template <class flccPrecision>
flccResult    lcorr3_host                     (flccPrecision *,  flccSize,  flccPrecision *,  flccSize,  flccPrecision *,  int,  int);


template <class flccPrecision>
void          randMatrix                      (flccPrecision *,  int);
template <class flccPrecision>
flccPrecision         findMean                (flccPrecision *,  int);
template <class flccPrecision>
flccPrecision         findStd                 (flccPrecision,  flccPrecision *,  int);
template <class flccPrecision>
void          normalize                       (flccPrecision,  flccPrecision,  flccPrecision *,  int);
int           alignSize2d                     (int);
int           alignSize3d                     (int);

template <class flccPrecision>
void *        padSignal2d_both                (void *);
template <class flccPrecision>
void *        padSignal3d_both                (void *);
template <class flccPrecision>
void *        convolution2d_thread_generic    (void *);
template <class flccPrecision>
void *        convolution3d_thread_generic    (void *);
template <class flccPrecision>
void *        lcorrelation2d_thread_generic   (void *);
template <class flccPrecision>
void *        lcorrelation3d_thread_generic   (void *);
 

GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   2,  2 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   3,  3 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   4,  4 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   5,  5 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   6,  6 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   7,  7 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   8,  8 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,   9,  9 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  10, 10 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  11, 11 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  12, 12 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  13, 13 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  14, 14 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  15, 15 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  16, 16 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  17, 17 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  18, 18 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  19, 19 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  20, 20 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  21, 21 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  22, 22 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  23, 23 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  24, 24 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  25, 25 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  26, 26 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  27, 27 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  28, 28 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  29, 29 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  30, 30 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  31, 31 );
GEN_DIRECT2_THREAD_PROTOTYPE ( convolution2d_thread,  32, 32 );

GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  2,  2 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  3,  3 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  4,  4 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  5,  5 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  6,  6 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  7,  7 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  8,  8 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread,  9,  9 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 10, 10 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 11, 11 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 12, 12 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 13, 13 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 14, 14 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 15, 15 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 16, 16 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 17, 17 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 18, 18 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 19, 19 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 20, 20 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 21, 21 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 22, 22 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 23, 23 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 24, 24 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 25, 25 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 26, 26 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 27, 27 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 28, 28 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 29, 29 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 30, 30 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 31, 31 );
GEN_DIRECT2_THREAD_PROTOTYPE ( lcorrelation2d_thread, 32, 32 );


GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   2,  2,  2 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   3,  3,  3 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   4,  4,  4 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   5,  5,  5 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   6,  6,  6 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   7,  7,  7 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   8,  8,  8 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,   9,  9,  9 );
GEN_DIRECT3_THREAD_PROTOTYPE ( convolution3d_thread,  10, 10, 10 ); 

GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  2,  2,  2 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  3,  3,  3 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  4,  4,  4 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  5,  5,  5 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  6,  6,  6 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  7,  7,  7 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  8,  8,  8 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread,  9,  9,  9 );
GEN_DIRECT3_THREAD_PROTOTYPE ( lcorrelation3d_thread, 10, 10, 10 );


#ifndef NO_FFTW
	#ifndef NO_CUDA	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult conv_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult lcorr_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult conv_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *conv, int imCnt);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult lcorr_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *lcc, int imCnt);
	
	#else
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	flccResult conv_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	flccResult lcorr_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	flccResult conv_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *conv, int imCnt);
	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	flccResult lcorr_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *lcc, int imCnt);
	
	#endif
	
template <class flccPrecision, class fftw_plan_t>
flccResult conv_destroy_t (flccPlan *plan);

template <class flccPrecision, class fftw_plan_t>
flccResult lcorr_destroy_t (flccPlan *plan);
	
#else
	#ifndef NO_CUDA	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult conv_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult lcorr_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult conv_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *conv, int imCnt);
	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	flccResult lcorr_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *lcc, int imCnt);
	
	#else
	template <class flccPrecision>
	flccResult conv_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision>
	flccResult lcorr_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
	
	template <class flccPrecision>
	flccResult conv_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *conv, int imCnt);
	
	template <class flccPrecision>
	flccResult lcorr_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *lcc, int imCnt);
	
	#endif

template <class flccPrecision>
flccResult conv_destroy_t (flccPlan *plan);

template <class flccPrecision>
flccResult lcorr_destroy_t (flccPlan *plan);
#endif


#endif // FLCC_GENERAL_DEFS_H
