/* 
 * FILE:
 * gen_direct2_kernel.h 
 * version 1.0
 *
 * DESCRIPTION:
 * It includes macro "function-generators" which generate
 * the optimized GPU kernels for the 2d direct method.
 * NOTE: In current FLCC version the kernels are generated by m-files instead.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * July 2010
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


/////////////////
#define GEN_DIRECT2_KERNEL_PROTOTYPE( FUNCTION_NAME, TH, TW )				\
template <class flccPrecision>												\
__global__ void FUNCTION_NAME ## _ ## TH ## x ## TW (						\
		flccPrecision *imageDev, 											\
		flccPrecision *templateDev,											\
		flccPrecision *outputDev,											\
		int SW, int BW )

/////////////////
#define GEN_DIRECT2_KERNEL_INSTANCE( FUNCTION_NAME, TH, TW, type )			\
template __global__															\
void FUNCTION_NAME ## _ ## TH ## x ## TW <type> (							\
		type *imageDev, 													\
		type *templateDev,													\
		type *outputDev,													\
		int SW, int BW );
		
/////////////////
#define GEN_DIRECT2_KERNEL_VAR_DEFINITIONS( TH, TW ) 						\
	const int u0 = IMUL(blockIdx.x, blockDim.x);							\
	const int v0 = IMUL(blockIdx.y, blockDim.y);							\
	const int PH = THREADS_2D+(TH)-1;										\
	const int PW = THREADS_2D+(TW)-1;										\
	const int P_SIZE = PH*PW;												\
	const int T_SIZE = (TH)*(TW);

/////////////////
#define GEN_DIRECT2_KERNEL_MOVE_TO_SHARED_MEM( TH, TW ) 					\
	__shared__ flccPrecision panelSh[P_SIZE];								\
	const int i_idx = u0 + IMUL(v0, BW);									\
	for (int j=threadIdx.y; j<PH; j+=THREADS_2D) {							\
        for (int i=threadIdx.x; i<PW; i+=THREADS_2D) {						\
			panelSh[IMUL(j, PW) + i] = 										\
				imageDev[i_idx + IMUL(j, BW) + i]; } }						\
	__shared__ flccPrecision templateSh[T_SIZE];							\
	for (int j=threadIdx.y; j<(TH); j+=THREADS_2D) {						\
        for (int i=threadIdx.x; i<(TW); i+=THREADS_2D) {					\
			const int jtwi = IMUL(j, (TW)) + i;								\
			templateSh[jtwi] = templateDev[jtwi]; } }						\
    __syncthreads();
		
/////////////////	
#define GEN_DIRECT2_KERNEL_CONV( TH, TW ) 									\
GEN_DIRECT2_KERNEL_PROTOTYPE( convolution2d_kernel, TH, TW )				\
{																			\
	GEN_DIRECT2_KERNEL_VAR_DEFINITIONS( TH, TW ) 							\
	GEN_DIRECT2_KERNEL_MOVE_TO_SHARED_MEM( TH, TW ) 						\
	register flccPrecision conv = 0.0;										\
	int p_idx = threadIdx.x + IMUL(threadIdx.y, PW);						\
	int t_idx = T_SIZE - 1;													\
	for (int j=0; j<(TH); j++) {											\
		for (int i=0; i<(TW); i++) {										\
			conv += panelSh[p_idx + i] * templateSh[t_idx - i]; }			\
		p_idx += PW; t_idx -= (TW); }										\
	outputDev[(u0+threadIdx.x) + IMUL((v0+threadIdx.y), SW)] = conv;		\
}

/////////////////	
#define GEN_DIRECT2_KERNEL_CONV_INSTANCE( TH, TW, type )					\
GEN_DIRECT2_KERNEL_INSTANCE( convolution2d_kernel, TH, TW, type )

/////////////////
#define GEN_DIRECT2_KERNEL_LCORR( TH, TW ) 									\
GEN_DIRECT2_KERNEL_PROTOTYPE( lcorrelation2d_kernel, TH, TW )				\
{																			\
	GEN_DIRECT2_KERNEL_VAR_DEFINITIONS( TH, TW ) 							\
	GEN_DIRECT2_KERNEL_MOVE_TO_SHARED_MEM( TH, TW ) 						\
	register flccPrecision corr   = 0.0;									\
	register flccPrecision sum    = 0.0;									\
	register flccPrecision sum_sq = 0.0;									\
	int p_idx = threadIdx.x + IMUL(threadIdx.y, PW);						\
	int t_idx = 0;															\
	for (int j=0; j<(TH); j++) {											\
		for (int i=0; i<(TW); i++) {										\
			const flccPrecision im = panelSh[p_idx + i];					\
			corr   += im * templateSh[t_idx + i];							\
			sum    += im;													\
			sum_sq += im * im; } 											\
		p_idx += PW; t_idx += (TW); }										\
	const flccPrecision var = sum_sq - sum*sum/T_SIZE;						\
	const int out_idx = (u0+threadIdx.x) + IMUL((v0+threadIdx.y), SW);		\
	if (var == 0)	outputDev[out_idx] = 0.0;								\
	else			outputDev[out_idx] = corr / sqrtf(var);					\
}

/////////////////	
#define GEN_DIRECT2_KERNEL_LCORR_INSTANCE( TH, TW, type )					\
GEN_DIRECT2_KERNEL_INSTANCE( lcorrelation2d_kernel, TH, TW, type )
