/* 
 * FILE:
 * gen_direct3_kernel.h 
 * version 1.0
 *
 * DESCRIPTION:
 * It includes macro "function-generators" which generate
 * the optimized GPU kernels for the 3d direct method.
 * NOTE: In current FLCC version the kernels are generated by m-files instead.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * July 2010
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 
 
/////////////////
#define GEN_DIRECT3_KERNEL_PROTOTYPE( FUNCTION_NAME, TH, TW, TD )			\
template <class flccPrecision>												\
__global__ void FUNCTION_NAME ## _ ## TH ## x ## TW ## x ## TD (			\
		flccPrecision *imageDev, 											\
		flccPrecision *templateDev,											\
		flccPrecision *outputDev,											\
		int SW, int SD,														\
		int BW, int BD )

/////////////////
#define GEN_DIRECT3_KERNEL_INSTANCE( FUNCTION_NAME, TH, TW, TD, type )		\
template __global__ 														\
void FUNCTION_NAME ## _ ## TH ## x ## TW ## x ## TD <type> (				\
		type *imageDev, 													\
		type *templateDev,													\
		type *outputDev,													\
		int SW, int SD,														\
		int BW, int BD )
		
/////////////////
#define GEN_DIRECT3_KERNEL_VAR_DEFINITIONS( TH, TW, TD )					\
	const int block_num_x = SD/blockDim.x;									\
    const int u0 = IMUL(blockDim.x, blockIdx.x%block_num_x);				\
    const int v0 = IMUL(blockDim.y, blockIdx.x/block_num_x);				\
    const int w0 = IMUL(blockDim.z, blockIdx.y);							\
	const int PH = THREADS_3D+(TH)-1;										\
	const int PW = THREADS_3D+(TW)-1;										\
	const int PD = THREADS_3D+(TD)-1;										\
	const int P_SIZE = PH*PW*PD;											\
	const int T_SIZE = (TH)*(TW)*(TD);

/////////////////
#define GEN_DIRECT3_KERNEL_MOVE_TO_SHARED_MEM( TH, TW, TD )					\
	__shared__ flccPrecision panelSh[P_SIZE];								\
	const int i_idx = u0 + IMUL(v0,BD) + IMUL(w0,IMUL(BD,BW));				\
	for (int k=threadIdx.z; k<PH; k+=THREADS_3D) {							\
        for (int j=threadIdx.y; j<PW; j+=THREADS_3D) {						\
			for (int i=threadIdx.x; i<PD; i+=THREADS_3D) {					\
				panelSh[i + IMUL(j, PD) + IMUL(k, IMUL(PD, PW))] = 			\
					imageDev[i_idx + i + IMUL(j,BD) + IMUL(k,IMUL(BD,BW))];	\
			} } }                 											\
	__shared__ flccPrecision templateSh[T_SIZE];							\
	for (int k=threadIdx.z; k<(TH); k+=THREADS_3D) {						\
        for (int j=threadIdx.y; j<(TW); j+=THREADS_3D) {					\
			for (int i=threadIdx.x; i<(TD); i+=THREADS_3D) {				\
				const int ijk = i + IMUL(j,(TD)) + IMUL(k,IMUL((TD),(TW)));	\
				templateSh[ijk] = templateDev[ijk]; 						\
			} } }                 											\
    __syncthreads();

/////////////////	
#define GEN_DIRECT3_KERNEL_CONV( TH, TW, TD )								\
GEN_DIRECT3_KERNEL_PROTOTYPE( convolution3d_kernel, TH, TW, TD )			\
{																			\
	GEN_DIRECT3_KERNEL_VAR_DEFINITIONS( TH, TW, TD )						\
	GEN_DIRECT3_KERNEL_MOVE_TO_SHARED_MEM( TH, TW, TD )						\
	register flccPrecision conv = 0.0;										\
	int p_idx = threadIdx.x + IMUL(threadIdx.y,PD) + 						\
		IMUL(threadIdx.z,IMUL(PD,PW));										\
	int t_idx = T_SIZE - 1;													\
	for (int k=0; k<(TH); k++) {											\
		for (int j=0; j<(TW); j++) {										\
			for (int i=0; i<(TD); i++) {									\
				conv += panelSh[p_idx+i] * templateSh[t_idx-i]; } 			\
			p_idx += PD; t_idx -= (TD); } 									\
		p_idx += PD*(PW-(TD)); }											\
	outputDev[(u0+threadIdx.x) + IMUL((v0+threadIdx.y),SD) + 				\
		IMUL((w0+threadIdx.z),IMUL(SD,SW))] = conv;							\
}

/////////////////
#define GEN_DIRECT3_KERNEL_CONV_INSTANCE( TH, TW, TD, type )				\
GEN_DIRECT3_KERNEL_INSTANCE( convolution3d_kernel, TH, TW, TD, type )

/////////////////
#define GEN_DIRECT3_KERNEL_LCORR( TH, TW, TD )								\
GEN_DIRECT3_KERNEL_PROTOTYPE( lcorrelation3d_kernel, TH, TW, TD )			\
{																			\
	GEN_DIRECT3_KERNEL_VAR_DEFINITIONS( TH, TW, TD )						\
	GEN_DIRECT3_KERNEL_MOVE_TO_SHARED_MEM( TH, TW, TD )						\
	register flccPrecision corr   = 0.0;									\
	register flccPrecision sum    = 0.0;									\
	register flccPrecision sum_sq = 0.0;									\
	int p_idx = threadIdx.x + IMUL(threadIdx.y,PD) + 						\
		IMUL(threadIdx.z,IMUL(PD,PW));										\
	int t_idx = 0;															\
	for (int k=0; k<(TH); k++) {											\
		for (int j=0; j<(TW); j++) {										\
			for (int i=0; i<(TD); i++) {									\
				const flccPrecision im = panelSh[p_idx+i];					\
				corr   += im * templateSh[t_idx+i];	  						\
				sum    += im;												\
				sum_sq += im*im; }											\
			p_idx += PD; t_idx += (TD); }									\
		p_idx += PD*(PW-(TD)); }											\
	const flccPrecision var = sum_sq - sum*sum/T_SIZE;						\
	const int out_idx = (u0+threadIdx.x) + IMUL((v0+threadIdx.y),SD) + 		\
			IMUL((w0+threadIdx.z),IMUL(SD,SW));								\
	if (var == 0)	outputDev[out_idx] = 0.0;								\
	else			outputDev[out_idx] = corr / sqrtf(var);					\
}

/////////////////
#define GEN_DIRECT3_KERNEL_LCORR_INSTANCE( TH, TW, TD, type )				\
GEN_DIRECT3_KERNEL_INSTANCE( lcorrelation3d_kernel, TH, TW, TD, type )	
