/* 
 * FILE:
 * lcorr_exec.c
 * version 1.1
 *
 * DESCRIPTION:
 * Matlab mex function for lcorr_exec.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * January 2013
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 */

#include "mex.h"
#include "matrix.h"
#include <flcc.h>
#include <string.h>

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	if (nrhs!=3) {mexErrMsgTxt("Incorrect number of inputs.");}
	if (nlhs!=1) {mexErrMsgTxt("Incorrect number of outputs.");}
	
	// get FLCC plan
	flccPlan **plan = (flccPlan **)mxGetData(prhs[0]);
	
	// get image
	int imCount;
	mwSize dim;
	const mwSize *imDims;
	void *image;
	mxClassID imclass, argclass = mxGetClassID(prhs[1]);
	int elemSize;
	if (argclass==mxSINGLE_CLASS || argclass==mxDOUBLE_CLASS) {
		imCount = 1;  
		dim = mxGetNumberOfDimensions(prhs[1]);
		imDims = mxGetDimensions(prhs[1]);
		image = mxGetData(prhs[1]);
		imclass = argclass;
	} 
	else if (argclass==mxCELL_CLASS) {
		imCount = mxGetNumberOfElements(prhs[1]);
		const mxArray *image0 = mxGetCell(prhs[1],0);
		const int imSize = mxGetNumberOfElements(image0);
		dim = mxGetNumberOfDimensions(image0);
		imDims = mxGetDimensions(image0);
		imclass = mxGetClassID(image0);
		switch (imclass) {
		case mxSINGLE_CLASS:
			elemSize = sizeof(float);
			break;
		case mxDOUBLE_CLASS:
			elemSize = sizeof(double);
			break;
		default:
			mexErrMsgTxt("Image array is not of a valid type."); 
		}
		image = flcc_malloc(imCount*imSize*elemSize);
		if (image == NULL) {mexErrMsgTxt("Memory allocation on host failed.");}
		memcpy(image, mxGetData(image0), imSize*elemSize);
		for (int i=1; i<imCount; i++) {
			memcpy((char *)image+i*imSize*elemSize, mxGetData(mxGetCell(prhs[1],i)), imSize*elemSize);
		}
	} 
	else {
		mexErrMsgTxt("Image array is not of a valid type."); 
	}
	
	// get template
	if (dim != mxGetNumberOfDimensions(prhs[2])) {mexErrMsgTxt("Image and template dimensions don't match.");}
	if (imclass != mxGetClassID(prhs[2]))        {mexErrMsgTxt("Image and template are not of the same class.");}
	const mwSize *temDims = mxGetDimensions(prhs[2]);
	const void *templat = mxGetData(prhs[2]);
	
	// create output
	mwSize convDims[3];
	int convSize;
	switch (dim) {
	case 2:
		convDims[0] = imDims[0] + temDims[0] - 1;
		convDims[1] = imDims[1] + temDims[1] - 1;
		convSize = convDims[0]*convDims[1];
		break;
	case 3:
		convDims[0] = imDims[0] + temDims[0] - 1;
		convDims[1] = imDims[1] + temDims[1] - 1;
		convDims[2] = imDims[2] + temDims[2] - 1;
		convSize = convDims[0]*convDims[1]*convDims[2];
		break;
	default:
		mexErrMsgTxt("Incorrect dimensions."); 
	}
	void *conv;
	if (argclass==mxSINGLE_CLASS || argclass==mxDOUBLE_CLASS) {
		plhs[0] = mxCreateNumericArray(dim, convDims, imclass, mxREAL);
		conv = mxGetData(plhs[0]);
	}
	else if (argclass==mxCELL_CLASS) {
		plhs[0] = mxCreateCellArray(mxGetNumberOfDimensions(prhs[1]), mxGetDimensions(prhs[1]));
		conv = flcc_malloc(imCount*convSize*elemSize);
		if (conv == NULL) {mexErrMsgTxt("Memory allocation on host failed.");}
	}
	else {
		mexErrMsgTxt("Image array is not of a valid type."); 
	}
	
	// execute
	flccResult rc;
	switch (imclass) {
	case mxSINGLE_CLASS:
		#ifndef NO_SINGLE_PREC
		rc = lcorr_exec_f(**plan, (float *)image, (float *)templat, (float *)conv, imCount);
		#else
		mexErrMsgTxt("Single precision is not installed.");
		#endif
		break;
	case mxDOUBLE_CLASS:
		#ifndef NO_DOUBLE_PREC
		rc = lcorr_exec_d(**plan, (double *)image, (double *)templat, (double *)conv, imCount);
		#else
		mexErrMsgTxt("Double precision is not installed.");
		#endif
		break;
	default:
		mexErrMsgTxt("Image array is not of a valid type."); 
	}	
	switch (rc)	{
	case FLCC_SUCCESS:
		break;
	case FLCC_INVALID_PLAN:
		mexErrMsgTxt("Invalid plan.");
		break;
	case FLCC_INVALID_VALUE:
		mexErrMsgTxt("Invalid arrays.");
		break;
	case FLCC_ALLOC_FAILED:
		mexErrMsgTxt("FLCC failed to allocate memory.");
		break;
	default:
		mexErrMsgTxt("FLCC execution failed.");
	}
	if (argclass==mxCELL_CLASS) {
		for (int i=0; i<imCount; i++) {
			mxSetCell(plhs[0], i, mxCreateNumericArray(dim, convDims, imclass, mxREAL));
			memcpy(mxGetData(mxGetCell(plhs[0],i)), (char *)conv+i*convSize*elemSize, convSize*elemSize);
		}
		flcc_free(image);
		flcc_free(conv);
	}
}
