/* 
 * FILE:
 * lcorr_plan.c
 * version 1.1
 *
 * DESCRIPTION:
 * Matlab mex function for lcorr_plan.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * January 2013
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 */

#include "mex.h"
#include "matrix.h"
#include <flcc.h>
#include <string.h>
#include <stdlib.h>

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	if (nrhs<2 || nrhs>5) {mexErrMsgTxt("Incorrect number of inputs.");}
	if (nlhs!=1)          {mexErrMsgTxt("Incorrect number of outputs.");}
	
	// get image and template dimensions
	flccSize imSize, temSize;
	double *bufDims;
	int dim = mxGetNumberOfElements(prhs[0]);
	if (dim != mxGetNumberOfElements(prhs[1])) {mexErrMsgTxt("Image and template dimensions don't match.");}
	switch (dim) {
	case 2:
		bufDims = (double *)mxGetData(prhs[0]); 
		imSize.w = (int)bufDims[0]; 
		imSize.h = (int)bufDims[1]; 
		bufDims = (double *)mxGetData(prhs[1]);
		temSize.w = (int)bufDims[0];
		temSize.h = (int)bufDims[1];
		break;
	case 3:
		bufDims = (double *)mxGetData(prhs[0]); 
		imSize.d = (int)bufDims[0]; 
		imSize.w = (int)bufDims[1]; 
		imSize.h = (int)bufDims[2]; 
		bufDims = (double *)mxGetData(prhs[1]);
		temSize.d = (int)bufDims[0];
		temSize.w = (int)bufDims[1];
		temSize.h = (int)bufDims[2];
		break;	
	default:
		mexErrMsgTxt("Incorrect number of dimensions."); 
	}
	
	// get FLCC type
	flccType type;
	char bufStr[12];
	if (nrhs<3) {type = FLCC_SINGLE;}
	else {
		mxGetString(prhs[2], bufStr, 12);
		if      (!strcmp(bufStr, "FLCC_SINGLE")) {type = FLCC_SINGLE;}
		else if (!strcmp(bufStr, "FLCC_STREAM")) {type = FLCC_STREAM;}
		else {mexErrMsgTxt("Incorrect FLCC type.");}
	}
	
	// get FLCC platform
	flccPlatform platform;
	if (nrhs<4) {platform = FLCC_ANY;}
	else {
		mxGetString(prhs[3], bufStr, 12);
		if      (!strcmp(bufStr, "FLCC_HOST"))   {platform = FLCC_HOST;}
		else if (!strcmp(bufStr, "FLCC_DEVICE")) {platform = FLCC_DEVICE;}
		else if (!strcmp(bufStr, "FLCC_ANY"))    {platform = FLCC_ANY;}
		else {mexErrMsgTxt("Incorrect FLCC platform.");}
	}
	
	// create FLCC plan
	plhs[0] = mxCreateNumericMatrix(1, 1, mxUINT64_CLASS, mxREAL);
	flccPlan **plan = (flccPlan **)mxGetData(plhs[0]);
	*plan = (flccPlan *)malloc(sizeof(flccPlan));
	if (*plan == NULL) {mexErrMsgTxt("Memory allocation on host failed.");}
	
	// do the planning
	flccResult rc;
	if (nrhs<5) {
		#ifndef NO_SINGLE_PREC
		rc = lcorr_plan(*plan, dim, imSize, temSize, type, platform);
		#else
		mexErrMsgTxt("Single precision is not installed.");
		#endif
	}
	else {
		mxGetString(prhs[4], bufStr, 12);
		if (!strcmp(bufStr, "single")) {
			#ifndef NO_SINGLE_PREC
			rc = lcorr_plan_f(*plan, dim, imSize, temSize, type, platform);
			#else
			mexErrMsgTxt("Single precision is not installed.");
			#endif
		}
		else if (!strcmp(bufStr, "double")) {
			#ifndef NO_DOUBLE_PREC
			rc = lcorr_plan_d(*plan, dim, imSize, temSize, type, platform);
			#else
			mexErrMsgTxt("Double precision is not installed.");
			#endif
		}
		else {mexErrMsgTxt("Incorrect precision.");}
	}
	switch (rc)	{
	case FLCC_SUCCESS:
		break;
	case FLCC_INVALID_DIMENSION:
		mexErrMsgTxt("Incorrect number of dimensions.");
		break;
	case FLCC_INVALID_SIZE:
		mexErrMsgTxt("Incorrect array size.");
		break;
	case FLCC_INVALID_TYPE:
		mexErrMsgTxt("Incorrect FLCC type.");
		break;
	case FLCC_INVALID_PLATFORM:
		mexErrMsgTxt("Incorrect FLCC platform.");
		break;
	case FLCC_ALLOC_FAILED:
		mexErrMsgTxt("FLCC failed to allocate memory.");
		break;
	default:
		mexErrMsgTxt("FLCC planning failed.");
	}
}
