function [time, err] = conv_test_single(imgsize, temsize, platform, precision)
% [time, err] = conv_test_single(imgsize, temsize, platform, precision)
% Tests the MATLAB routines for convolution on a single image.
%
% INPUTS:
%     imgsize    array of image size
%     temsize    array of template size
%     platform   execution platform: 'FLCC_HOST', 'FLCC_DEVICE' or 'FLCC_ANY'
%                (optional, default 'FLCC_ANY')
%     precision  data precision: 'single' or 'double'
%                (optional, default 'single')
%
% OUTPUTS:
%     time       array with execution times
%     err        array with errors
%
% Georgios Papamakarios (gpapamak@auth.gr)
% Electrical and Computer Engineering AUTH
% January 2013


% set to true to test MATLAB on GPU
do_matlab_on_gpu = false;

% default arguments
if nargin<4
	precision = 'single';
end
if nargin<3
	platform = 'FLCC_ANY';
end
if nargin<2
	error('Incorrect number of arguments.');
end

% reset the state of the random number generator
RandStream.setGlobalStream(RandStream('mt19937ar', 'seed', sum(100*clock)));

% include mex/bin
addpath([pwd '/../bin']);

% data
img = rand(imgsize);
tem = rand(temsize);
if strcmp(precision,'single')
	img = single(img);
	tem = single(tem);
end

% FLCC
plan = conv_plan(size(img), size(tem), 'FLCC_SINGLE', platform, precision);
tic;
con1 = conv_exec(plan, img, tem); 
time(1) = toc;
conv_destroy(plan, precision);
fprintf('FLCC:\t%f sec\n', time(1));

% MATLAB
tic;
con2 = convn(img, tem);
time(2) = toc;
if strcmp(precision,'single')
	con2 = single(con2);
end
err(1) = norm(con1(:)-con2(:), 'fro');
fprintf('MATLAB:\t%f sec\n', time(2));
fprintf('Error with FLCC:\t%f\n', err(1));

if ~do_matlab_on_gpu
	return
end

% MATLAB ON GPU
reset(gpuDevice(2));
tic;
con3 = gather(convn(gpuArray(img), gpuArray(tem)));
time(3) = toc;
if strcmp(precision,'single')
	con3 = single(con3);
end
err(2) = norm(con1(:)-con3(:), 'fro');
err(3) = norm(con2(:)-con3(:), 'fro');
fprintf('MATLAB ON GPU:\t%f sec\n', time(3));
fprintf('Error with FLCC:\t%f\n', err(2));
fprintf('Error with MATLAB:\t%f\n', err(3));

