function [time, err] = conv_test_stream(imgsize, temsize, imcount, platform, precision)
% [time, err] = conv_test_stream(imgsize, temsize, imcount, platform, precision)
% Tests the MATLAB routines for convolution on a stream of images.
%
% INPUTS:
%     imgsize    array of image size
%     temsize    array of template size
%     imcount    number of images
%     platform   execution platform: 'FLCC_HOST', 'FLCC_DEVICE' or 'FLCC_ANY'
%                (optional, default 'FLCC_ANY')
%     precision  data precision: 'single' or 'double'
%                (optional, default 'single')
%
% OUTPUTS:
%     time       array with execution times
%     err        array with errors
%
% Georgios Papamakarios (gpapamak@auth.gr)
% Electrical and Computer Engineering AUTH
% January 2013


% set to true to test MATLAB on GPU
do_matlab_on_gpu = false;

% default arguments
if nargin<5
	precision = 'single';
end
if nargin<4
	platform = 'FLCC_ANY';
end
if nargin<3
	error('Incorrect number of arguments.');
end

% reset the state of the random number generator
RandStream.setGlobalStream(RandStream('mt19937ar', 'seed', sum(100*clock)));

% include mex/bin
addpath([pwd '/../bin']);

% data
img = cell(1,imcount);
for i=1:imcount
	img{i} = rand(imgsize);
end
tem = rand(temsize);
if strcmp(precision,'single')
	for i=1:imcount
		img{i} = single(img{i});
	end
	tem = single(tem);
end

% FLCC
plan = conv_plan(size(img{1}), size(tem), 'FLCC_STREAM', platform, precision);
tic;
con1 = conv_exec(plan, img, tem); 
time(1) = toc;
conv_destroy(plan, precision);
fprintf('FLCC:\t%f sec\n', time(1));

% MATLAB
con2 = cell(1,imcount);
tic;
for i=1:imcount
	con2{i} = convn(img{i}, tem);
end
time(2) = toc;
if strcmp(precision,'single')
	for i=1:imcount
		con2{i} = single(con2{i});
	end
end
err(1) = 0;
for i=1:imcount
	err(1) = err(1) + norm(con1{i}(:)-con2{i}(:), 'fro')^2;
end
err(1) = sqrt(err(1));
fprintf('MATLAB:\t%f sec\n', time(2));
fprintf('Error with FLCC:\t%f\n', err(1));

if ~do_matlab_on_gpu
	return
end

% MATLAB ON GPU
reset(gpuDevice(2));
con3 = cell(1,imcount);
tic;
tem_d = gpuArray(tem);
for i=1:imcount
	con3{i} = gather(convn(gpuArray(img{i}), tem_d));
end
time(3) = toc;
if strcmp(precision,'single')
	for i=1:imcount
		con3{i} = single(con3{i});
	end
end
err(2) = 0;
err(3) = 0;
for i=1:imcount
	err(2) = err(2) + norm(con1{i}(:)-con3{i}(:), 'fro')^2;
	err(3) = err(3) + norm(con2{i}(:)-con3{i}(:), 'fro')^2;
end
err(2) = sqrt(err(2));
err(3) = sqrt(err(3));
fprintf('MATLAB ON GPU:\t%f sec\n', time(3));
fprintf('Error with FLCC:\t%f\n', err(2));
fprintf('Error with MATLAB:\t%f\n', err(3));

