/* 
 * FILE:
 * lcorr_exec_t.c
 * version 1.5
 *
 * DESCRIPTION:
 * This function computes the LCC distribution of an image with a template according to the plan specified.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Faculty of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * April 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#ifndef NO_CUDA	
#include <FLCC_cuda_defs.h>
#endif
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif

// Template function for the execution of LCC problem using a pre-made plan
#ifndef NO_FFTW
	#ifndef NO_CUDA	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	#else
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	#endif
#else
	#ifndef NO_CUDA	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	#else
	template <class flccPrecision>
	#endif
#endif
flccResult lcorr_exec_t (flccPlan plan, flccPrecision *image, flccPrecision *templat, flccPrecision *lcc, int imCnt)
{
	int imCount; // Number of images to be processed along with a template to calculate their LCC
	flccResult rc;
	#ifndef NO_CUDA
	cufftHandle *c2r_plan_dev;
	cufftHandle *r2c_plan_dev;
	#endif
	#ifndef NO_FFTW
	fftw_plan_t *c2r_plan_host_diff;
	fftw_plan_t *r2c_plan_host_diff;
	#endif

	// Check validity of arguments
	if (image==NULL || templat==NULL || lcc==NULL)
		return FLCC_INVALID_VALUE;
		
	// Check type
	switch (plan.type)
	{
	case FLCC_SINGLE:
		imCount = 1; // Number of images is set to one to avoid processing of garbage and logical error
		break;
		
	case FLCC_STREAM:
		if (imCnt<1)
			return FLCC_INVALID_SIZE; // Check validity of number of images
			
		imCount = imCnt;
		break;
		
	default:
		return FLCC_INVALID_PLAN;
	}

	// Execute according to plan
	switch (plan.method)
	{
	
	#ifndef NO_CUDA	
	
	case FLCC_DIRECT2_DEV:
		// Launch function
		rc = lcorr2_dev<flccPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, imCount );
					
		break;
		
	case FLCC_DIRECT3_DEV:
		// Launch function
		rc = lcorr3_dev<flccPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, imCount );
					
		break;

	case FLCC_FAST2_DEV:
	
		c2r_plan_dev = (cufftHandle *)plan.c2r_plan_dev;
		r2c_plan_dev = (cufftHandle *)plan.r2c_plan_dev;
	
		// Launch function
		rc = flcorr2_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, imCount,
					*c2r_plan_dev,
					*r2c_plan_dev);
					
		break;
		
	case FLCC_FAST3_DEV:
	
		c2r_plan_dev = (cufftHandle *)plan.c2r_plan_dev;
		r2c_plan_dev = (cufftHandle *)plan.r2c_plan_dev;
	
		// Launch function
		rc = flcorr3_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, imCount,
					*c2r_plan_dev,
					*r2c_plan_dev);
					
		break;
		
	#endif
		
	case FLCC_DIRECT2_HOST:
		// Launch function
		rc = lcorr2_host<flccPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, 
					imCount, plan.numCpuCores );
					
		break;
		
	case FLCC_DIRECT3_HOST:
		// Launch function
		rc = lcorr3_host<flccPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, 
					imCount, plan.numCpuCores );
					
		break;
		
	#ifndef NO_FFTW	

	case FLCC_FAST2_HOST:
	
		c2r_plan_host_diff = (fftw_plan_t *)plan.c2r_plan_host_diff;
		r2c_plan_host_diff = (fftw_plan_t *)plan.r2c_plan_host_diff;
	
		// Launch function
		rc = flcorr2_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, 
					imCount, plan.numCpuCores,  
					*c2r_plan_host_diff,
					*r2c_plan_host_diff );
					
		break;
		
	case FLCC_FAST3_HOST:
	
		c2r_plan_host_diff = (fftw_plan_t *)plan.c2r_plan_host_diff;
		r2c_plan_host_diff = (fftw_plan_t *)plan.r2c_plan_host_diff;
	
		// Launch function
		rc = flcorr3_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
					image, plan.imageSize, 
					templat, plan.templateSize, 
					lcc, 
					imCount, plan.numCpuCores,  
					*c2r_plan_host_diff,
					*r2c_plan_host_diff );
					
		break;
		
	#endif
		
	default:
		return FLCC_INVALID_PLAN;
	}
	
	return rc;
}


// make instants
#ifndef NO_FFTW
	#ifndef NO_CUDA	
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_exec_t<float,  fftwf_complex, fftwf_plan, cufftReal,       cufftComplex>       (flccPlan plan, float  *image, float *templat,  float *lcc, int imCnt);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_exec_t<double, fftw_complex,  fftw_plan,  cufftDoubleReal, cufftDoubleComplex> (flccPlan plan, double *image, double *templat, double *lcc, int imCnt);
		#endif
	#else
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_exec_t<float,  fftwf_complex, fftwf_plan> (flccPlan plan, float  *image, float *templat,  float *lcc, int imCnt);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_exec_t<double, fftw_complex,  fftw_plan>  (flccPlan plan, double *image, double *templat, double *lcc, int imCnt);
		#endif
	#endif
#else
	#ifndef NO_CUDA	
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_exec_t<float,  cufftReal,       cufftComplex>       (flccPlan plan, float  *image, float *templat,  float *lcc, int imCnt);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_exec_t<double, cufftDoubleReal, cufftDoubleComplex> (flccPlan plan, double *image, double *templat, double *lcc, int imCnt);
		#endif
	#else
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_exec_t<float>  (flccPlan plan, float  *image, float *templat,  float *lcc, int imCnt);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_exec_t<double> (flccPlan plan, double *image, double *templat, double *lcc, int imCnt);
		#endif
	#endif
#endif
