/* 
 * FILE:
 * lcorr_plan_t.c
 * version 1.5
 *
 * DESCRIPTION:
 * This non-API template function selects among the available routines
 * for LCC computation according to the size and dimension specified.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Faculty of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#ifndef NO_CUDA	
#include <FLCC_cuda_defs.h>
#endif
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif


// Template function for the setting of plan parameters to be used in convolution computation
#ifndef NO_FFTW
	#ifndef NO_CUDA	
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t, class cufftRealPrecision, class cufftComplexPrecision>
	#else
	template <class flccPrecision, class fftwComplexPrecision, class fftw_plan_t>
	#endif
#else
	#ifndef NO_CUDA	
	template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
	#else
	template <class flccPrecision>
	#endif
#endif
flccResult lcorr_plan_t (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform)
{
	// Declare host and device variables
	flccPrecision *randomTemplate, *randomImage, *randomLcc;
	
	#ifndef NO_CUDA	
	flccPrecision *randomArrayHost;
	flccPrecision *randomTemplate_dev, *randomImage_dev;
	cufftComplexPrecision *fft_templateDev, *fft_BtDev; 
	int cufft_plan_ok = 0;
	cufftHandle *c2r_plan_dev;
	cufftHandle *r2c_plan_dev;
	MALLOC_SAFE_CALL ( c2r_plan_dev = (cufftHandle *)malloc(sizeof(cufftHandle)) );
	MALLOC_SAFE_CALL ( r2c_plan_dev = (cufftHandle *)malloc(sizeof(cufftHandle)) );
	#endif
	
	#ifndef NO_FFTW
	fftwComplexPrecision *fft_templateHost, *fft_BtHost; 
	fftwComplexPrecision *planFFTW_fft;
	flccPrecision *planFFTW_real;
	int fftw_plan_ok = 0;
	fftw_plan_t *c2r_plan_host_diff;
	fftw_plan_t *r2c_plan_host_diff;
	MALLOC_SAFE_CALL ( c2r_plan_host_diff = (fftw_plan_t *)malloc(sizeof(fftw_plan_t)) );
	MALLOC_SAFE_CALL ( r2c_plan_host_diff = (fftw_plan_t *)malloc(sizeof(fftw_plan_t)) );
	#endif
	
	// Declare timing variables
	ticks start, end;
	double time, best_time = MAX_TIME;
	
	flccResult result;
	
	// Declare size variables
	int SH, SW, SD;
	int T_SIZE, I_SIZE, S_SIZE, s_size, H_SIZE;
	
	// Check validity of arguments
	if (plan==NULL)
		return FLCC_INVALID_PLAN; // An object type flccPlan must have been created beforehand. It is only set in this function
	
	if (dim!=2 && dim!=3)
		return FLCC_INVALID_DIMENSION; // FLCC library only supports 2d and 3d computations as of yet
	
	if (type!=FLCC_SINGLE && type!=FLCC_STREAM)	
		return FLCC_INVALID_TYPE; // A proper type must be given
		
	if (platform!=FLCC_HOST && platform!=FLCC_DEVICE && platform!=FLCC_ANY)	
		return FLCC_INVALID_PLATFORM; // A proper platform must be designated
		
		
	plan->method = FLCC_NULL_METHOD;
	plan->numCpuCores   = NUM_CPUS;		
	plan->numGpuDevices = 1; 			
	
	
	switch (dim)
	{
	case 2: // Fill image and template with random values - 2-dimensions
	
		if (imSize.h<2 || imSize.w<2 || temSize.h<2 || temSize.w<2)		
			return FLCC_INVALID_SIZE; // Each image and template dimension size must have at least two elements
			
		if (temSize.h>imSize.h || temSize.w>imSize.w)
			return FLCC_INVALID_SIZE; // Each dimension of an image must be greater in size than its corresponding template size
		
		// Sizes are computed
		T_SIZE = temSize.h * temSize.w;
		I_SIZE = imSize.h  * imSize.w;
		s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1);
		SH     = alignSize2d (imSize.h+temSize.h-1); // LCC sizes are aligned to the next power of two so as to perform faster ffts.
		SW     = alignSize2d (imSize.w+temSize.w-1);
		S_SIZE = SH * SW;
		H_SIZE = SH * (SW/2 + 1);
		
		// Allocate memory for arrays
		MALLOC_SAFE_CALL ( randomTemplate = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*T_SIZE) ); 
		MALLOC_SAFE_CALL ( randomImage    = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*I_SIZE) ); 
		MALLOC_SAFE_CALL ( randomLcc      = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*s_size) ); 
		
		// Fill arrays with random elements
		randMatrix<flccPrecision> (randomTemplate, T_SIZE);
		randMatrix<flccPrecision> (randomImage,    I_SIZE);
		
		// Measure times on host
		if (platform==FLCC_HOST || platform==FLCC_ANY)
		{
			
			// Start timer
			start = getticks();
			
			// Launch function
			result = lcorr2_host<flccPrecision> (
					randomImage, imSize, 
					randomTemplate, temSize, 
					randomLcc, 
					1, plan->numCpuCores );
			end = getticks(); // End timer
			time = elapsed (end, start);
			
			// If execution was correct and time elapsed less than previous best time, set this method as optimal
			if (result==FLCC_SUCCESS && time<best_time)
			{
				best_time = time;
				plan->method = FLCC_DIRECT2_HOST;
			}
			
			
			#ifndef NO_FFTW
			
			if (fftw_init_threads_t<flccPrecision>() != 0) 
				fftw_plan_with_nthreads_t<flccPrecision> (plan->numCpuCores); // Set FFTW thread number as the number of cpu cores on host
	
			// Allocate memory for FFTW planning arrays
			MALLOC_SAFE_CALL ( planFFTW_fft = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) ); 
			MALLOC_SAFE_CALL ( planFFTW_real = (flccPrecision *)fftw_malloc_t<flccPrecision> (sizeof(flccPrecision)*S_SIZE) );
			
			// Create FFTW plans
			*c2r_plan_host_diff = fftw_plan_dft_c2r_2d_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (SH, SW, planFFTW_fft, planFFTW_real, FFTW_MEASURE); 
			*r2c_plan_host_diff = fftw_plan_dft_r2c_2d_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (SH, SW, planFFTW_real, planFFTW_fft, FFTW_MEASURE);
			
			// Free FFTW planning arrays
			fftw_free_t<flccPrecision> (planFFTW_fft);
			fftw_free_t<flccPrecision> (planFFTW_real);
			
			// Check validity of FFTW plans
			if (*c2r_plan_host_diff!=NULL && *r2c_plan_host_diff!=NULL)
				fftw_plan_ok =1;
			
			if (fftw_plan_ok == 1) // If valid FFTW plans were created
			{
				switch (type)
				{
				case FLCC_SINGLE: // Measure single LCC
				
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr2_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
							randomImage, imSize, 
							randomTemplate, temSize, 
							randomLcc, 
							1, plan->numCpuCores, 
							*c2r_plan_host_diff,
							*r2c_plan_host_diff );
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST2_HOST;
					}
						
					break;
					
				case FLCC_STREAM: // Measure stream LCC
				
					// Allocate memory for arrays to be used in transient part of FLCC and fill with random elements
					MALLOC_SAFE_CALL ( fft_templateHost = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) );
					randMatrix_fftw<flccPrecision, fftwComplexPrecision> (fft_templateHost, H_SIZE);
					
					MALLOC_SAFE_CALL ( fft_BtHost = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) );
					randMatrix_fftw<flccPrecision, fftwComplexPrecision> (fft_BtHost, H_SIZE);
					
					
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr2_testStream_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
							randomImage, imSize, 
							fft_templateHost, fft_BtHost, temSize, 
							randomLcc, plan->numCpuCores,
							*c2r_plan_host_diff,
							*r2c_plan_host_diff );
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST2_HOST;
					}
					
					// Free memory for arrays to be used in transient part of FLCC
					fftw_free_t<flccPrecision> (fft_templateHost);
					fftw_free_t<flccPrecision> (fft_BtHost);
						
					break;
				}
			}
			#endif // NO_FFTW
		}
		
		#ifndef NO_CUDA	
		
		// Measure times on GPU
		if (platform==FLCC_DEVICE || platform==FLCC_ANY)
		{
							
			// Check validity of CUFFT plans
			if ( cufftPlan2d_C2R_t<flccPrecision>(c2r_plan_dev, SH, SW) == CUFFT_SUCCESS &&  cufftPlan2d_R2C_t<flccPrecision>(r2c_plan_dev, SH, SW) == CUFFT_SUCCESS)
				cufft_plan_ok = 1;
			
			switch (type)
			{
			case FLCC_SINGLE: // Measure single LCC
				
				// Start timer
				start = getticks();
				
				// Launch function
				result = lcorr2_dev<flccPrecision> (
						randomImage, imSize, 
						randomTemplate, temSize, 
						randomLcc, 1 );
				end = getticks(); // End timer
				time = elapsed (end, start);
				
				// If execution was correct and time elapsed less than previous best time, set this method as optimal
				if (result==FLCC_SUCCESS && time<best_time)
				{
					best_time = time;
					plan->method = FLCC_DIRECT2_DEV;
				}
			
			
				if (cufft_plan_ok == 1) // If valid CUFFT plan was created
				{
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr2_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
							randomImage, imSize, 
							randomTemplate, temSize, 
							randomLcc, 1, 
							*c2r_plan_dev,
							*r2c_plan_dev);
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST2_DEV;
					}
				}
						
				break;
					
			case FLCC_STREAM: // Measure stream LCC
			
				// Move template to device
				CUDA_SAFE_CALL ( cudaMalloc ((void **)&randomTemplate_dev, T_SIZE*sizeof(flccPrecision)) ); 
				CUDA_SAFE_CALL ( cudaMemcpy (randomTemplate_dev, randomTemplate, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) ); 
				
				// Move image to device
				CUDA_SAFE_CALL ( cudaMalloc ((void **)&randomImage_dev, I_SIZE*sizeof(flccPrecision)) ); 
				CUDA_SAFE_CALL ( cudaMemcpy (randomImage_dev, randomImage, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
			
			
				// Start timer
				start = getticks();
				
				// Launch function
				result = lcorr2_testStream_dev<flccPrecision> (
						randomImage_dev, imSize, 
						randomTemplate_dev, temSize );
				end = getticks(); // End timer
				time = elapsed (end, start);
				
				// If execution was correct and time elapsed less than previous best time, set this method as optimal
				if (result==FLCC_SUCCESS && time<best_time)
				{
					best_time = time;
					plan->method = FLCC_DIRECT2_DEV;
				}
				
				// Free template on device
				CUDA_SAFE_CALL ( cudaFree (randomTemplate_dev) );
				
				
				if (cufft_plan_ok == 1) // If valid CUFFT plan was created
				{
					// Allocate memory for arrays to be used in transient part of FLCC and fill with random elements
					MALLOC_SAFE_CALL ( randomArrayHost = (flccPrecision *)malloc (sizeof(flccPrecision)*2*H_SIZE) ); 
					
					randMatrix<flccPrecision> (randomArrayHost, 2*H_SIZE); 
					CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_templateDev, H_SIZE*sizeof(cufftComplexPrecision)) ); 
					CUDA_SAFE_CALL ( cudaMemcpy (fft_templateDev, randomArrayHost, H_SIZE*sizeof(cufftComplexPrecision), cudaMemcpyHostToDevice) ); 
					
					randMatrix<flccPrecision> (randomArrayHost, 2*H_SIZE); 
					CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_BtDev, H_SIZE*sizeof(cufftComplexPrecision)) );
					CUDA_SAFE_CALL ( cudaMemcpy (fft_BtDev, randomArrayHost, H_SIZE*sizeof(cufftComplexPrecision), cudaMemcpyHostToDevice) ); 
					
					free (randomArrayHost);
					
		
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr2_testStream_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
							randomImage_dev, imSize, 
							fft_templateDev, fft_BtDev, temSize, 
							*c2r_plan_dev,
							*r2c_plan_dev);
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST2_DEV;
					}
					
					// Free memory for arrays to be used in transient part of FLCC
					CUDA_SAFE_CALL ( cudaFree (fft_templateDev) );
					CUDA_SAFE_CALL ( cudaFree (fft_BtDev) );
				}
				
				// Free image on device
				CUDA_SAFE_CALL ( cudaFree (randomImage_dev) );
				
				break;
			}
		}
		
		#endif // NO_CUDA

		break;
		
	case 3: // Fill image and template with random values - 3-dimensions
		
		if (imSize.h<2 || imSize.w<2 || imSize.d<2 || temSize.h<2 || temSize.w<2 || temSize.d<2)
			return FLCC_INVALID_SIZE; // Each image and template dimension size must have at least two elements
			
		if (temSize.h>imSize.h || temSize.w>imSize.w || temSize.d>imSize.d)
			return FLCC_INVALID_SIZE; // Each dimension of an image must be greater in size than its corresponding template size
		
		// Sizes are computed
		T_SIZE = temSize.h * temSize.w * temSize.d;
		I_SIZE = imSize.h  * imSize.w  * imSize.d;
		s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1) * (imSize.d+temSize.d-1);
		SH     = alignSize3d (imSize.h+temSize.h-1); // LCC sizes are aligned to the next power of two so as to perform faster ffts.
		SW     = alignSize3d (imSize.w+temSize.w-1);
		SD     = alignSize3d (imSize.d+temSize.d-1);
		S_SIZE = SH * SW * SD;
		H_SIZE = SH * SW * (SD/2 + 1);
		
		// Allocate memory for arrays
		MALLOC_SAFE_CALL ( randomTemplate = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*T_SIZE) ); 
		MALLOC_SAFE_CALL ( randomImage    = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*I_SIZE) ); 
		MALLOC_SAFE_CALL ( randomLcc      = (flccPrecision *)flcc_malloc (sizeof(flccPrecision)*s_size) ); 
		
		// Fill arrays with random elements
		randMatrix<flccPrecision> (randomTemplate, T_SIZE);
		randMatrix<flccPrecision> (randomImage,    I_SIZE);
		
		// Measure times on host
		if (platform==FLCC_HOST || platform==FLCC_ANY)
		{
							
			// Start timer
			start = getticks();
			
			// Launch function
			result = lcorr3_host<flccPrecision> (
					randomImage, imSize, 
					randomTemplate, temSize, 
					randomLcc, 
					1, plan->numCpuCores );
			end = getticks(); // End timer
			time = elapsed (end, start);
			
			// If execution was correct and time elapsed less than previous best time, set this method as optimal
			if (result==FLCC_SUCCESS && time<best_time)
			{
				best_time = time;
				plan->method = FLCC_DIRECT3_HOST;
			}
			
			
			#ifndef NO_FFTW
			
			if (fftw_init_threads_t<flccPrecision>() != 0) 
				fftw_plan_with_nthreads_t<flccPrecision> (plan->numCpuCores); // Set FFTW thread number as the number of cpu cores on host
		
			// Allocate memory for FFTW planning arrays
			MALLOC_SAFE_CALL ( planFFTW_fft = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) );
			MALLOC_SAFE_CALL ( planFFTW_real = (flccPrecision *)fftw_malloc_t<flccPrecision> (sizeof(flccPrecision)*S_SIZE) );
			
			// Create FFTW plans
			*c2r_plan_host_diff = fftw_plan_dft_c2r_3d_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (SH, SW, SD, planFFTW_fft, planFFTW_real, FFTW_MEASURE); 
			*r2c_plan_host_diff = fftw_plan_dft_r2c_3d_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (SH, SW, SD, planFFTW_real, planFFTW_fft, FFTW_MEASURE);
			
			// Free FFTW planning arrays
			fftw_free_t<flccPrecision> (planFFTW_fft);
			fftw_free_t<flccPrecision> (planFFTW_real);
			
			// Check validity of FFTW plans
			if (*c2r_plan_host_diff!=NULL && *r2c_plan_host_diff!=NULL)
				fftw_plan_ok = 1;
			
			if (fftw_plan_ok == 1) // If valid FFTW plans were created
			{
				switch (type)
				{
				case FLCC_SINGLE: // Measure single LCC
			
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr3_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
							randomImage, imSize, 
							randomTemplate, temSize, 
							randomLcc, 
							1, plan->numCpuCores, 
							*c2r_plan_host_diff,
							*r2c_plan_host_diff );
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST3_HOST;
					}
						
					break;
					
				case FLCC_STREAM: // Measure stream LCC
				
					// Allocate memory for arrays to be used in transient part of FLCC and fill with random elements
					MALLOC_SAFE_CALL ( fft_templateHost = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) );
					randMatrix_fftw<flccPrecision, fftwComplexPrecision> (fft_templateHost, H_SIZE);
					
					MALLOC_SAFE_CALL ( fft_BtHost = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (sizeof(fftwComplexPrecision)*H_SIZE) );
					randMatrix_fftw<flccPrecision, fftwComplexPrecision> (fft_BtHost, H_SIZE);
					
					
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr3_testStream_host<flccPrecision, fftw_plan_t, fftwComplexPrecision> (
							randomImage, imSize, 
							fft_templateHost, fft_BtHost, temSize, 
							randomLcc, plan->numCpuCores,
							*c2r_plan_host_diff,
							*r2c_plan_host_diff );
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST3_HOST;
					}
					
					// Free memory for arrays to be used in transient part of FLCC
					fftw_free_t<flccPrecision> (fft_templateHost);
					fftw_free_t<flccPrecision> (fft_BtHost);
						
					break;
				}
			}
			#endif // NO_FFTW			
		}
		
		#ifndef NO_CUDA	
		
		// Measure times on GPU
		if (platform==FLCC_DEVICE || platform==FLCC_ANY)
		{
			
			// Check validity of CUFFT plans
			if ( cufftPlan3d_C2R_t<flccPrecision>(c2r_plan_dev, SH, SW, SD) == CUFFT_SUCCESS && cufftPlan3d_R2C_t<flccPrecision>(r2c_plan_dev, SH, SW, SD) == CUFFT_SUCCESS )
				cufft_plan_ok = 1;
			
			switch (type)
			{
			case FLCC_SINGLE: // Measure single LCC
			
				// Start timer
				start = getticks();
				
				// Launch function
				result = lcorr3_dev<flccPrecision> (
						randomImage, imSize, 
						randomTemplate, temSize, 
						randomLcc, 1 );
				end = getticks(); // End timer
				time = elapsed (end, start);
				
				// If execution was correct and time elapsed less than previous best time, set this method as optimal
				if (result==FLCC_SUCCESS && time<best_time)
				{
					best_time = time;
					plan->method = FLCC_DIRECT3_DEV;
				}
			
			
				if (cufft_plan_ok == 1) // If valid CUFFT plan was created
				{
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr3_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
							randomImage, imSize, 
							randomTemplate, temSize, 
							randomLcc, 1, 
							*c2r_plan_dev,
							*r2c_plan_dev);
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST3_DEV;
					}
				}
						
				break;
				
			case FLCC_STREAM: // Measure stream LCC
				
				// Move template to device
				CUDA_SAFE_CALL ( cudaMalloc ((void **)&randomTemplate_dev, T_SIZE*sizeof(flccPrecision)) ); 
				CUDA_SAFE_CALL ( cudaMemcpy (randomTemplate_dev, randomTemplate, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) ); 
				
				// Move image to device
				CUDA_SAFE_CALL ( cudaMalloc ((void **)&randomImage_dev, I_SIZE*sizeof(flccPrecision)) ); 
				CUDA_SAFE_CALL ( cudaMemcpy (randomImage_dev, randomImage, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
				
				
				// Start timer
				start = getticks();
				
				// Launch function
				result = lcorr3_testStream_dev<flccPrecision> (
						randomImage_dev, imSize, 
						randomTemplate_dev, temSize );
				end = getticks(); // End timer
				time = elapsed (end, start);
				
				// If execution was correct and time elapsed less than previous best time, set this method as optimal
				if (result==FLCC_SUCCESS && time<best_time)
				{
					best_time = time;
					plan->method = FLCC_DIRECT3_DEV;
				}
				
				// Free template on device
				CUDA_SAFE_CALL ( cudaFree (randomTemplate_dev) );
				
				
				if (cufft_plan_ok == 1) // If valid CUFFT plan was created
				{
					// Allocate memory for arrays to be used in transient part of FLCC and fill with random elements
					MALLOC_SAFE_CALL ( randomArrayHost = (flccPrecision *)malloc (sizeof(flccPrecision)*2*H_SIZE) ); 
					
					randMatrix<flccPrecision> (randomArrayHost, 2*H_SIZE); 
					CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_templateDev, H_SIZE*sizeof(cufftComplexPrecision)) ); 
					CUDA_SAFE_CALL ( cudaMemcpy (fft_templateDev, randomArrayHost, H_SIZE*sizeof(cufftComplexPrecision), cudaMemcpyHostToDevice) ); 
					
					randMatrix<flccPrecision> (randomArrayHost, 2*H_SIZE); 
					CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_BtDev, H_SIZE*sizeof(cufftComplexPrecision)) ); 
					CUDA_SAFE_CALL ( cudaMemcpy (fft_BtDev, randomArrayHost, H_SIZE*sizeof(cufftComplexPrecision), cudaMemcpyHostToDevice) ); 
					
					free (randomArrayHost);
					
					
					// Start timer
					start = getticks();
					
					// Launch function
					result = flcorr3_testStream_dev<flccPrecision, cufftRealPrecision, cufftComplexPrecision> (
							randomImage_dev, imSize, 
							fft_templateDev, fft_BtDev, temSize, 
							*c2r_plan_dev,
							*r2c_plan_dev);
					end = getticks(); // End timer
					time = elapsed (end, start);
					
					// If execution was correct and time elapsed less than previous best time, set this method as optimal
					if (result==FLCC_SUCCESS && time<best_time)
					{
						best_time = time;
						plan->method = FLCC_FAST3_DEV;
					}
					
					// Free memory for arrays to be used in transient part of FLCC
					CUDA_SAFE_CALL ( cudaFree (fft_templateDev) );
					CUDA_SAFE_CALL ( cudaFree (fft_BtDev) );
				}
				
				// Free image on device
				CUDA_SAFE_CALL ( cudaFree (randomImage_dev) );
				
				break;
			}
		}
		
		#endif // NO_CUDA

		break;
	}

	// Free host variables
	flcc_free (randomTemplate);
	flcc_free (randomImage);
	flcc_free (randomLcc);
	
	#ifndef NO_CUDA
	// Destroy CUFFT plans if not needed in execution
	if (plan->method!=FLCC_FAST2_DEV && plan->method!=FLCC_FAST3_DEV)
	{
		if (cufft_plan_ok==1)
		{
			CUFFT_SAFE_CALL ( cufftDestroy (*c2r_plan_dev) );
			CUFFT_SAFE_CALL ( cufftDestroy (*r2c_plan_dev) );
		}
		free (c2r_plan_dev);
		free (r2c_plan_dev);
	}
	else
	{
		plan->c2r_plan_dev = (void *)c2r_plan_dev;
		plan->r2c_plan_dev = (void *)r2c_plan_dev;
	}
	#endif
		
	#ifndef NO_FFTW
	// Destroy FFTW plans if not needed in execution
	if (plan->method!=FLCC_FAST2_HOST && plan->method!=FLCC_FAST3_HOST)
	{
		if (fftw_plan_ok==1)
		{
			fftw_destroy_plan_t<fftw_plan_t> (*c2r_plan_host_diff);
			fftw_destroy_plan_t<fftw_plan_t> (*r2c_plan_host_diff);
		}
		free (c2r_plan_host_diff);
		free (r2c_plan_host_diff);
	}
	else
	{
		plan->c2r_plan_host_diff = (void *)c2r_plan_host_diff;
		plan->r2c_plan_host_diff = (void *)r2c_plan_host_diff;
	}
	#endif
	
	// Loose ends
	if (plan->method == FLCC_NULL_METHOD)
		return FLCC_EXEC_FAILED;
	
	// Set rest of flccPlan parameters
	plan->dimension    = dim;
	plan->imageSize    = imSize;
	plan->templateSize = temSize;
	plan->type         = type;
	plan->platform	   = platform;
		

	return FLCC_SUCCESS;
}


// make instants
#ifndef NO_FFTW
	#ifndef NO_CUDA	
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_plan_t<float,  fftwf_complex, fftwf_plan, cufftReal,       cufftComplex>       (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_plan_t<double, fftw_complex,  fftw_plan,  cufftDoubleReal, cufftDoubleComplex> (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
	#else
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_plan_t<float,  fftwf_complex, fftwf_plan> (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_plan_t<double, fftw_complex,  fftw_plan>  (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
	#endif
#else
	#ifndef NO_CUDA	
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_plan_t<float,  cufftReal,       cufftComplex>       (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_plan_t<double, cufftDoubleReal, cufftDoubleComplex> (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
	#else
		#ifndef NO_SINGLE_PREC
		template flccResult lcorr_plan_t<float>  (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
		#ifndef NO_DOUBLE_PREC
		template flccResult lcorr_plan_t<double> (flccPlan *plan, int dim, flccSize imSize, flccSize temSize, flccType type, flccPlatform platform);
		#endif
	#endif
#endif
