/* 
 * FILE:
 * FLCC_host_functions.c
 * version 1.0
 *
 * DESCRIPTION:
 * These are the helper host functions for the FLCC library.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2009
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif


// Fill a float array with random values
template <class flccPrecision>
void randMatrix (flccPrecision *array, int size)
{
	srand ( time(NULL) );
	for (int i=0; i<size; i++) {
		array[i] = (flccPrecision)rand()/RAND_MAX;
	}
} 
#ifndef NO_SINGLE_PREC
template void randMatrix<float>  (float  *array, int size);
#endif
#ifndef NO_DOUBLE_PREC
template void randMatrix<double> (double *array, int size);
#endif


#ifndef NO_FFTW
// Fill an fftwf_complex array with random values
template <class flccPrecision, class fftwComplexPrecision>
void randMatrix_fftw (fftwComplexPrecision *array, int size)
{
	srand ( time(NULL) );
	for (int i=0; i<size; i++) {
		array[i][0] = (flccPrecision)rand()/RAND_MAX;
		array[i][1] = (flccPrecision)rand()/RAND_MAX;
	}
} 
#ifndef NO_SINGLE_PREC
template void randMatrix_fftw<float,  fftwf_complex> (fftwf_complex *array, int size);
#endif
#ifndef NO_DOUBLE_PREC
template void randMatrix_fftw<double, fftw_complex>  (fftw_complex  *array, int size);
#endif
#endif // NO_FFTW


// Compute the mean value of an array
template <class flccPrecision>
flccPrecision findMean (flccPrecision *array, int size)
{
	flccPrecision sum = 0.0;
	for (int i=0; i<size; i++) {
		sum += array[i];
	}
	return sum/size;
} 
#ifndef NO_SINGLE_PREC
template float  findMean<float>  (float  *array, int size);
#endif
#ifndef NO_DOUBLE_PREC
template double findMean<double> (double *array, int size);
#endif


// Compute the standard deviation of an array (but no division with the number of elements)
template <class flccPrecision>
flccPrecision findStd (flccPrecision mean, flccPrecision *array, int size)
{
	flccPrecision sum = 0.0;
	for (int i=0; i<size; i++) {
		sum += array[i]*array[i];
	}
	return sqrt(sum - size*mean*mean);
} 
#ifndef NO_SINGLE_PREC
template float  findStd<float>  (float  mean, float  *array, int size);
#endif
#ifndef NO_DOUBLE_PREC
template double findStd<double> (double mean, double *array, int size);
#endif


// Normalize an array to zero mean and unit standard deviation
template <class flccPrecision>
void normalize (flccPrecision mean, flccPrecision std, flccPrecision *array, int size)
{
	for (int i=0; i<size; i++) {
		array[i] = (array[i] - mean) / std;
	}	
} 
#ifndef NO_SINGLE_PREC
template void normalize<float>  (float  mean, float  std, float  *array, int size);
#endif
#ifndef NO_DOUBLE_PREC
template void normalize<double> (double mean, double std, double *array, int size);
#endif


// Align size of 2d fft transform in order to compute using faster algorithms
int alignSize2d (int size)
{        
	// Align size to a multiple of block dimension
	size = (size%THREADS_2D != 0) ? (size - size%THREADS_2D + THREADS_2D) : size; 
		
	unsigned int pot = 2;
	
	// If the size is small enough, align to the nearest higher power of two
	while (pot <= 1024) {
		if (size <= pot)
			return pot;
		else 
			pot = pot << 1;
	}
	// If not, align only to the nearest higher multiple of 512
	return (size%512 != 0) ? (size - size%512 + 512) : size; 
} 


// Align size of 3d fft transform in order to compute using faster algorithms
int alignSize3d (int size)
{        
	// Align size to a multiple of block dimension
	size = (size%THREADS_3D != 0) ? (size - size%THREADS_3D + THREADS_3D) : size; 
		
	unsigned int pot = 2;
	
	// If the size is small enough, align to the nearest higher power of two
	while (pot <= 128) {
		if (size <= pot)
			return pot;
		else 
			pot = pot << 1;
	}
	// If not, align only to the nearest higher multiple of 64
	return (size%64 != 0) ? (size - size%64 + 64) : size; 
} 

