/* 
 * FILE:
 * FLCC_kernels_2d.cu
 * version 1.1
 *
 * DESCRIPTION:
 * These are the 2D GPU kernel functions for the FLCC library.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * March 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>

 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for conv2_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Optimized kernels
#include "opt/convolution2d_kernel.c"


// Perform the convolution direct method between 2D image and template on a GPU without using its shared memory
template <class flccPrecision>
__global__ void convolution2d_kernel_generic (
		flccPrecision *imageDev, 
		flccPrecision *templateDev,
		flccPrecision *outputDev,
		int TH, int TW,
		int SW, int BW )
{
	const int u = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int v = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;

	register flccPrecision conv = 0.0;
	const int i_idx = u + IMUL(v, BW);	
	const int t_idx = IMUL(TH, TW) - 1;	
	for (int j=0; j<TH; j++) {
		for (int i=0; i<TW; i++) {
			conv += imageDev[i_idx + i + IMUL(j, BW)] * 
				templateDev[t_idx - i - IMUL(j, TW)]; } }
	outputDev[u + IMUL(v, SW)] = conv;
} 

#ifndef NO_SINGLE_PREC
template __global__ void convolution2d_kernel_generic<float> (
		float *imageDev, 
		float *templateDev,
		float *outputDev,
		int TH, int TW,
		int SW, int BW );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void convolution2d_kernel_generic<double> (
		double *imageDev, 
		double *templateDev,
		double *outputDev,
		int TH, int TW,
		int SW, int BW );
#endif


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for lcorr2_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Optimized kernels
#include "opt/lcorrelation2d_kernel.c"
	
	
// Perform the LCC direct method between 2D image and template on a GPU without using its shared memory
template <class flccPrecision>
__global__ void lcorrelation2d_kernel_generic (
		flccPrecision *imageDev, 
		flccPrecision *templateDev,
		flccPrecision *outputDev,
		int TH, int TW,
		int SW, int BW )
{
	const int u = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int v = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;

	register flccPrecision corr   = 0.0;
	register flccPrecision sum    = 0.0;
	register flccPrecision sum_sq = 0.0;
	const int i_idx = u + IMUL(v, BW);
	for (int j=0; j<TH; j++) {
		for (int i=0; i<TW; i++) {
			const flccPrecision im = imageDev[i_idx + i + IMUL(j, BW)];
			corr   += im * templateDev[i + IMUL(j, TW)];
			sum    += im;
			sum_sq += im*im; } } 
	const flccPrecision var = sum_sq - sum*sum/IMUL(TH,TW);
	const int out_idx = u + IMUL(v, SW);
	if (var == 0)	outputDev[out_idx] = 0.0;
	else			outputDev[out_idx] = corr / sqrtf(var);
}

#ifndef NO_SINGLE_PREC
template __global__ void lcorrelation2d_kernel_generic<float> (
		float *imageDev, 
		float *templateDev,
		float *outputDev,
		int TH, int TW,
		int SW, int BW );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void lcorrelation2d_kernel_generic<double> (
		double *imageDev, 
		double *templateDev,
		double *outputDev,
		int TH, int TW,
		int SW, int BW );
#endif
		

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for fconv2_dev & flcorr2_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Normalize template and create spatial support arrays
template <class flccPrecision, class cufftRealPrecision>
__global__ void makePermanent2d (
		cufftRealPrecision *templateDev, 
		flccPrecision *tempTemplateDev, 
		cufftRealPrecision *BtDev, 
		flccPrecision mean, flccPrecision std,
		int TH, int TW,
		int SW )
{
    const int i = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int j = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;
	
	const int s_indx = i + IMUL(j, SW);
	
	if (i<TW && j<TH) {
		templateDev[s_indx] = ( tempTemplateDev[IMUL(TH,TW)-1 - i-IMUL(j,TW)] - mean)/ std;
		BtDev[s_indx]       = 1.0; 
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makePermanent2d<float, cufftReal> (
		cufftReal *templateDev, 
		float *tempTemplateDev, 
		cufftReal *BtDev, 
		float mean, float std,
		int TH, int TW,
		int SW );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makePermanent2d<double, cufftDoubleReal> (
		cufftDoubleReal *templateDev, 
		double *tempTemplateDev, 
		cufftDoubleReal *BtDev, 
		double mean, double std,
		int TH, int TW,
		int SW );
#endif


// Pad frame and squared frame value array
template <class flccPrecision, class cufftRealPrecision>
__global__ void makeTransient2d (
		cufftRealPrecision *sceneDev, 
		cufftRealPrecision *sceneSqDev, 
		flccPrecision *tempSceneDev,
		int IH, int IW,
		int SW )
{
    const int i = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int j = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;
	
	const int s_indx = i + IMUL(j, SW);
	
    if (i<IW && j<IH) {
		const flccPrecision im = tempSceneDev[i + IMUL(j, IW)];
		sceneDev[s_indx]   = im;
		sceneSqDev[s_indx] = im*im;
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makeTransient2d<float, cufftReal> (
		cufftReal *sceneDev, 
		cufftReal *sceneSqDev, 
		float *tempSceneDev,
		int IH, int IW,
		int SW );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makeTransient2d<double, cufftDoubleReal> (
		cufftDoubleReal *sceneDev, 
		cufftDoubleReal *sceneSqDev, 
		double *tempSceneDev,
		int IH, int IW,
		int SW );
#endif
		

// Perform a point-to-point multiplication
template <class cufftComplexPrecision>
__global__ void complexProduct2d (
		cufftComplexPrecision *prod_fft, 
		cufftComplexPrecision *fft1, 
		cufftComplexPrecision *fft2,
		int SW, int S_SIZE )
{
    const int i = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int j = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;

	const int indx = i + IMUL(j, SW);

	if (i<SW) {
		const cufftComplexPrecision f1 = fft1[indx];
		const cufftComplexPrecision f2 = fft2[indx];
		cufftComplexPrecision f;
	
		// Divide by number of elements in data set to get back original data
		f.x = (f1.x*f2.x - f1.y*f2.y) / S_SIZE;
		f.y = (f1.x*f2.y + f1.y*f2.x) / S_SIZE;
		
		prod_fft[indx] = f;
	}
} 

#ifndef NO_SINGLE_PREC
template __global__ void complexProduct2d<cufftComplex> (
		cufftComplex *prod_fft, 
		cufftComplex *fft1, 
		cufftComplex *fft2,
		int SW, int S_SIZE );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void complexProduct2d<cufftDoubleComplex> (
		cufftDoubleComplex *prod_fft, 
		cufftDoubleComplex *fft1, 
		cufftDoubleComplex *fft2,
		int SW, int S_SIZE );
#endif
		

// Compute LCC array using all precalculated arrays
template <class flccPrecision, class cufftRealPrecision>
__global__ void makeLccArray2d (
		flccPrecision *lccDev, 
		cufftRealPrecision *SceneTimesBtDev,
		cufftRealPrecision *SceneSqTimesBtDev,
		cufftRealPrecision *SceneTimesTDev, 
		int sh, int sw,
		int SW, int T_SIZE )
{
    const int i = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
	const int j = IMUL(blockIdx.y, blockDim.y) + threadIdx.y;

	const int S_indx = i + IMUL(j, SW);
	const int s_indx = i + IMUL(j, sw);
	
	if (i<sw && j<sh) {
		const flccPrecision SBt = SceneTimesBtDev[S_indx];
		const flccPrecision nom   =    SceneTimesTDev[S_indx];
		const flccPrecision denom = SceneSqTimesBtDev[S_indx] - SBt*SBt/T_SIZE;

		if (denom == 0)	lccDev[s_indx] = 0.0;
		else			lccDev[s_indx] = nom/sqrtf(denom);
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makeLccArray2d<float, cufftReal> (
		float *lccDev, 
		cufftReal *SceneTimesBtDev,
		cufftReal *SceneSqTimesBtDev,
		cufftReal *SceneTimesTDev, 
		int sh, int sw,
		int SW, int T_SIZE );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makeLccArray2d<double, cufftDoubleReal> (
		double *lccDev, 
		cufftDoubleReal *SceneTimesBtDev,
		cufftDoubleReal *SceneSqTimesBtDev,
		cufftDoubleReal *SceneTimesTDev, 
		int sh, int sw,
		int SW, int T_SIZE );
#endif
