/* 
 * FILE:
 * FLCC_kernels_3d.cu
 * version 1.1
 *
 * DESCRIPTION:
 * These are the 3D GPU kernel functions for the FLCC library.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * March 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for conv3_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Optimized convolution kernels
#include "opt/convolution3d_kernel.c"


// Perform the convolution direct method between 3D image and template on a GPU without using its shared memory
template <class flccPrecision>
__global__ void convolution3d_kernel_generic (
		flccPrecision *imageDev, 
		flccPrecision *templateDev,
		flccPrecision *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD )
{
    const int block_num_x = SD/blockDim.x;
    const int u = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int v = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int w = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;
	
	register flccPrecision conv = 0.0;
	const int i_idx = u + IMUL(v, BD) + IMUL(w, IMUL(BD, BW));										
	const int t_idx = IMUL(TH, IMUL(TD, TW)) - 1;										
	for (int k=0; k<TH; k++) {
		for (int j=0; j<TW; j++) {
			for (int i=0; i<TD; i++) {
				conv += imageDev[i_idx + i + IMUL(j, BD) + IMUL(k, IMUL(BD, BW))] 
					* templateDev[t_idx - i - IMUL(j, TD) - IMUL(k, IMUL(TD, TW))];
			} } }
	outputDev[u + IMUL(v, SD) + IMUL(w, IMUL(SD, SW))] = conv;
} 

#ifndef NO_SINGLE_PREC
template __global__ void convolution3d_kernel_generic<float> (
		float *imageDev, 
		float *templateDev,
		float *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void convolution3d_kernel_generic<double> (
		double *imageDev, 
		double *templateDev,
		double *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD );
#endif
		

// Pad a 3D array on both sides
template <class flccPrecision>
__global__ void padSignal3d_both (
		flccPrecision *padded, 
		flccPrecision *source,
		int TH, int TW, int TD, 
		int IH, int IW, int ID,
		int BH, int BW, int BD ) 
{
	const int block_num_x = (BD+blockDim.x-1)/blockDim.x;       
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	const int indx = i + IMUL(j, BD) + IMUL(k, IMUL(BD, BW));

	if (i>TD-2 && i<ID+TD-1 && j>TW-2 && j<IW+TW-1 && k>TH-2 && k<IH+TH-1) {
		padded[indx] = source[(i-TD+1) + IMUL((j-TW+1), ID) + IMUL((k-TH+1), IMUL(ID, IW))];
	}
	else if (i<BD && j<BW && k<BH) {
		padded[indx] = 0.0;
	}
} 

#ifndef NO_SINGLE_PREC
template __global__ void padSignal3d_both<float> (
		float *padded, 
		float *source,
		int TH, int TW, int TD, 
		int IH, int IW, int ID,
		int BH, int BW, int BD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void padSignal3d_both<double> (
		double *padded, 
		double *source,
		int TH, int TW, int TD, 
		int IH, int IW, int ID,
		int BH, int BW, int BD );
#endif
		

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for fconv3_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Pad a 3D signal
template <class flccPrecision, class cufftRealPrecision>
__global__ void padSignal3d (
		cufftRealPrecision *signalDev, 
		flccPrecision *tempSignalDev, 
		int IH, int IW, int ID,
		int SW, int SD )
{
	const int block_num_x = SD/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;
	
	const int s_indx = i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW));
	
	if (i<ID && j<IW && k<IH) {
		signalDev[s_indx] = tempSignalDev[i + IMUL(j, ID) + IMUL(k, IMUL(ID, IW))];
	} 
	else {
		signalDev[s_indx] = 0.0; 
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void padSignal3d<float, cufftReal> (
		cufftReal *signalDev, 
		float *tempSignalDev, 
		int IH, int IW, int ID,
		int SW, int SD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void padSignal3d<double, cufftDoubleReal> (
		cufftDoubleReal *signalDev, 
		double *tempSignalDev, 
		int IH, int IW, int ID,
		int SW, int SD );
#endif
		

// Depad a 3D signal
template <class flccPrecision>
__global__ void depadSignal3d (
		flccPrecision *depadded, 
		flccPrecision *padded, 
		int sh, int sw, int sd,
		int SW, int SD )
{
	const int block_num_x = SD/blockDim.x;       
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	if (i<sd && j<sw && k<sh) {
		depadded[i + IMUL(j, sd) + IMUL(k, IMUL(sd, sw))] = padded[i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW))];
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void depadSignal3d<float> (
		float *depadded, 
		float *padded, 
		int sh, int sw, int sd,
		int SW, int SD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void depadSignal3d<double> (
		double *depadded, 
		double *padded, 
		int sh, int sw, int sd,
		int SW, int SD );
#endif


// Compute convolution array using all precalculated arrays
template <class flccPrecision, class cufftRealPrecision>
__global__ void makeConvArray3d (
		flccPrecision *convDev, 
		cufftRealPrecision *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD)
{
	const int block_num_x = (sd+blockDim.x-1)/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	if (i<sd && j<sw && k<sh) {
		convDev[i + IMUL(j, sd) + IMUL(k, IMUL(sd, sw))] = SceneTimesTDev[i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW))];
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makeConvArray3d<float, cufftReal> (
		float *convDev, 
		cufftReal *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD);
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makeConvArray3d<double, cufftDoubleReal> (
		double *convDev, 
		cufftDoubleReal *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD);
#endif
		

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for lcorr3_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Optimized kernels
#include "opt/lcorrelation3d_kernel.c"

// Perform the LCC direct method between 3D image and template on a GPU without using its shared memory
template <class flccPrecision>
__global__ void lcorrelation3d_kernel_generic (
		flccPrecision *imageDev, 
		flccPrecision *templateDev,
		flccPrecision *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD )
{
    const int block_num_x = SD/blockDim.x;       
    const int u = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int v = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int w = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	register flccPrecision corr   = 0.0;
	register flccPrecision sum    = 0.0;
	register flccPrecision sum_sq = 0.0;
	const int i_idx = u + IMUL(v, BD) + IMUL(w, IMUL(BD, BW));										
	for (int k=0; k<TH; k++) {
		for (int j=0; j<TW; j++) {
			for (int i=0; i<TD; i++) {
				const flccPrecision im = imageDev[i_idx + i + IMUL(j, BD) + IMUL(k, IMUL(BD, BW))];
				corr   += im * templateDev[i + IMUL(j, TD) + IMUL(k, IMUL(TD, TW))];
				sum    += im;
				sum_sq += im*im;
			} } }
	const flccPrecision var = sum_sq - sum*sum/IMUL(TH,IMUL(TW,TD));
	const int out_idx = u + IMUL(v, SD) + IMUL(w, IMUL(SD, SW));
	if (var == 0)	outputDev[out_idx] = 0.0;
	else			outputDev[out_idx] = corr / sqrtf(var);
}

#ifndef NO_SINGLE_PREC
template __global__ void lcorrelation3d_kernel_generic<float> (
		float *imageDev, 
		float *templateDev,
		float *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void lcorrelation3d_kernel_generic<double> (
		double *imageDev, 
		double *templateDev,
		double *outputDev,
		int TH, int TW, int TD,
		int SW, int SD,
		int BW, int BD );
#endif
		

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Kernels for flcorr3_dev
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Normalize template and create spatial support arrays
template <class flccPrecision, class cufftRealPrecision>
__global__ void makePermanent3d (
		cufftRealPrecision *templateDev, 
		flccPrecision *tempTemplateDev, 
		cufftRealPrecision *BtDev, 
		flccPrecision mean, flccPrecision std,
		int TH, int TW, int TD,
		int SW, int SD )
{
	const int block_num_x = SD/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;
	
	const int s_indx = i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW));
	
	if (i<TD && j<TW && k<TH) {
		templateDev[s_indx] = ( tempTemplateDev[IMUL(TH,IMUL(TW,TD))-1 - (i+IMUL(j,TD)+IMUL(k,IMUL(TD,TW)))] - mean)/ std;
		BtDev[s_indx]       = 1.0; 
	}
	else {
		templateDev[s_indx] = 0.0; 
		BtDev[s_indx]       = 0.0; 
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makePermanent3d <float, cufftReal> (
		cufftReal *templateDev, 
		float *tempTemplateDev, 
		cufftReal *BtDev, 
		float mean, float std,
		int TH, int TW, int TD,
		int SW, int SD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makePermanent3d <double, cufftDoubleReal> (
		cufftDoubleReal *templateDev, 
		double *tempTemplateDev, 
		cufftDoubleReal *BtDev, 
		double mean, double std,
		int TH, int TW, int TD,
		int SW, int SD );
#endif


// Pad frame and square frame value array
template <class flccPrecision, class cufftRealPrecision>
__global__ void makeTransient3d (
		cufftRealPrecision *sceneDev, 
		cufftRealPrecision *sceneSqDev, 
		flccPrecision *tempSceneDev,
		int IH, int IW, int ID,
		int SW, int SD )
{
	const int block_num_x = SD/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;
	
	const int s_indx = i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW));
	
    if (i<ID && j<IW && k<IH) {
		const flccPrecision im = tempSceneDev[i + IMUL(j, ID) + IMUL(k, IMUL(ID, IW))];
		sceneDev[s_indx]   = im;
		sceneSqDev[s_indx] = im*im;
	} 
	else {
		sceneDev[s_indx]   = 0.0;
		sceneSqDev[s_indx] = 0.0;
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makeTransient3d<float, cufftReal> (
		cufftReal *sceneDev, 
		cufftReal *sceneSqDev, 
		float *tempSceneDev,
		int IH, int IW, int ID,
		int SW, int SD );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makeTransient3d<double, cufftDoubleReal> (
		cufftDoubleReal *sceneDev, 
		cufftDoubleReal *sceneSqDev, 
		double *tempSceneDev,
		int IH, int IW, int ID,
		int SW, int SD );
#endif
		

// Perform a point-to-point multiplication
template <class cufftComplexPrecision>
__global__ void complexProduct3d (
		cufftComplexPrecision *prod_fft, 
		cufftComplexPrecision *fft1, 
		cufftComplexPrecision *fft2,
		int SW, int SD,
		int S_SIZE)
{
	const int block_num_x = (SD+blockDim.x-1)/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	const int indx = i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW));
	
	if (i<SD) {
		const cufftComplexPrecision f1 = fft1[indx];
		const cufftComplexPrecision f2 = fft2[indx];
		cufftComplexPrecision f;
	
		// Divide by number of elements in data set to get back original data
		f.x = (f1.x*f2.x - f1.y*f2.y) / S_SIZE;
		f.y = (f1.x*f2.y + f1.y*f2.x) / S_SIZE;
		
		prod_fft[indx] = f;
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void complexProduct3d<cufftComplex> (
		cufftComplex *prod_fft, 
		cufftComplex *fft1, 
		cufftComplex *fft2,
		int SW, int SD,
		int S_SIZE);
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void complexProduct3d<cufftDoubleComplex> (
		cufftDoubleComplex *prod_fft, 
		cufftDoubleComplex *fft1, 
		cufftDoubleComplex *fft2,
		int SW, int SD,
		int S_SIZE);
#endif
		

// Compute LCC array using all precalculated arrays
template <class flccPrecision, class cufftRealPrecision>
__global__ void makeLccArray3d (
		flccPrecision *lccDev, 
		cufftRealPrecision *SceneTimesBtDev,
		cufftRealPrecision *SceneSqTimesBtDev,
		cufftRealPrecision *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD, int T_SIZE )
{
	const int block_num_x = (sd+blockDim.x-1)/blockDim.x;
    const int i = IMUL(blockDim.x, blockIdx.x%block_num_x) + threadIdx.x;
    const int j = IMUL(blockDim.y, blockIdx.x/block_num_x) + threadIdx.y;
    const int k = IMUL(blockDim.z, blockIdx.y)             + threadIdx.z;

	const int S_indx = i + IMUL(j, SD) + IMUL(k, IMUL(SD, SW));
	const int s_indx = i + IMUL(j, sd) + IMUL(k, IMUL(sd, sw));
	
	if (i<sd && j<sw && k<sh) {
		const flccPrecision SBt = SceneTimesBtDev[S_indx];
		const flccPrecision nom   =    SceneTimesTDev[S_indx];
		const flccPrecision denom = SceneSqTimesBtDev[S_indx] - SBt*SBt/T_SIZE;

		if (denom == 0)	lccDev[s_indx] = 0.0;
		else			lccDev[s_indx] = nom/sqrtf(denom);
	} 
} 

#ifndef NO_SINGLE_PREC
template __global__ void makeLccArray3d<float, cufftReal> (
		float *lccDev, 
		cufftReal *SceneTimesBtDev,
		cufftReal *SceneSqTimesBtDev,
		cufftReal *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD, int T_SIZE );
#endif
#ifndef NO_DOUBLE_PREC
template __global__ void makeLccArray3d<double, cufftDoubleReal> (
		double *lccDev, 
		cufftDoubleReal *SceneTimesBtDev,
		cufftDoubleReal *SceneSqTimesBtDev,
		cufftDoubleReal *SceneTimesTDev, 
		int sh, int sw, int sd,	
		int SW, int SD, int T_SIZE );
#endif
