/* 
 * FILE:
 * FLCC_thread_functions_3d.c
 * version 1.0
 *
 * DESCRIPTION:
 * These are the 3D CPU thread functions for the FLCC library.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2009
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Thread functions for conv3_host
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Optimized threads
GEN_DIRECT3_THREAD_CONV (  2,  2,  2 )
GEN_DIRECT3_THREAD_CONV (  3,  3,  3 )
GEN_DIRECT3_THREAD_CONV (  4,  4,  4 )
GEN_DIRECT3_THREAD_CONV (  5,  5,  5 )
GEN_DIRECT3_THREAD_CONV (  6,  6,  6 )
GEN_DIRECT3_THREAD_CONV (  7,  7,  7 )
GEN_DIRECT3_THREAD_CONV (  8,  8,  8 )
GEN_DIRECT3_THREAD_CONV (  9,  9,  9 )
GEN_DIRECT3_THREAD_CONV ( 10, 10, 10 )

#ifndef NO_SINGLE_PREC
GEN_DIRECT3_THREAD_CONV_INSTANCE (  2,  2,  2, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  3,  3,  3, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  4,  4,  4, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  5,  5,  5, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  6,  6,  6, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  7,  7,  7, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  8,  8,  8, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  9,  9,  9, float )
GEN_DIRECT3_THREAD_CONV_INSTANCE ( 10, 10, 10, float )
#endif

#ifndef NO_DOUBLE_PREC
GEN_DIRECT3_THREAD_CONV_INSTANCE (  2,  2,  2, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  3,  3,  3, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  4,  4,  4, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  5,  5,  5, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  6,  6,  6, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  7,  7,  7, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  8,  8,  8, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE (  9,  9,  9, double )
GEN_DIRECT3_THREAD_CONV_INSTANCE ( 10, 10, 10, double )
#endif


// Perform the convolution direct method between 3D image and template on CPU
template <class flccPrecision>
void *convolution3d_thread_generic (void *arg)
{
	const corrParm<flccPrecision> *c = (corrParm<flccPrecision> *)arg;
	
	const int rem = c->td%4;
	const int TD  = c->td-rem;
	
	for (int n=0; n<c->offset; n++) {
		const int indx = c->initialPosition + n;
		const int u = (indx%(c->sd*c->sw))%c->sd;
		const int v = (indx%(c->sd*c->sw))/c->sd;
		const int w = indx/(c->sd*c->sw);
		
		register flccPrecision conv = 0.0;
		
		int t_indx = c->th*c->tw*c->td-1;
		
		for (int k=0; k<c->th; k++) {
			const int kk = u + v*c->bd + (w+k)*c->bw*c->bd;
			for (int j=0; j<c->tw; j++) {
				int jj = kk + j*c->bd;
				for (int i=0; i<TD; i+=4) {
					conv += c->image[jj  ] * c->templat[t_indx  ];
					conv += c->image[jj+1] * c->templat[t_indx-1];
					conv += c->image[jj+2] * c->templat[t_indx-2];
					conv += c->image[jj+3] * c->templat[t_indx-3];
					
					jj += 4;
					t_indx -= 4;
				}
				
				switch (rem) {
					case 3: conv += c->image[jj+2] * c->templat[t_indx-2];
					case 2: conv += c->image[jj+1] * c->templat[t_indx-1];
					case 1: conv += c->image[jj  ] * c->templat[t_indx  ];
				}
			
				t_indx -= rem;
			}
		}

		c->lcc[indx] = conv;
	} 
	
	pthread_exit (NULL);	
} 

#ifndef NO_SINGLE_PREC
template void *convolution3d_thread_generic<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *convolution3d_thread_generic<double> (void *arg);
#endif


// Pad a 3D signal on both sides on CPU
template <class flccPrecision>
void *padSignal3d_both (void *arg)
{
	const padParm<flccPrecision> *p = (padParm<flccPrecision> *)arg;
	
	for (int n=0; n<p->offset; n++) {
		const int indx = p->initialPosition + n;
		const int x = (indx%(p->bd*p->bw))%p->bd;
		const int y = (indx%(p->bd*p->bw))/p->bd;
		const int z = indx/(p->bd*p->bw);
		
		if ( x>(p->td - 2) && x<(p->sd) && y>(p->tw - 2) && y<(p->sw) && z>(p->th - 2) && z<(p->sh) ) {
			p->padded[indx] = p->unpadded[(x - p->td + 1) + (y - p->tw + 1)*(p->id) + (z - p->th + 1)*(p->iw*p->id)];
		}
		else {
			p->padded[indx] = 0.0;
		}
	} 
	
	pthread_exit (NULL);	
} 

#ifndef NO_SINGLE_PREC
template void *padSignal3d_both<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *padSignal3d_both<double> (void *arg);
#endif


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Thread functions for lcorr3_host
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Optimized threads
GEN_DIRECT3_THREAD_LCORR (  2,  2,  2 )
GEN_DIRECT3_THREAD_LCORR (  3,  3,  3 )
GEN_DIRECT3_THREAD_LCORR (  4,  4,  4 )
GEN_DIRECT3_THREAD_LCORR (  5,  5,  5 )
GEN_DIRECT3_THREAD_LCORR (  6,  6,  6 )
GEN_DIRECT3_THREAD_LCORR (  7,  7,  7 )
GEN_DIRECT3_THREAD_LCORR (  8,  8,  8 )
GEN_DIRECT3_THREAD_LCORR (  9,  9,  9 )
GEN_DIRECT3_THREAD_LCORR ( 10, 10, 10 )

#ifndef NO_SINGLE_PREC
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  2,  2,  2, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  3,  3,  3, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  4,  4,  4, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  5,  5,  5, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  6,  6,  6, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  7,  7,  7, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  8,  8,  8, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  9,  9,  9, float )
GEN_DIRECT3_THREAD_LCORR_INSTANCE ( 10, 10, 10, float )
#endif

#ifndef NO_DOUBLE_PREC
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  2,  2,  2, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  3,  3,  3, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  4,  4,  4, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  5,  5,  5, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  6,  6,  6, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  7,  7,  7, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  8,  8,  8, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE (  9,  9,  9, double )
GEN_DIRECT3_THREAD_LCORR_INSTANCE ( 10, 10, 10, double )
#endif


// Performs the LCC direct method between 3D image and template on CPU
template <class flccPrecision>
void *lcorrelation3d_thread_generic (void *arg)
{
	const corrParm<flccPrecision> *c = (corrParm<flccPrecision> *)arg;
	
	const int rem = c->td%4;
	const int TD  = c->td-rem;
	
	for (int n=0; n<c->offset; n++) {
		const int indx = c->initialPosition + n;
		const int u = (indx%(c->sd*c->sw))%c->sd;
		const int v = (indx%(c->sd*c->sw))/c->sd;
		const int w = indx/(c->sd*c->sw);
		
		register flccPrecision corr   = 0.0;
		register flccPrecision sum    = 0.0;
		register flccPrecision sum_sq = 0.0;
		
		register flccPrecision im;
		
		int t_indx = 0;
		
		for (int k=0; k<c->th; k++) {
			const int kk = u + v*c->bd + (w+k)*c->bw*c->bd;
			for (int j=0; j<c->tw; j++) {
				int jj = kk + j*c->bd;
				for (int i=0; i<TD; i+=4) {
					im = c->image[jj  ];  corr += im * c->templat[t_indx  ];  sum += im;  sum_sq += im * im;
					im = c->image[jj+1];  corr += im * c->templat[t_indx+1];  sum += im;  sum_sq += im * im;
					im = c->image[jj+2];  corr += im * c->templat[t_indx+2];  sum += im;  sum_sq += im * im;
					im = c->image[jj+3];  corr += im * c->templat[t_indx+3];  sum += im;  sum_sq += im * im;
					
					jj += 4;
					t_indx += 4;
				} 
				
				switch (rem) {
					case 3: im = c->image[jj+2];  corr += im * c->templat[t_indx+2];  sum += im;  sum_sq += im * im;
					case 2: im = c->image[jj+1];  corr += im * c->templat[t_indx+1];  sum += im;  sum_sq += im * im;
					case 1: im = c->image[jj  ];  corr += im * c->templat[t_indx  ];  sum += im;  sum_sq += im * im;
				}
				
				t_indx += rem;
			}
		}
		
		const flccPrecision var = sum_sq - sum*sum/((c->tw)*(c->th)*(c->td));
		if (var == 0)	c->lcc[indx] = 0.0;
		else			c->lcc[indx] = corr / sqrt(var);
	} 
	
	pthread_exit (NULL);	
} 

#ifndef NO_SINGLE_PREC
template void *lcorrelation3d_thread_generic<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *lcorrelation3d_thread_generic<double> (void *arg);
#endif


#ifndef NO_FFTW


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Thread functions for fconv3_host
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Pad a 3D signal
template <class flccPrecision>
void *padSignal3d (void *arg)
{
	const padSigParm<flccPrecision> *padSig = (padSigParm<flccPrecision> *)arg;
	
	for (int n=0; n<padSig->offset; n++) {
		const int indx = padSig->initialPosition + n;
		const int x = (indx%(padSig->sw*padSig->sd))%padSig->sd;
		const int y = (indx%(padSig->sw*padSig->sd))/padSig->sd;
		const int z = indx/(padSig->sw*padSig->sd);
	
		if (x<padSig->d && y<padSig->w && z<padSig->h) {
			padSig->signal[indx] = padSig->tempSignal[x + y*padSig->d + z*padSig->w*padSig->d];
		} 
		else {
			padSig->signal[indx] = 0.0;
		} 
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *padSignal3d<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *padSignal3d<double> (void *arg);
#endif


// Perform point-to-point multiplication
template <class fftwComplexPrecision>
void *complexProduct3d (void *arg)
{
	const prodParm<fftwComplexPrecision> *prod = (prodParm<fftwComplexPrecision> *)arg;

	for (int n=0; n<prod->offset; n++) {
		const int indx = prod->initialPosition + n;
		const fftwComplexPrecision f1 = {prod->fft1[indx][0], prod->fft1[indx][1]};
		const fftwComplexPrecision f2 = {prod->fft2[indx][0], prod->fft2[indx][1]};
		fftwComplexPrecision f;
		
		f[0] = (f1[0] * f2[0] - f1[1] * f2[1]) / prod->s_size;
		f[1] = (f1[0] * f2[1] + f1[1] * f2[0]) / prod->s_size; 
		
		prod->prod_fft[indx][0] = f[0]; 
		prod->prod_fft[indx][1] = f[1]; 
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *complexProduct3d<fftwf_complex> (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *complexProduct3d<fftw_complex>  (void *arg);
#endif


// Create a 3D convolution array
template <class flccPrecision>
void *makeConvArray3d (void *arg)
{
	const makeConvParm<flccPrecision> *makeConv = (makeConvParm<flccPrecision> *)arg;
	
	for (int n=0; n<makeConv->offset; n++) {
		const int indx = makeConv->initialPosition + n;
		const int x = (indx%(makeConv->sw*makeConv->sd))%makeConv->sd;
		const int y = (indx%(makeConv->sw*makeConv->sd))/makeConv->sd;
		const int z = indx/(makeConv->sw*makeConv->sd);
		
		makeConv->conv[indx] = makeConv->sceneTimesT[x + y*makeConv->SD + z*makeConv->SW*makeConv->SD];
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *makeConvArray3d<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *makeConvArray3d<double> (void *arg);
#endif


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Thread functions for flcorr3_host
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Normalize template and create template spatial support
template <class flccPrecision>
void *makePermanent3d (void *arg)
{
	const permParm<flccPrecision> *perm = (permParm<flccPrecision> *)arg;
	
	for (int n=0; n<perm->offset; n++) {
		const int indx = perm->initialPosition + n;
		const int x = (indx%(perm->sw*perm->sd))%perm->sd;
		const int y = (indx%(perm->sw*perm->sd))/perm->sd;
		const int z = indx/(perm->sw*perm->sd);
	
		if (x<perm->td && y<perm->tw && z<perm->th) {
			perm->templat[indx] = ( perm->tempTemplate[perm->tw*perm->th*perm->td - 1 - (x + y*perm->td + z*perm->tw*perm->td)] - perm->mean ) / perm->std; // Normalize here
			perm->Bt[indx]      = 1.0;
		}
		else {
			perm->templat[indx] = 0.0;
			perm->Bt[indx]      = 0.0;
		} 
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *makePermanent3d<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *makePermanent3d<double> (void *arg);
#endif

	
// Pad image and create square-valued image
template <class flccPrecision>
void *makeTransient3d (void *arg)
{
	const tranParm<flccPrecision> *tran = (tranParm<flccPrecision> *)arg;
	
	for (int n=0; n<tran->offset; n++) {
		const int indx = tran->initialPosition + n;
		const int x = (indx%(tran->sw*tran->sd))%tran->sd;
		const int y = (indx%(tran->sw*tran->sd))/tran->sd;
		const int z = indx/(tran->sw*tran->sd);
	
		if (x<tran->id && y<tran->iw && z<tran->ih) {
			const flccPrecision im = tran->image[x + y*tran->id + z*tran->iw*tran->id];
			tran->scene[indx]   = im;
			tran->sceneSq[indx] = im * im;
		} 
		else {
			tran->scene[indx]   = 0.0;
			tran->sceneSq[indx] = 0.0;
		} 
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *makeTransient3d<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *makeTransient3d<double> (void *arg);
#endif

	
// Calculate a 2D LCC array from precalculated arrays
template <class flccPrecision>
void *makeLccArray3d (void *arg)
{
	const makeLccParm<flccPrecision> *makeLcc = (makeLccParm<flccPrecision> *)arg;

	for (int n=0; n<makeLcc->offset; n++) {
		const int indx = makeLcc->initialPosition + n;
		const int x = (indx%(makeLcc->sw*makeLcc->sd))%makeLcc->sd;
		const int y = (indx%(makeLcc->sw*makeLcc->sd))/makeLcc->sd;
		const int z = indx/(makeLcc->sw*makeLcc->sd);
		const int INDX = x + y*makeLcc->SD + z*makeLcc->SW*makeLcc->SD;
		const flccPrecision SBt = makeLcc->sceneTimesBt[INDX];
		
		const flccPrecision nom   =    makeLcc->sceneTimesT[INDX];
		const flccPrecision denom = makeLcc->sceneSqTimesBt[INDX] - (SBt * SBt)/(makeLcc->t_size);
		
		if (denom == 0)	makeLcc->lcc[indx] = 0.0;
		else 			makeLcc->lcc[indx] = nom/sqrt(denom);
	} 
	
	pthread_exit (NULL);
} 

#ifndef NO_SINGLE_PREC
template void *makeLccArray3d<float>  (void *arg);
#endif
#ifndef NO_DOUBLE_PREC
template void *makeLccArray3d<double> (void *arg);
#endif


#endif // NO_FFTW

