/* 
 * FILE:
 * conv3_dev.cu
 * version 1.1
 *
 * DESCRIPTION:
 * This function computes the 3D convolution given an image and a template on a GPU device.
 * It implements the direct method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult conv3_dev (
		flccPrecision *image,   flccSize imSize, 
		flccPrecision *templat, flccSize temSize, 
		flccPrecision *conv, int imCount )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int TD = temSize.d;
	const int T_SIZE = TH*TW*TD;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int ID = imSize.d;
	const int I_SIZE = IH*IW*ID;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int rh = (THREADS_3D - sh%THREADS_3D)%THREADS_3D;
	const int SH = sh + rh;
	const int sw = IW+TW-1;
	const int rw = (THREADS_3D - sw%THREADS_3D)%THREADS_3D;
	const int SW = sw + rw;
	const int sd = ID+TD-1;
	const int rd = (THREADS_3D - sd%THREADS_3D)%THREADS_3D;
	const int SD = sd + rd;
	const int s_size = sh*sw*sd;
	const int S_SIZE = SH*SW*SD;
	
	// Double padded arrays dimensions
	const int BH = SH+TH-1;
	const int BW = SW+TW-1;
	const int BD = SD+TD-1;
	const int B_SIZE = BH*BW*BD;
	
	
	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_3D, THREADS_3D, THREADS_3D);

	// Find number of blocks in grid
	dim3 dimGrid_B ( ((BD+dimBlock.x-1)/dimBlock.x) * ((BW+dimBlock.y-1)/dimBlock.y), (BH+dimBlock.z-1)/dimBlock.z );
	dim3 dimGrid_S ( (SD/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );

	// Declare streams
	cudaStream_t host2devMemcpy, dev2hostMemcpy, workExec;
	
	
	// Array with function pointers to kernels
	void (*convolution3d_kernel[])(flccPrecision *, flccPrecision *, flccPrecision *, int, int, int, int) = {
		convolution3d_kernel_2x2x2<flccPrecision>,    convolution3d_kernel_3x3x3<flccPrecision>,    convolution3d_kernel_4x4x4<flccPrecision>,   
		convolution3d_kernel_5x5x5<flccPrecision>,    convolution3d_kernel_6x6x6<flccPrecision>,    convolution3d_kernel_7x7x7<flccPrecision>,
		convolution3d_kernel_8x8x8<flccPrecision> 
	};
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Transfer template to device
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	flccPrecision *templateDev;
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&templateDev, T_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMemcpy (templateDev, templat, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make requisite arrays
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Device variables' declaration
	flccPrecision *imageDev, *tempImageDev, *convDev, *convDevDepadded;

	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&imageDev,        B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempImageDev,    I_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDev,         S_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDevDepadded, s_size*sizeof(flccPrecision)) );

	// Create streams
	CUDA_SAFE_CALL ( cudaStreamCreate(&host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&workExec) );
	
	
	// Transfer first image
	CUDA_SAFE_CALL ( cudaMemcpy (tempImageDev, image, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
	
	// Begin streaming of images
	for (int k=0; k<imCount; k++)
	{
		// Pad image with zeros
		padSignal3d_both<flccPrecision> <<<dimGrid_B, dimBlock, 0, host2devMemcpy>>> (
				imageDev, 
				tempImageDev,
				TH, TW, TD,
				IH, IW, ID,
				BH, BW, BD );
	
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Memory transfer
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Write out previous result
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (conv+(k-1)*s_size, convDevDepadded, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost, dev2hostMemcpy) );
		
		// Transfer next image
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (tempImageDev, image+(k+1)*I_SIZE, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice, host2devMemcpy) );

	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Convolution
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Launch kernel function
		if (TH==TW && TH==TD && TH<=KERNEL_MAX_SIZE_3D)
		{
			convolution3d_kernel[TH-2] <<<dimGrid_S, dimBlock, 0, workExec>>> (
					imageDev, 
					templateDev,
					convDev,
					SW, SD,
					BW, BD );
		}	
		else
		{
			convolution3d_kernel_generic<flccPrecision> <<<dimGrid_S, dimBlock, 0, workExec>>> (
					imageDev, 
					templateDev,
					convDev,
					TH, TW, TD,
					SW, SD,
					BW, BD );
		}

		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		// Depad image
		depadSignal3d<flccPrecision> <<<dimGrid_S, dimBlock, 0, dev2hostMemcpy>>> (
				convDevDepadded, 
				convDev,
				sh, sw, sd,
				SW, SD );	
	}

	// Write out last result
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	CUDA_SAFE_CALL ( cudaMemcpy (conv+(imCount-1)*s_size, convDevDepadded, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Destroy streams
	CUDA_SAFE_CALL ( cudaStreamDestroy (host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (workExec) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (tempImageDev) );
	CUDA_SAFE_CALL ( cudaFree (imageDev) );
	CUDA_SAFE_CALL ( cudaFree (templateDev) );
	CUDA_SAFE_CALL ( cudaFree (convDev) );
	CUDA_SAFE_CALL ( cudaFree (convDevDepadded) );

	
	return FLCC_SUCCESS;
}   


// make instances
#ifndef NO_SINGLE_PREC
template flccResult conv3_dev<float> (
		float *image,   flccSize imSize, 
		float *templat, flccSize temSize, 
		float *conv, int imCount );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult conv3_dev<double> (
		double *image,   flccSize imSize, 
		double *templat, flccSize temSize, 
		double *conv, int imCount );
#endif
