/* 
 * FILE:
 * conv3_host.c
 * version 1.0
 *
 * DESCRIPTION:
 * This function computes the 3D convolution given an image and a template on the CPU host.
 * It implements the direct method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * July 2010
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult conv3_host (
		flccPrecision *image,   flccSize imageDims, 
		flccPrecision *templat, flccSize templateDims, 
		flccPrecision *conv, 
		int imCount, int n )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = templateDims.h;
	const int TW = templateDims.w;
	const int TD = templateDims.d;
	
	// Image dimensions
	const int IH = imageDims.h;
	const int IW = imageDims.w;
	const int ID = imageDims.d;
	const int I_SIZE = IH*IW*ID;
	
	// Double padded arrays dimensions
	const int BH = IH+2*TH-2;
	const int BW = IW+2*TW-2;
	const int BD = ID+2*TD-2;
	const int B_SIZE = BH*BW*BD;
	
	// Single padded arrays dimensions
	const int SH = IH+TH-1;
	const int SW = IW+TW-1;
	const int SD = ID+TD-1;
	const int S_SIZE = SH*SW*SD;
	
	// Auxiliary variables
	int i, k;
	int p_temp;
	int c_temp;
	
	// Declare and allocate for Array padding
	flccPrecision *paddedImage;
	MALLOC_SAFE_CALL ( paddedImage = (flccPrecision *)malloc (B_SIZE*sizeof(flccPrecision)) );
	
	// Thread variables' declaration and allocation
	pthread_t *threads;
	pthread_attr_t pthread_custom_attr;
	padParm<flccPrecision>  *p;
	corrParm<flccPrecision> *c;
	void *status;
	
	MALLOC_SAFE_CALL ( threads = (pthread_t *)malloc (n*sizeof(pthread_t)) );
	
	PTHREAD_SAFE_CALL ( pthread_attr_init (&pthread_custom_attr) );
	PTHREAD_SAFE_CALL ( pthread_attr_setdetachstate (&pthread_custom_attr, PTHREAD_CREATE_JOINABLE) );
	
	MALLOC_SAFE_CALL ( p =  (padParm<flccPrecision> *)malloc (sizeof(padParm<flccPrecision>)*n)  );
	MALLOC_SAFE_CALL ( c = (corrParm<flccPrecision> *)malloc (sizeof(corrParm<flccPrecision>)*n) );
	
	
	// Array with function pointers to threads
	void *(*convolution3d_thread[])(void *) = {
		convolution3d_thread_2x2x2<flccPrecision>,    convolution3d_thread_3x3x3<flccPrecision>,    convolution3d_thread_4x4x4<flccPrecision>,
		convolution3d_thread_5x5x5<flccPrecision>,    convolution3d_thread_6x6x6<flccPrecision>,    convolution3d_thread_7x7x7<flccPrecision>,
		convolution3d_thread_8x8x8<flccPrecision>,    convolution3d_thread_9x9x9<flccPrecision>,    convolution3d_thread_10x10x10<flccPrecision>
	};
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make arguments for threads
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	p_temp = 0;
	c_temp = 0;
	
	for (i=0; i<n; i++)
	{
		p[i].offset = B_SIZE/n;
		
		if (i < (B_SIZE%n))
			p[i].offset++;
		
	    p[i].initialPosition = p_temp;
	    p_temp += p[i].offset;
		
		p[i].padded   = paddedImage;
		
		p[i].th = TH;
		p[i].tw = TW;
		p[i].td = TD;
		p[i].sh = SH;
		p[i].sw = SW;
		p[i].sd = SD;
		p[i].iw = IW;
		p[i].id = ID;
		p[i].bw = BW;
		p[i].bd = BD;
		
		c[i].offset = S_SIZE/n;
		
		if (i < (S_SIZE%n))
			c[i].offset++;
			
	    c[i].initialPosition = c_temp;
	    c_temp += c[i].offset;
		
		c[i].image = paddedImage;
		c[i].templat = templat;
		
		c[i].th = TH;
		c[i].tw = TW;
		c[i].td = TD;
		c[i].sw = SW;
		c[i].sd = SD;
		c[i].bw = BW;
		c[i].bd = BD;
		
    } 
	
	
	// Begin streaming of images
	for (k=0; k<imCount; k++)
	{
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Array padding
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Start threads
		for (i=0; i<n; i++)
		{
			p[i].unpadded = image + k*I_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, padSignal3d_both<flccPrecision>, (void *)(p+i)) );
		}
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
		
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Convolution
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		void *(*convolve3d)(void *);
		if (TH==TW && TH==TD && TH<=THREAD_MAX_SIZE_3D)	convolve3d = convolution3d_thread[TH-2];
		else							                convolve3d = convolution3d_thread_generic<flccPrecision>;
	
		// Start threads
		for (i=0; i<n; i++)
		{
			c[i].lcc = conv + k*S_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, convolve3d, (void *)(c+i)) );
		}
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	}
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Destroy pthread attribute
	PTHREAD_SAFE_CALL ( pthread_attr_destroy (&pthread_custom_attr) );
	
	// Free threads' variables
	free (threads);
	free (p);
	free (c);
	
	// Free for Array padding
	free (paddedImage);
	
	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult conv3_host<float> (
		float *image,   flccSize imageDims, 
		float *templat, flccSize templateDims, 
		float *conv, 
		int imCount, int n );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult conv3_host<double> (
		double *image,   flccSize imageDims, 
		double *templat, flccSize templateDims, 
		double *conv, 
		int imCount, int n );
#endif
