/* 
 * FILE:
 * fconv2_dev.cu
 * version 1.2
 *
 * DESCRIPTION:
 * This function computes the 2D convolution given an image and a template on a GPU device.
 * It computes the convolution using the FFT method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 
 
 // System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult fconv2_dev (
		flccPrecision *image,   flccSize imSize, 
		flccPrecision *templat, flccSize temSize, 
		flccPrecision *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int I_SIZE = IH*IW;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize2d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize2d(sw);
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	const int H_SIZE = SH*(SW/2 + 1);
	
	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_H ( ((SW/2 + 1)+dimBlock.x-1)/dimBlock.x, SH/dimBlock.y );
	
	// Declare streams
	cudaStream_t host2devMemcpy, dev2hostMemcpy, workExec;
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage A (computation once for all)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision *templateDev; // real array in device containing the template (padded - size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&templateDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMemset (templateDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Transfer data to device
	CUDA_SAFE_CALL ( cudaMemcpy2D (templateDev, SW*sizeof(cufftRealPrecision), templat, TW*sizeof(flccPrecision), TW*sizeof(flccPrecision), TH, cudaMemcpyHostToDevice) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(T)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variable declaration
	cufftComplexPrecision *fft_templateDev; // complex array containing the FFT of the template
	
	// Allocate memory in device for requisite array
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_templateDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	
	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, templateDev, fft_templateDev) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (templateDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage B (computation for each scene)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad scene
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *sceneDev; // real array in device containing the scene (padded - size S)
	flccPrecision        *tempSceneDev; // real array in device containing the scene (not padded - size I)
	cufftComplexPrecision *fft_sceneDev; // complex array containing the FFT of the scene
	cufftComplexPrecision *fft_SceneTimesTDev; // complex array containing F(S) * F(T)
	cufftRealPrecision    *time_SceneTimesTDev; // real array containing IF( F(S) * F(T) )
	flccPrecision        *convDev;
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&sceneDev,            S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempSceneDev,        I_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,        H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_SceneTimesTDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_SceneTimesTDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDev,             s_size*sizeof(flccPrecision)) );
	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (sceneDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Create streams for concurrent memory transfer
	CUDA_SAFE_CALL ( cudaStreamCreate(&host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&workExec) );
	
	
	// Transfer first image
	CUDA_SAFE_CALL ( cudaMemcpy2D (sceneDev, SW*sizeof(cufftRealPrecision), image, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyHostToDevice) );
	
	// Begin streaming of images
	for (int k=0; k<imCount; k++)
	{
		// Pad image with zeros (unless it's the first one)
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpy2DAsync (sceneDev, SW*sizeof(cufftRealPrecision), tempSceneDev, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyDeviceToDevice, host2devMemcpy) );
				
		// Synchronize among threads
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
			
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Memory transfer
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Transfer back previous image
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (conv+(k-1)*s_size, convDev, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost, dev2hostMemcpy) );
		
		// Transfer next image
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (tempSceneDev, image+(k+1)*I_SIZE, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice, host2devMemcpy) );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make F(S)
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream (planR2C, workExec) );
		CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, sceneDev, fft_sceneDev) );
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( F(T) * F(S) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Launch kernel function
		complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock, 0, workExec>>> (
				fft_SceneTimesTDev, 
				fft_sceneDev, 
				fft_templateDev,
				(SW/2 + 1), S_SIZE);
			
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaStreamSynchronize(workExec) );

		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream (planC2R, workExec) );
		CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_SceneTimesTDev, time_SceneTimesTDev) );
		
		// Synchronize among threads
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		// Depad image (unless it's the last one)
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpy2DAsync (convDev, sw*sizeof(flccPrecision), time_SceneTimesTDev, SW*sizeof(cufftRealPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToDevice, dev2hostMemcpy) );
	} 
	
	// Write out last result
	CUDA_SAFE_CALL ( cudaMemcpy2D (conv+(imCount-1)*s_size, sw*sizeof(flccPrecision), time_SceneTimesTDev, SW*sizeof(cufftRealPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToHost) );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Associate plans with the default stream again before destroying streams
	CUFFT_SAFE_CALL ( cufftSetStream (planR2C, 0) );
	CUFFT_SAFE_CALL ( cufftSetStream (planC2R, 0) );
	
	// Destroy streams
	CUDA_SAFE_CALL ( cudaStreamDestroy (host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (workExec) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (fft_templateDev) );
	CUDA_SAFE_CALL ( cudaFree (tempSceneDev) );
	CUDA_SAFE_CALL ( cudaFree (sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (convDev) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult fconv2_dev<float, cufftReal, cufftComplex> (
		float *image,   flccSize imSize, 
		float *templat, flccSize temSize, 
		float *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult fconv2_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *image,   flccSize imSize, 
		double *templat, flccSize temSize, 
		double *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
