/* 
 * FILE:
 * fconv2_host.c
 * version 1.0
 *
 * DESCRIPTION:
 * This function computes the 2D convolution given an image and a template on the CPU host.
 * It computes the convolution using the FFT method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * May 2010
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_fftw_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class fftw_plan_t, class fftwComplexPrecision>
flccResult fconv2_host (
		flccPrecision *image,   flccSize imageDims, 
		flccPrecision *templat, flccSize templateDims, 
		flccPrecision *conv, 
		int imCount, int n, 
		fftw_plan_t c2r_plan_diff,
		fftw_plan_t r2c_plan_diff ) 
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = templateDims.h;
	const int TW = templateDims.w;
	
	// Image dimensions
	const int IH = imageDims.h;
	const int IW = imageDims.w;
	const int I_SIZE = IH*IW;
	
	// Single padded arrays dimensions
	const int sh = IH+TH-1;		
	const int SH = alignSize2d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize2d(sw);
	const int s_size = sh*sw;		  // Logical FFT size
	const int S_SIZE = SH*SW;         // FFT size after padding
	const int H_SIZE = SH*(SW/2 + 1); // Complex Hermitian size

	// Auxiliary variables
	int i, k;
	int S_offset, s_offset, H_offset;
	int S_temp, s_temp, H_temp;
	void *status;
	
	// Declare and allocate for Pad template (T) in the time domain
	flccPrecision *time_template;
	MALLOC_SAFE_CALL ( time_template = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate for Make F(T)
	fftwComplexPrecision *fft_template;
	MALLOC_SAFE_CALL ( fft_template = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	
	// Declare and allocate for Make scene (S) in the time domain
	flccPrecision *time_scene;
	MALLOC_SAFE_CALL ( time_scene = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate for Make F(S)
	fftwComplexPrecision *fft_scene;
	MALLOC_SAFE_CALL ( fft_scene = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	
	// Declare and allocate for Make IF ( F(T) * F(S) )
	fftwComplexPrecision *fft_sceneTimesT;
	MALLOC_SAFE_CALL ( fft_sceneTimesT = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	
	flccPrecision *time_sceneTimesT;
	MALLOC_SAFE_CALL ( time_sceneTimesT = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate thread variables
	pthread_t *threads;
	pthread_attr_t pthread_custom_attr;
	padSigParm<flccPrecision>   *padSig;
	prodParm<fftwComplexPrecision> *prod;
	makeConvParm<flccPrecision> *makeConv;
	
	MALLOC_SAFE_CALL ( threads = (pthread_t *)malloc (n*sizeof(pthread_t)) );
	PTHREAD_SAFE_CALL ( pthread_attr_init (&pthread_custom_attr) );
	PTHREAD_SAFE_CALL ( pthread_attr_setdetachstate (&pthread_custom_attr, PTHREAD_CREATE_JOINABLE) );
	
	MALLOC_SAFE_CALL ( padSig = (padSigParm<flccPrecision> *)malloc (sizeof(padSigParm<flccPrecision>)*n) );
	MALLOC_SAFE_CALL ( prod = (prodParm<fftwComplexPrecision> *)malloc (sizeof(prodParm<fftwComplexPrecision>)*n) );
	MALLOC_SAFE_CALL ( makeConv = (makeConvParm<flccPrecision> *)malloc (sizeof(makeConvParm<flccPrecision>)*n) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage A (computation once for all)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Make arguments for threads
	S_temp = 0;
	s_temp = 0;
	H_temp = 0;
	for (i=0; i<n; i++)
	{
		S_offset = S_SIZE/n;
		s_offset = s_size/n;
		H_offset = H_SIZE/n;
		if (i < (S_SIZE%n))	 S_offset++;
		if (i < (s_size%n))	 s_offset++;
		if (i < (H_SIZE%n))	 H_offset++;
			
		padSig[i].offset   = S_offset;
		prod[i].offset     = H_offset;
		makeConv[i].offset = s_offset;
		
	    padSig[i].initialPosition   = S_temp;
		prod[i].initialPosition     = H_temp;
		makeConv[i].initialPosition = s_temp;
		
	    S_temp += S_offset;
		s_temp += s_offset;
		H_temp += H_offset;
		
		//--------
		
		padSig[i].signal     = time_template;
		padSig[i].tempSignal = templat;
		
		padSig[i].w  = TW;
		padSig[i].h  = TH;
		padSig[i].sw = SW;
		
		//--------
		
		prod[i].prod_fft = fft_sceneTimesT;
		prod[i].fft1     = fft_scene;
		prod[i].fft2     = fft_template;
			
		prod[i].s_size = S_SIZE;
		
		//--------

		makeConv[i].sceneTimesT = time_sceneTimesT;
		
		makeConv[i].sw = sw;
		makeConv[i].SW = SW;
		
    } 
	
	// Start threads
	for (i=0; i<n; i++)
		PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, padSignal2d<flccPrecision>, (void *)(padSig+i)) );
	
	// Synchronize the completion of each thread
	for (i=0; i<n; i++)
		PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(T)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Use plan to transform data
	fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_template, fft_template);
	
	// Free for Pad template (T) in the time domain
	fftw_free_t<flccPrecision> (time_template);

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage B (computation for each scene)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad scene
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	for (i=0; i<n; i++)
	{
		padSig[i].w = IW;
		padSig[i].h = IH;
		padSig[i].signal = time_scene;
	} 
	
	for (k=0; k<imCount; k++)
	{
		// Start threads
		for (i=0; i<n; i++)
		{
			padSig[i].tempSignal = image + k*I_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, padSignal2d<flccPrecision>, (void *)(padSig+i)) );
		} // end for
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make F(S)
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Use plan to transform data in place 
		fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_scene, fft_scene);
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( F(T) * F(S) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Start threads
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, complexProduct2d<fftwComplexPrecision>, (void *)(prod+i)) );
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	
		if (S_SIZE == s_size)
		{
			// Use plan to transform data
			fftw_execute_dft_c2r_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (c2r_plan_diff, fft_sceneTimesT, conv + k*s_size);
		}
		else 
		{
			// Use plan to transform data
			fftw_execute_dft_c2r_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (c2r_plan_diff, fft_sceneTimesT, time_sceneTimesT);
		
			// Start threads
			for (i=0; i<n; i++)
			{
				makeConv[i].conv = conv + k*s_size;
				PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, makeConvArray2d<flccPrecision>, (void *)(makeConv+i)) );
			} // end for
		
			// Synchronize the completion of each thread
			for (i=0; i<n; i++)
				PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
		}
	
	} 
	

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Destroy pthread attribute
	PTHREAD_SAFE_CALL ( pthread_attr_destroy (&pthread_custom_attr) );
	
	// Free thread variables
	free (threads);
	free (padSig);
	free (prod);
	free (makeConv);
	
	
	// Free for Make F(T)
	fftw_free_t<flccPrecision> (fft_template);
	
	// Free for Make scene (S) in the time domain
	fftw_free_t<flccPrecision> (time_scene);
	
	// Free for Make F(S)
	fftw_free_t<flccPrecision> (fft_scene);
	
	// Free for Make IF ( F(T) * F(S) )
	fftw_free_t<flccPrecision> (fft_sceneTimesT);
	fftw_free_t<flccPrecision> (time_sceneTimesT);
	
	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult fconv2_host<float, fftwf_plan, fftwf_complex> (
		float *image,   flccSize imageDims, 
		float *templat, flccSize templateDims, 
		float *conv, 
		int imCount, int n, 
		fftwf_plan c2r_plan_diff,
		fftwf_plan r2c_plan_diff );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult fconv2_host<double, fftw_plan, fftw_complex> (
		double *image,   flccSize imageDims, 
		double *templat, flccSize templateDims, 
		double *conv, 
		int imCount, int n, 
		fftw_plan c2r_plan_diff,
		fftw_plan r2c_plan_diff );
#endif
