/* 
 * FILE:
 * fconv2_testStream_dev.cu
 * version 1.2
 *
 * DESCRIPTION:
 * This function executes only the transient stage of the convolution computation using FFTs and it is used as a benchmark 
 * for performance measurements of the 2D image streaming process. It runs on the GPU.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult fconv2_testStream_dev (
		flccPrecision *tempSceneDev, flccSize imSize, 
		cufftComplexPrecision *fft_templateDev, flccSize temSize, 
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize2d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize2d(sw);
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	const int H_SIZE = SH*(SW/2 + 1);

	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_H ( ((SW/2 + 1)+dimBlock.x-1)/dimBlock.x, SH/dimBlock.y );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage B (computation for each scene)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad scene
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *sceneDev;            // real array in device containing the scene (padded - size S)
	cufftComplexPrecision *fft_sceneDev;        // complex array containing the FFT of the scene
	cufftComplexPrecision *fft_SceneTimesTDev;  // complex array containing F(S) * F(T)
	cufftRealPrecision    *time_SceneTimesTDev; // real array containing IF( F(S) * F(T) )
	flccPrecision        *convDev;             // not trimmed (size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&sceneDev,     	    S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,        H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_SceneTimesTDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_SceneTimesTDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDev,             s_size*sizeof(flccPrecision)) );

	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (sceneDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Pad image with zeros
	CUDA_SAFE_CALL ( cudaMemcpy2D (sceneDev, SW*sizeof(cufftRealPrecision), tempSceneDev, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyDeviceToDevice) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(S)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, sceneDev, fft_sceneDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make IF ( F(T) * F(S) )
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
	// Launch kernel function
	complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock>>> (
			fft_SceneTimesTDev, 
			fft_sceneDev, 
			fft_templateDev,
			(SW/2 + 1), S_SIZE);
		
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_SceneTimesTDev, time_SceneTimesTDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute Convolution Array
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	CUDA_SAFE_CALL ( cudaMemcpy2D (convDev, sw*sizeof(flccPrecision), time_SceneTimesTDev, SW*sizeof(cufftRealPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToDevice) );
	

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (convDev) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult fconv2_testStream_dev<float, cufftReal, cufftComplex> (
		float *tempSceneDev, flccSize imSize, 
		cufftComplex *fft_templateDev, flccSize temSize, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult fconv2_testStream_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *tempSceneDev, flccSize imSize, 
		cufftDoubleComplex *fft_templateDev, flccSize temSize, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
