/* 
 * FILE:
 * fconv3_dev.cu
 * version 1.1
 *
 * DESCRIPTION:
 * This function computes the 3D convolution given an image and a template on a GPU device.
 * It computes the convolution using the FFT method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 
 
 // System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult fconv3_dev (
		flccPrecision *image,   flccSize imSize, 
		flccPrecision *templat, flccSize temSize, 
		flccPrecision *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int TD = temSize.d;
	const int T_SIZE = TH*TW*TD;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int ID = imSize.d;
	const int I_SIZE = IH*IW*ID;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize3d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize3d(sw);
	const int sd = ID+TD-1;
	const int SD = alignSize3d(sd);
	const int s_size = sh*sw*sd;
	const int S_SIZE = SH*SW*SD;
	const int H_SIZE = SH*SW*(SD/2 + 1);

	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_3D, THREADS_3D, THREADS_3D);

	// Find number of blocks in grid
	dim3 dimGrid   ( (SD/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );
	dim3 dimGrid_s ( ((sd+dimBlock.x-1)/dimBlock.x) * ((sw+dimBlock.y-1)/dimBlock.y), (sh+dimBlock.z-1)/dimBlock.z );
	dim3 dimGrid_H ( (((SD/2 + 1)+dimBlock.x-1)/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );
	
	// Declare streams
	cudaStream_t host2devMemcpy, dev2hostMemcpy, workExec;
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage A (computation once for all)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision *templateDev; // real array in device containing the template (padded - size S)
	flccPrecision     *tempTemplateDev; // real array in device containing the template (not padded - size T)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&templateDev,     S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempTemplateDev, T_SIZE*sizeof(flccPrecision)) );
	
	// Transfer data to device
	CUDA_SAFE_CALL ( cudaMemcpy (tempTemplateDev, templat, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
	
	// Launch kernel function
	padSignal3d<flccPrecision, cufftRealPrecision> <<<dimGrid, dimBlock>>> (
			templateDev, 
			tempTemplateDev, 
			TH, TW, TD,
			SW, SD );
	
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (tempTemplateDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(T)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variable declaration
	cufftComplexPrecision *fft_templateDev; // complex array containing the FFT of the template
	
	// Allocate memory in device for requisite array
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_templateDev, H_SIZE*sizeof(cufftComplexPrecision)) );
	
	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, templateDev, fft_templateDev) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (templateDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage B (computation for each scene)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad scene
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *sceneDev; // real array in device containing the scene (padded - size S)
	flccPrecision        *tempSceneDev; // real array in device containing the scene (not padded - size I)
	cufftComplexPrecision *fft_sceneDev; // complex array containing the FFT of the scene
	cufftComplexPrecision *fft_SceneTimesTDev; // complex array containing F(S) * F(T)
	cufftRealPrecision    *time_SceneTimesTDev; // real array containing IF( F(S) * F(T) )
	flccPrecision        *convDev; // not trimmed (size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&sceneDev,            S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempSceneDev,        I_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,        H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_SceneTimesTDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_SceneTimesTDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDev,             s_size*sizeof(flccPrecision)) );
	
	// Create streams for concurrent memory transfer
	CUDA_SAFE_CALL ( cudaStreamCreate(&host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&workExec) );
	
	
	// Transfer first image
	CUDA_SAFE_CALL ( cudaMemcpy (tempSceneDev, image, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
	
	// Begin streaming of images
	for (int k=0; k<imCount; k++)
	{
		// Pad image with zeros
		padSignal3d<flccPrecision, cufftRealPrecision> <<<dimGrid, dimBlock, 0, host2devMemcpy>>> (
				sceneDev, 
				tempSceneDev, 
				IH, IW, ID,
				SW, SD );
	
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Memory transfer
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Write out previous result
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (conv+(k-1)*s_size, convDev, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost, dev2hostMemcpy) );
		
		// Transfer next image
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (tempSceneDev, image+(k+1)*I_SIZE, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice, host2devMemcpy) );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make F(S)
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream(planR2C, workExec) );
		CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, sceneDev, fft_sceneDev) );
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( F(T) * F(S) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Launch kernel function
		complexProduct3d<cufftComplexPrecision> <<<dimGrid_H, dimBlock, 0, workExec>>> (
				fft_SceneTimesTDev, 
				fft_sceneDev, 
				fft_templateDev,
				SW, (SD/2 + 1),
				S_SIZE);
			
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaStreamSynchronize(workExec) );

		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream(planC2R, workExec) );
		CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_SceneTimesTDev, time_SceneTimesTDev) );

		// Synchronize among threads
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		// Depad image 
		makeConvArray3d<flccPrecision, cufftRealPrecision> <<<dimGrid_s, dimBlock, 0, dev2hostMemcpy>>> (
				convDev, 
				time_SceneTimesTDev, 
				sh, sw, sd,
				SW, SD );
	} 

	// Write out last result
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	CUDA_SAFE_CALL ( cudaMemcpy (conv+(imCount-1)*s_size, convDev, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Associate plans with the default stream again before destroying streams
	CUFFT_SAFE_CALL ( cufftSetStream(planR2C, 0) );
	CUFFT_SAFE_CALL ( cufftSetStream(planC2R, 0) );
	
	// Destroy streams
	CUDA_SAFE_CALL ( cudaStreamDestroy (host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (workExec) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (fft_templateDev) );
	CUDA_SAFE_CALL ( cudaFree (tempSceneDev) );
	CUDA_SAFE_CALL ( cudaFree (sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (convDev) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult fconv3_dev<float, cufftReal, cufftComplex> (
		float *image,   flccSize imSize, 
		float *templat, flccSize temSize, 
		float *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult fconv3_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *image,   flccSize imSize, 
		double *templat, flccSize temSize, 
		double *conv, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
