/* 
 * FILE:
 * fconv3_testStream_dev.cu
 * version 1.1
 *
 * DESCRIPTION:
 * This function executes only the transient stage of the convolution computation using FFTs and it is used as a benchmark 
 * for performance measurements of the 3D image streaming process. It runs on the GPU.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult fconv3_testStream_dev (
		flccPrecision *tempSceneDev, flccSize imSize, 
		cufftComplexPrecision *fft_templateDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int TD = temSize.d;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int ID = imSize.d;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize3d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize3d(sw);
	const int sd = ID+TD-1;
	const int SD = alignSize3d(sd);
	const int s_size = sh*sw*sd;
	const int S_SIZE = SH*SW*SD;
	const int H_SIZE = SH*SW*(SD/2 + 1);

	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_3D, THREADS_3D, THREADS_3D);

	// Find number of blocks in grid
	dim3 dimGrid   ( (SD/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );
	dim3 dimGrid_s ( ((sd+dimBlock.x-1)/dimBlock.x) * ((sw+dimBlock.y-1)/dimBlock.y), (sh+dimBlock.z-1)/dimBlock.z );
	dim3 dimGrid_H ( (((SD/2 + 1)+dimBlock.x-1)/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage B (computation for each scene)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Pad scene
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *sceneDev;            // real array in device containing the scene (padded - size S)
	cufftComplexPrecision *fft_sceneDev;        // complex array containing the FFT of the scene
	cufftComplexPrecision *fft_SceneTimesTDev;  // complex array containing F(S) * F(T)
	cufftRealPrecision    *time_SceneTimesTDev; // real array containing IF( F(S) * F(T) )
	flccPrecision        *convDev;             // not trimmed (size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&sceneDev,            S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,        H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_SceneTimesTDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_SceneTimesTDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&convDev,             s_size*sizeof(flccPrecision)) );

	// Launch kernel function
	padSignal3d<flccPrecision, cufftRealPrecision> <<<dimGrid, dimBlock>>> (
			sceneDev, 
			tempSceneDev, 
			IH, IW, ID,
			SW, SD );
	
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(S)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, sceneDev, fft_sceneDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make IF ( F(T) * F(S) )
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
	// Launch kernel function
	complexProduct3d<cufftComplexPrecision> <<<dimGrid_H, dimBlock>>> (
			fft_SceneTimesTDev, 
			fft_sceneDev, 
			fft_templateDev,
			SW, (SD/2 + 1),
			S_SIZE);
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_SceneTimesTDev, time_SceneTimesTDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute Convolution Array
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Launch kernel function
	makeConvArray3d<flccPrecision, cufftRealPrecision> <<<dimGrid_s, dimBlock>>> (
			convDev, 
			time_SceneTimesTDev, 
			sh, sw, sd,
			SW, SD );
		
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_SceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (convDev) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult fconv3_testStream_dev<float, cufftReal, cufftComplex> (
		float *tempSceneDev, flccSize imSize, 
		cufftComplex *fft_templateDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult fconv3_testStream_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *tempSceneDev, flccSize imSize, 
		cufftDoubleComplex *fft_templateDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
