/* 
 * FILE:
 * flcorr2_dev.cu
 * version 1.3
 *
 * DESCRIPTION:
 * This function computes the 2D LCC distribution given an image and a template on a GPU device.
 * It implements the FLCC algorithm.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult flcorr2_dev (
		flccPrecision *image,   flccSize imSize, 
		flccPrecision *templat, flccSize temSize, 
		flccPrecision *lcc, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int T_SIZE = TH*TW;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int I_SIZE = IH*IW;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize2d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize2d(sw);
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	const int H_SIZE = SH*(SW/2 + 1);

	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_T ( (TW+dimBlock.x-1)/dimBlock.x, (TH+dimBlock.y-1)/dimBlock.y );	
	dim3 dimGrid_I ( (IW+dimBlock.x-1)/dimBlock.x, (IH+dimBlock.y-1)/dimBlock.y );	
	dim3 dimGrid_s ( (sw+dimBlock.x-1)/dimBlock.x, (sh+dimBlock.y-1)/dimBlock.y );
	dim3 dimGrid_H ( ((SW/2 + 1)+dimBlock.x-1)/dimBlock.x, SH/dimBlock.y );
	
	// Declare streams
	cudaStream_t host2devMemcpy, dev2hostMemcpy, workExec;
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage A (computation once for all)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make permanent arrays (T and Bt) in the time domain
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision *time_templateDev; // real array in device containing the template (padded - size S)
	cufftRealPrecision *time_BtDev;       // real binary array of spatial support of T (padded - size S)
	flccPrecision     *tempTemplateDev;  // real array in device containing the template (not padded - size T)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_templateDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_BtDev,       S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempTemplateDev,  T_SIZE*sizeof(flccPrecision)) );
	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (time_templateDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMemset (time_BtDev,       0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Transfer data to device
	CUDA_SAFE_CALL ( cudaMemcpy (tempTemplateDev, templat, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
	
	// Normalize template
	flccPrecision mean = findMean (templat, T_SIZE);
	flccPrecision std  = findStd (mean, templat, T_SIZE);
	if (std == 0)	return FLCC_INVALID_VALUE;
	//normalize(mean, std, templat, T_SIZE); // is made parallel
	
	// Launch kernel function
	makePermanent2d<flccPrecision, cufftRealPrecision> <<<dimGrid_T, dimBlock>>> (
			time_templateDev, 
			tempTemplateDev, 
			time_BtDev, 
			mean, std,
			TH, TW,
			SW );
	
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (tempTemplateDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(T) and F(Bt)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftComplexPrecision *fft_templateDev;  // complex array containing the FFT of the template (size H)
	cufftComplexPrecision *fft_BtDev;        // complex array containing the FFT of Bt (size H)
	
	// Allocate memory in device for requisite array
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_templateDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_BtDev,        H_SIZE*sizeof(cufftComplexPrecision)) );
	
	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_templateDev, fft_templateDev) );
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_BtDev, fft_BtDev) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (time_templateDev) );
	CUDA_SAFE_CALL ( cudaFree (time_BtDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stages B-D
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make transient arrays (S and S^2) in the time domain
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *time_sceneDev;	      // real array in device containing the scene (padded - size S)
	cufftRealPrecision    *time_sceneSqDev;	      // real array in device containing the scene^2 (padded - size S)
	flccPrecision        *tempSceneDev;           // real array in device containing the scene (not padded - size I)
	cufftComplexPrecision *fft_sceneDev;           // complex array in device containing F(S) (size H)
	cufftComplexPrecision *fft_sceneSqDev;         // complex array in device containing F(S^2) (size H)
	cufftComplexPrecision *fft_sceneTimesTDev;     // complex array containing F(S) * F(T) (size H)
	cufftRealPrecision    *time_sceneTimesTDev;    // real array containing IF( F(S) * F(T) ) (size H)
	cufftComplexPrecision *fft_sceneTimesBtDev;    // complex array containing F(S) * F(Bt) (size H)
	cufftComplexPrecision *fft_sceneSqTimesBtDev;  // complex array containing F(S^2) * F(Bt) (size H)
	cufftRealPrecision    *time_sceneTimesBtDev;   // real array containing IF( F(S) * F(Bt) ) (size S)
	cufftRealPrecision    *time_sceneSqTimesBtDev; // real array containing IF( F(S^2) * F(Bt) ) (size S)
	flccPrecision        *lccDev;                 // not trimmed (size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneDev,          S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneSqDev,        S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempSceneDev,           I_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,           H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneSqDev,         H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneTimesTDev,     H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneTimesTDev,    S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneTimesBtDev,    H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneSqTimesBtDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneTimesBtDev,   S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneSqTimesBtDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDev,                 s_size*sizeof(flccPrecision)) );
	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (time_sceneDev,   0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMemset (time_sceneSqDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Create streams for concurrent memory transfer
	CUDA_SAFE_CALL ( cudaStreamCreate(&host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&workExec) );
	
	
	// Transfer first image
	CUDA_SAFE_CALL ( cudaMemcpy (tempSceneDev, image, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );
	
	// Begin streaming of images
	for (int k=0; k<imCount; k++)
	{
		// Launch kernel function
		makeTransient2d<flccPrecision, cufftRealPrecision> <<<dimGrid_I, dimBlock, 0, host2devMemcpy>>> (
				time_sceneDev, 
				time_sceneSqDev, 
				tempSceneDev,
				IH, IW, 
				SW );
		
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
			
			
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Memory transfer
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					
		// Transfer data to host
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (lcc+(k-1)*s_size, lccDev, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost, dev2hostMemcpy) );
		
		// Transfer next image
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (tempSceneDev, image+(k+1)*I_SIZE, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice, host2devMemcpy) );
		
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make F(S) and F(S^2)
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream(planR2C, workExec) );
		CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_sceneDev, fft_sceneDev) );
		CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_sceneSqDev, fft_sceneSqDev) );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( F(T) * F(S) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Launch kernel function
		complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock, 0, workExec>>> (
				fft_sceneTimesTDev, 
				fft_sceneDev, 
				fft_templateDev,
				(SW/2 + 1), S_SIZE);
			
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaStreamSynchronize(workExec) );

		// Use plan to transform data
		CUFFT_SAFE_CALL ( cufftSetStream(planC2R, workExec) );
		CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneTimesTDev, time_sceneTimesTDev) );
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( (F(S) * F(Bt)) ) and IF( (F(S^2) * F(Bt)) ) 
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Launch kernel function
		complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock, 0, workExec>>> (
				fft_sceneTimesBtDev, 
				fft_sceneDev, 
				fft_BtDev,
				(SW/2 + 1), S_SIZE);
			
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaStreamSynchronize(workExec) );

		// Use plans to transform data
		CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneTimesBtDev, time_sceneTimesBtDev) );

		// Launch kernel function
		complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock, 0, workExec>>> (
				fft_sceneSqTimesBtDev, 
				fft_sceneSqDev, 
				fft_BtDev,
				(SW/2 + 1), S_SIZE);
			
		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaStreamSynchronize(workExec) );

		// Use plans to transform data
		CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneSqTimesBtDev, time_sceneSqTimesBtDev) );
		
		// Synchronize among threads
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		// Launch kernel function
		makeLccArray2d<flccPrecision, cufftRealPrecision> <<<dimGrid_s, dimBlock, 0, dev2hostMemcpy>>> (
				lccDev, 
				time_sceneTimesBtDev,
				time_sceneSqTimesBtDev,
				time_sceneTimesTDev, 
				sh, sw,
				SW, T_SIZE );
	}
	
	// Transfer data to host
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	CUDA_SAFE_CALL ( cudaMemcpy (lcc+(imCount-1)*s_size, lccDev, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Associate plans with the default stream again before destroying streams
	CUFFT_SAFE_CALL ( cufftSetStream(planR2C, 0) );
	CUFFT_SAFE_CALL ( cufftSetStream(planC2R, 0) );
	
	// Destroy streams
	CUDA_SAFE_CALL ( cudaStreamDestroy (host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (workExec) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (tempSceneDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneSqDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneSqDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneSqTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneSqTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_BtDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_templateDev) );
	
	
	return FLCC_SUCCESS;
}


// make instances
#ifndef NO_SINGLE_PREC
template flccResult flcorr2_dev<float, cufftReal, cufftComplex> (
		float *image,   flccSize imSize, 
		float *templat, flccSize temSize, 
		float *lcc, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C ); 
#endif
#ifndef NO_DOUBLE_PREC
template flccResult flcorr2_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *image,   flccSize imSize, 
		double *templat, flccSize temSize, 
		double *lcc, int imCount, 
		cufftHandle planC2R,
		cufftHandle planR2C ); 
#endif
