/* 
 * FILE:
 * flcorr2_testStream_dev.cu
 * version 1.3
 *
 * DESCRIPTION:
 * This function executes only the transient stage of the LCC computation using FFTs and it is used as a benchmark 
 * for performance measurements of the 2D image streaming process. It runs on the GPU.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */


// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class cufftRealPrecision, class cufftComplexPrecision>
flccResult flcorr2_testStream_dev (
		flccPrecision *tempSceneDev, flccSize imSize, 
		cufftComplexPrecision *fft_templateDev, cufftComplexPrecision *fft_BtDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int T_SIZE = TH*TW;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int SH = alignSize2d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize2d(sw);
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	const int H_SIZE = SH*(SW/2 + 1);

	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_I ( (IW+dimBlock.x-1)/dimBlock.x, (IH+dimBlock.y-1)/dimBlock.y );		
	dim3 dimGrid_s ( (sw+dimBlock.x-1)/dimBlock.x, (sh+dimBlock.y-1)/dimBlock.y );
	dim3 dimGrid_H ( ((SW/2 + 1)+dimBlock.x-1)/dimBlock.x, SH/dimBlock.y );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stages B-D
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make transient arrays (S and S^2) in the time domain
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Device variables' declaration
	cufftRealPrecision    *time_sceneDev;	      // real array in device containing the scene (padded - size S)
	cufftRealPrecision    *time_sceneSqDev;	      // real array in device containing the scene^2 (padded - size S)
	cufftComplexPrecision *fft_sceneDev;           // complex array in device containing F(S) (size H)
	cufftComplexPrecision *fft_sceneSqDev;         // complex array in device containing F(S^2) (size H)
	cufftComplexPrecision *fft_sceneTimesTDev;     // complex array containing F(S) * F(T) (size H)
	cufftRealPrecision    *time_sceneTimesTDev;    // real array containing IF( F(S) * F(T) ) (size H)
	cufftComplexPrecision *fft_sceneTimesBtDev;    // complex array containing F(S) * F(Bt) (size H)
	cufftComplexPrecision *fft_sceneSqTimesBtDev;  // complex array containing F(S^2) * F(Bt) (size H)
	cufftRealPrecision    *time_sceneTimesBtDev;   // real array containing IF( F(S) * F(Bt) ) (size S)
	cufftRealPrecision    *time_sceneSqTimesBtDev; // real array containing IF( F(S^2) * F(Bt) ) (size S)
	flccPrecision        *lccDev;                 // not trimmed (size S)
	
	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneDev,          S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneSqDev,        S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneDev,           H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneSqDev,         H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneTimesTDev,     H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneTimesTDev,    S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneTimesBtDev,    H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&fft_sceneSqTimesBtDev,  H_SIZE*sizeof(cufftComplexPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneTimesBtDev,   S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&time_sceneSqTimesBtDev, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDev,                 s_size*sizeof(flccPrecision)) );

	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (time_sceneDev,   0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaMemset (time_sceneSqDev, 0, S_SIZE*sizeof(cufftRealPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	// Launch kernel function
	makeTransient2d<flccPrecision, cufftRealPrecision> <<<dimGrid_I, dimBlock>>> (
			time_sceneDev, 
			time_sceneSqDev, 
			tempSceneDev,
			IH, IW, 
			SW );
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(S) and F(S^2)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_sceneDev, fft_sceneDev) );
	CUFFT_SAFE_CALL ( cufftExecR2C_t (planR2C, time_sceneSqDev, fft_sceneSqDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make IF ( F(T) * F(S) )
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Launch kernel function
	complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock>>> (
			fft_sceneTimesTDev, 
			fft_sceneDev, 
			fft_templateDev,
			(SW/2 + 1), S_SIZE);
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Use plan to transform data
	CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneTimesTDev, time_sceneTimesTDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make IF ( (F(S) * F(Bt)) ) and IF( (F(S^2) * F(Bt)) ) 
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Launch kernel function
	complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock>>> (
			fft_sceneTimesBtDev, 
			fft_sceneDev, 
			fft_BtDev,
			(SW/2 + 1), S_SIZE);
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Use plans to transform data
	CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneTimesBtDev, time_sceneTimesBtDev) );

	// Launch kernel function
	complexProduct2d<cufftComplexPrecision> <<<dimGrid_H, dimBlock>>> (
			fft_sceneSqTimesBtDev, 
			fft_sceneSqDev, 
			fft_BtDev,
			(SW/2 + 1), S_SIZE);
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Use plans to transform data
	CUFFT_SAFE_CALL ( cufftExecC2R_t (planC2R, fft_sceneSqTimesBtDev, time_sceneSqTimesBtDev) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute LCC Array
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Launch kernel function
	makeLccArray2d<flccPrecision, cufftRealPrecision> <<<dimGrid_s, dimBlock>>> (
			lccDev, 
			time_sceneTimesBtDev,
			time_sceneSqTimesBtDev,
			time_sceneTimesTDev, 
			sh, sw,
			SW, T_SIZE );
			
	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (time_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneSqDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneSqDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneTimesTDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (fft_sceneSqTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (time_sceneSqTimesBtDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDev) );
	
	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult flcorr2_testStream_dev<float, cufftReal, cufftComplex> (
		float *tempSceneDev, flccSize imSize, 
		cufftComplex *fft_templateDev, cufftComplex *fft_BtDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult flcorr2_testStream_dev<double, cufftDoubleReal, cufftDoubleComplex> (
		double *tempSceneDev, flccSize imSize, 
		cufftDoubleComplex *fft_templateDev, cufftDoubleComplex *fft_BtDev, flccSize temSize,  
		cufftHandle planC2R,
		cufftHandle planR2C );
#endif
