/* 
 * FILE:
 * flcorr3_host.c
 * version 1.1
 *
 * DESCRIPTION:
 * This function computes the 3D LCC distribution given an image and a template on the CPU host.
 * It implements the FLCC algorithm.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * July 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */
 

// System include
#include <FLCC_general_defs.h>
#include <FLCC_fftw_defs.h>


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision, class fftw_plan_t, class fftwComplexPrecision>
flccResult flcorr3_host (
		flccPrecision *image,   flccSize imageDims, 
		flccPrecision *templat, flccSize templateDims, 
		flccPrecision *LCC, 
		int imCount, int n,
		fftw_plan_t c2r_plan_diff,
		fftw_plan_t r2c_plan_diff ) 
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = templateDims.h;
	const int TW = templateDims.w;
	const int TD = templateDims.d;
	const int T_SIZE = TH*TW*TD;
	
	// Image dimensions
	const int IH = imageDims.h;
	const int IW = imageDims.w;
	const int ID = imageDims.d;
	const int I_SIZE = IH*IW*ID;
	
	// Single padded arrays dimensions
	const int sh = IH+TH-1;		
	const int SH = alignSize3d(sh);
	const int sw = IW+TW-1;
	const int SW = alignSize3d(sw);
	const int sd = ID+TD-1;		
	const int SD = alignSize3d(sd);
	const int s_size = sh*sw*sd;
	const int S_SIZE = SH*SW*SD;
	const int H_SIZE = SH*SW*(SD/2 + 1);

	// Auxiliary variables
	int i, k;
	int S_offset, s_offset, H_offset;
	int S_temp, s_temp, H_temp;
	void *status;
	
	// Declare and allocate for Make pemanent arrays (T and Bt) in the time domain
	flccPrecision *time_template;
	MALLOC_SAFE_CALL ( time_template = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	flccPrecision *time_Bt;
	MALLOC_SAFE_CALL ( time_Bt = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate for Make F(T) and F(Bt)
	fftwComplexPrecision *fft_template;
	MALLOC_SAFE_CALL ( fft_template = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	fftwComplexPrecision *fft_Bt;
	MALLOC_SAFE_CALL ( fft_Bt = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	
	// Declare and allocate for Make transient arrays (S and S^2) in the time domain
	flccPrecision *time_scene;
	MALLOC_SAFE_CALL ( time_scene = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	flccPrecision *time_sceneSq;
	MALLOC_SAFE_CALL ( time_sceneSq = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate for Make F(S + i*S^2)
	fftwComplexPrecision *fft_scene;
	MALLOC_SAFE_CALL ( fft_scene = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	fftwComplexPrecision *fft_sceneSq;
	MALLOC_SAFE_CALL ( fft_sceneSq = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	
	// Declare and allocate for Make IF ( F(T) * F(S) )
	fftwComplexPrecision *fft_sceneTimesT;
	MALLOC_SAFE_CALL ( fft_sceneTimesT = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	flccPrecision *time_sceneTimesT;
	MALLOC_SAFE_CALL ( time_sceneTimesT = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	// Declare and allocate for Make IF ( (F(S) * F(Bt)) + i*(F(S^2) * F(Bt)) )
	fftwComplexPrecision *fft_sceneTimesBt;
	MALLOC_SAFE_CALL ( fft_sceneTimesBt = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	flccPrecision *time_sceneTimesBt;
	MALLOC_SAFE_CALL ( time_sceneTimesBt = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	fftwComplexPrecision *fft_sceneSqTimesBt;
	MALLOC_SAFE_CALL ( fft_sceneSqTimesBt = (fftwComplexPrecision *)fftw_malloc_t<flccPrecision> (H_SIZE*sizeof(fftwComplexPrecision)) );
	flccPrecision *time_sceneSqTimesBt;
	MALLOC_SAFE_CALL ( time_sceneSqTimesBt = (flccPrecision *)fftw_malloc_t<flccPrecision> (S_SIZE*sizeof(flccPrecision)) );
	
	
	// Declare and allocate thread variables
	pthread_t *threads;
	pthread_attr_t pthread_custom_attr;
	permParm<flccPrecision>    *perm;
	tranParm<flccPrecision>    *tran;
	prodParm<fftwComplexPrecision>    *prod;
	makeLccParm<flccPrecision> *makeLcc;
	
	MALLOC_SAFE_CALL ( threads = (pthread_t *)malloc (n*sizeof(pthread_t)) );
	PTHREAD_SAFE_CALL ( pthread_attr_init (&pthread_custom_attr) );
	PTHREAD_SAFE_CALL ( pthread_attr_setdetachstate (&pthread_custom_attr, PTHREAD_CREATE_JOINABLE) );
	
	MALLOC_SAFE_CALL ( perm = (permParm<flccPrecision> *)malloc (sizeof(permParm<flccPrecision>)*n) );
	MALLOC_SAFE_CALL ( tran = (tranParm<flccPrecision> *)malloc (sizeof(tranParm<flccPrecision>)*n) );
	MALLOC_SAFE_CALL ( prod = (prodParm<fftwComplexPrecision> *)malloc (sizeof(prodParm<fftwComplexPrecision>)*n) );
	MALLOC_SAFE_CALL ( makeLcc = (makeLccParm<flccPrecision> *)malloc (sizeof(makeLccParm<flccPrecision>)*n) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stage A (computation once for all)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make pemanent arrays (T and Bt) in the time domain
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Template normalization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template too small to parallelize normalization
	flccPrecision mean = findMean<flccPrecision> (templat, T_SIZE);
	flccPrecision std  = findStd<flccPrecision> (mean, templat, T_SIZE);
	if (std == 0)	return FLCC_INVALID_VALUE;
	//normalize<flccPrecision> (mean, std, templat, T_SIZE); // normalize<flccPrecision> in parallel
	
	
	// Make arguments for threads
	S_temp = 0;
	s_temp = 0;
	H_temp = 0;
	for (i=0; i<n; i++)
	{
		S_offset = S_SIZE/n;
		s_offset = s_size/n;
		H_offset = H_SIZE/n;
		if (i < (S_SIZE%n))	 S_offset++;
		if (i < (s_size%n))	 s_offset++;
		if (i < (H_SIZE%n))	 H_offset++;
			
		perm[i].offset  = S_offset;
		prod[i].offset  = H_offset;
		tran[i].offset  = S_offset;
		makeLcc[i].offset = s_offset;
		
	    perm[i].initialPosition  = S_temp;
		prod[i].initialPosition  = H_temp;
		tran[i].initialPosition  = S_temp;
		makeLcc[i].initialPosition = s_temp;
		
	    S_temp += S_offset;
		s_temp += s_offset;
		H_temp += H_offset;
		
		//--------
		
		perm[i].templat      = time_template;
		perm[i].tempTemplate = templat;
		perm[i].Bt           = time_Bt;
		
		perm[i].mean = mean;
		perm[i].std  = std;
		
		perm[i].tw = TW;
		perm[i].th = TH;
		perm[i].td = TD;
		perm[i].sw = SW;
		perm[i].sd = SD;
		
		//--------

		prod[i].s_size = S_SIZE;
		
		//--------
		
		tran[i].scene   = time_scene;
		tran[i].sceneSq = time_sceneSq;

		tran[i].iw = IW;
		tran[i].ih = IH;
		tran[i].id = ID;
		tran[i].sw = SW;
		tran[i].sd = SD;
		
		//--------

		makeLcc[i].sceneTimesBt   = time_sceneTimesBt;
		makeLcc[i].sceneSqTimesBt = time_sceneSqTimesBt;
		makeLcc[i].sceneTimesT    = time_sceneTimesT;
		
		makeLcc[i].sw     = sw;
		makeLcc[i].sd     = sd;
		makeLcc[i].SW     = SW;
		makeLcc[i].SD     = SD;
		makeLcc[i].t_size = T_SIZE;
		
    } 
	
	// Start threads
	for (i=0; i<n; i++)
		PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, makePermanent3d<flccPrecision>, (void *)(perm+i)) );
	
	// Synchronize the completion of each thread
	for (i=0; i<n; i++)
		PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make F(T) and F(Bt)
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
	// Use plan to transform data
	fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_template, fft_template);
	fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_Bt, fft_Bt);
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Stages B-D
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make transient arrays (S and S^2) in the time domain
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	for (k=0; k<imCount; k++)
	{
		// Start threads
		for (i=0; i<n; i++)
		{
			tran[i].image = image + k*I_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, makeTransient3d<flccPrecision>, (void *)(tran+i)) );
		} 
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
			
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make F(S + i*S^2)
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Use plan to transform data in place ( One FFT to compute S and S squared )
		fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_scene, fft_scene);
		fftw_execute_dft_r2c_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (r2c_plan_diff, time_sceneSq, fft_sceneSq);
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( F(T) * F(S) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Start threads
		for (i=0; i<n; i++)
		{
			prod[i].prod_fft = fft_sceneTimesT;
			prod[i].fft1     = fft_scene;
			prod[i].fft2     = fft_template;
		
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, complexProduct3d<fftwComplexPrecision>, (void *)(prod+i)) );
		} 
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
			
			
		// Use plan to transform data
		fftw_execute_dft_c2r_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (c2r_plan_diff, fft_sceneTimesT, time_sceneTimesT);
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Make IF ( (F(S) * F(Bt)) + i*(F(S^2) * F(Bt)) )
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Start threads
		for (i=0; i<n; i++)
		{
			prod[i].prod_fft = fft_sceneTimesBt;
			prod[i].fft1     = fft_scene;
			prod[i].fft2     = fft_Bt;
			
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, complexProduct3d<fftwComplexPrecision>, (void *)(prod+i)) );
		} 
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
		
		// Use plan to transform data
		fftw_execute_dft_c2r_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (c2r_plan_diff, fft_sceneTimesBt, time_sceneTimesBt);
		
		// Start threads
		for (i=0; i<n; i++)
		{
			prod[i].prod_fft = fft_sceneSqTimesBt;
			prod[i].fft1     = fft_sceneSq;
			prod[i].fft2     = fft_Bt;
			
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, complexProduct3d<fftwComplexPrecision>, (void *)(prod+i)) );
		} 
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	
		// Use plan to transform data
		fftw_execute_dft_c2r_t<flccPrecision, fftw_plan_t, fftwComplexPrecision> (c2r_plan_diff, fft_sceneSqTimesBt, time_sceneSqTimesBt);
	
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Compute LCC Array
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Start threads
		for (i=0; i<n; i++)
		{
			makeLcc[i].lcc = LCC + k*s_size;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, makeLccArray3d<flccPrecision>, (void *)(makeLcc+i)) );
		}
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	
	} 
	

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Destroy pthread attribute
	PTHREAD_SAFE_CALL ( pthread_attr_destroy (&pthread_custom_attr) );
	
	
	// Free thread variables
	free (threads);
	free (perm);
	free (tran);
	free (prod);
	free (makeLcc);
	
	// Free for Make pemanent arrays (T and Bt) in the time domain
	fftw_free_t<flccPrecision> (time_template);
	fftw_free_t<flccPrecision> (time_Bt);
	
	// Free for Make F(T) and F(Bt)
	fftw_free_t<flccPrecision> (fft_template);
	fftw_free_t<flccPrecision> (fft_Bt);
	
	// Free for Make transient arrays (S and S^2) in the time domain
	fftw_free_t<flccPrecision> (time_scene);
	fftw_free_t<flccPrecision> (time_sceneSq);
	
	// Free for Make F(S + i*S^2)
	fftw_free_t<flccPrecision> (fft_scene);
	fftw_free_t<flccPrecision> (fft_sceneSq);
	
	// Free for Make IF ( F(T) * F(S) )
	fftw_free_t<flccPrecision> (fft_sceneTimesT);
	fftw_free_t<flccPrecision> (time_sceneTimesT);
	
	// Free for Make IF ( (F(S) * F(Bt)) + i*(F(S^2) * F(Bt)) )
	fftw_free_t<flccPrecision> (fft_sceneTimesBt);
	fftw_free_t<flccPrecision> (time_sceneTimesBt);
	fftw_free_t<flccPrecision> (fft_sceneSqTimesBt);
	fftw_free_t<flccPrecision> (time_sceneSqTimesBt);
	
	
	return FLCC_SUCCESS;
}


// make instances
#ifndef NO_SINGLE_PREC
template flccResult flcorr3_host<float, fftwf_plan, fftwf_complex> (
		float *image,   flccSize imageDims, 
		float *templat, flccSize templateDims, 
		float *LCC, 
		int imCount, int n,
		fftwf_plan c2r_plan_diff,
		fftwf_plan r2c_plan_diff ); 
#endif
#ifndef NO_DOUBLE_PREC
template flccResult flcorr3_host<double, fftw_plan, fftw_complex> (
		double *image,   flccSize imageDims, 
		double *templat, flccSize templateDims, 
		double *LCC, 
		int imCount, int n,
		fftw_plan c2r_plan_diff,
		fftw_plan r2c_plan_diff ); 
#endif
