/* 
 * FILE:
 * lcorr2_dev.cu
 * version 1.2
 *
 * DESCRIPTION:
 * This function computes the 2D LCC distribution given an image and a template on a GPU device.
 * It implements the direct method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult lcorr2_dev (
		flccPrecision *image,   flccSize imSize, 
		flccPrecision *templat, flccSize temSize, 
		flccPrecision *lcc, int imCount )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int T_SIZE = TH*TW;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int I_SIZE = IH*IW;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int rh = (THREADS_2D - sh%THREADS_2D)%THREADS_2D;
	const int SH = sh + rh;
	const int sw = IW+TW-1;
	const int rw = (THREADS_2D - sw%THREADS_2D)%THREADS_2D;
	const int SW = sw + rw;
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	
	// Double padded arrays dimensions
	const int BH = SH+TH-1;
	const int BW = SW+TW-1;
	const int B_SIZE = BH*BW;
	
	
	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_S ( SW/dimBlock.x, SH/dimBlock.y );
	
	// Declare streams
	cudaStream_t host2devMemcpy, dev2hostMemcpy, workExec;
	
	
	// Array with function pointers to kernels
	void (*lcorrelation2d_kernel[])(flccPrecision *, flccPrecision *, flccPrecision *, int, int) = {
		lcorrelation2d_kernel_2x2,     lcorrelation2d_kernel_3x3,     lcorrelation2d_kernel_4x4,     lcorrelation2d_kernel_5x5,
		lcorrelation2d_kernel_6x6,     lcorrelation2d_kernel_7x7,     lcorrelation2d_kernel_8x8,     lcorrelation2d_kernel_9x9,
		lcorrelation2d_kernel_10x10,   lcorrelation2d_kernel_11x11,   lcorrelation2d_kernel_12x12,   lcorrelation2d_kernel_13x13,
		lcorrelation2d_kernel_14x14,   lcorrelation2d_kernel_15x15,   lcorrelation2d_kernel_16x16,   lcorrelation2d_kernel_17x17,
		lcorrelation2d_kernel_18x18,   lcorrelation2d_kernel_19x19,   lcorrelation2d_kernel_20x20,   lcorrelation2d_kernel_21x21,
		lcorrelation2d_kernel_22x22,   lcorrelation2d_kernel_23x23,   lcorrelation2d_kernel_24x24,   lcorrelation2d_kernel_25x25,
		lcorrelation2d_kernel_26x26,   lcorrelation2d_kernel_27x27,   lcorrelation2d_kernel_28x28,   lcorrelation2d_kernel_29x29,
		lcorrelation2d_kernel_30x30,   lcorrelation2d_kernel_31x31,   lcorrelation2d_kernel_32x32
	};

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Template normalization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	flccPrecision mean = findMean<flccPrecision> (templat, T_SIZE);
	flccPrecision std  = findStd<flccPrecision> (mean, templat, T_SIZE);
	if (std == 0)	return FLCC_INVALID_VALUE;

	normalize<flccPrecision> (mean, std, templat, T_SIZE);
	
	// Transfer template to device
	flccPrecision *templateDev;
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&templateDev, T_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMemcpy (templateDev, templat, T_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice) );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make requisite arrays
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Device variables' declaration
	flccPrecision *imageDev, *tempImageDev, *lccDev, *lccDevDepadded;

	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&imageDev,       B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&tempImageDev,   I_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDev,         S_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDevDepadded, s_size*sizeof(flccPrecision)) );
	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (imageDev, 0, B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );

	// Create streams
	CUDA_SAFE_CALL ( cudaStreamCreate(&host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamCreate(&workExec) );
	
	
	// Transfer first image
	CUDA_SAFE_CALL ( cudaMemcpy2D (imageDev+(TW-1)+(TH-1)*BW, BW*sizeof(flccPrecision), image, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyHostToDevice) );
	
	// Begin streaming of images
	for (int k=0; k<imCount; k++)
	{
		// Pad image with zeros (unless it's the first one)
		if ((k-1)>=0)
			CUDA_SAFE_CALL ( cudaMemcpy2DAsync (imageDev+(TW-1)+(TH-1)*BW, BW*sizeof(flccPrecision), tempImageDev, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyDeviceToDevice, host2devMemcpy) );
			
		// Synchronize among threads
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Memory transfer
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		// Transfer back previous image
		if ((k-1)>=0)	
			CUDA_SAFE_CALL ( cudaMemcpyAsync (lcc+(k-1)*s_size, lccDevDepadded, s_size*sizeof(flccPrecision), cudaMemcpyDeviceToHost, dev2hostMemcpy) );
				
		// Transfer next image
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpyAsync (tempImageDev, image+(k+1)*I_SIZE, I_SIZE*sizeof(flccPrecision), cudaMemcpyHostToDevice, host2devMemcpy) );

	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Correlation 
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		// Launch kernel function
		if (TH==TW && TH<=KERNEL_MAX_SIZE_2D)
		{
			lcorrelation2d_kernel[TH-2] <<<dimGrid_S, dimBlock, 0, workExec>>> (
					imageDev, 
					templateDev,
					lccDev,
					SW, BW );
		}	
		else
		{
			lcorrelation2d_kernel_generic<flccPrecision> <<<dimGrid_S, dimBlock, 0, workExec>>> (
					imageDev, 
					templateDev,
					lccDev,
					TH, TW,
					SW, BW );
		}

		CHECK_ERROR;
		CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
		
		// Depad image (unless it's the last one)
		if ((k+1)<imCount)
			CUDA_SAFE_CALL ( cudaMemcpy2DAsync (lccDevDepadded, sw*sizeof(flccPrecision), lccDev, SW*sizeof(flccPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToDevice, dev2hostMemcpy) );
	}

	// Write out last result
	CUDA_SAFE_CALL ( cudaMemcpy2D (lcc+(imCount-1)*s_size, sw*sizeof(flccPrecision), lccDev, SW*sizeof(flccPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToHost) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Destroy streams
	CUDA_SAFE_CALL ( cudaStreamDestroy (host2devMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (dev2hostMemcpy) );
	CUDA_SAFE_CALL ( cudaStreamDestroy (workExec) );
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (tempImageDev) );
	CUDA_SAFE_CALL ( cudaFree (imageDev) );
	CUDA_SAFE_CALL ( cudaFree (templateDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDevDepadded) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult lcorr2_dev<float> (
		float *image,   flccSize imSize, 
		float *templat, flccSize temSize, 
		float *lcc, int imCount );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult lcorr2_dev<double> (
		double *image,   flccSize imSize, 
		double *templat, flccSize temSize, 
		double *lcc, int imCount );
#endif
