/* 
 * FILE:
 * lcorr2_host.c
 * version 1.0
 *
 * DESCRIPTION:
 * This function computes the 2D LCC distribution given an image and a template on the CPU host.
 * It implements the direct method.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2009
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult lcorr2_host (
		flccPrecision *image,   flccSize imageDims, 
		flccPrecision *templat, flccSize templateDims, 
		flccPrecision *LCC, 
		int imCount, int n )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = templateDims.h;
	const int TW = templateDims.w;
	const int T_SIZE = TH*TW;
	
	// Image dimensions
	const int IH = imageDims.h;
	const int IW = imageDims.w;
	const int I_SIZE = IH*IW;
	
	// Double padded arrays dimensions
	const int BH = IH+2*TH-2;
	const int BW = IW+2*TW-2;
	const int B_SIZE = BH*BW;
	
	// Single padded arrays dimensions
	const int SH = IH+TH-1;
	const int SW = IW+TW-1;
	const int S_SIZE = SH*SW;
	
	// Auxiliary variables
	int i, k;
	int p_temp;
	int c_temp;
	
	// Declare and allocate for Array padding
	flccPrecision *paddedImage;
	MALLOC_SAFE_CALL ( paddedImage = (flccPrecision *)malloc (B_SIZE*sizeof(flccPrecision)) );
	
	// Thread variables' declaration and allocation
	pthread_t *threads;
	pthread_attr_t pthread_custom_attr;
	padParm<flccPrecision>  *p;
	corrParm<flccPrecision> *c;
	void *status;
	
	MALLOC_SAFE_CALL ( threads = (pthread_t *)malloc (n*sizeof(pthread_t)) );
	
	PTHREAD_SAFE_CALL ( pthread_attr_init (&pthread_custom_attr) );
	PTHREAD_SAFE_CALL ( pthread_attr_setdetachstate (&pthread_custom_attr, PTHREAD_CREATE_JOINABLE) );
	
	MALLOC_SAFE_CALL ( p =  (padParm<flccPrecision> *)malloc (sizeof(padParm<flccPrecision>)*n)  );
	MALLOC_SAFE_CALL ( c = (corrParm<flccPrecision> *)malloc (sizeof(corrParm<flccPrecision>)*n) );
	
	
	// Array with function pointers to threads
	void *(*lcorrelation2d_thread[])(void *) = {
		lcorrelation2d_thread_2x2<flccPrecision>,     lcorrelation2d_thread_3x3<flccPrecision>,     lcorrelation2d_thread_4x4<flccPrecision>,     lcorrelation2d_thread_5x5<flccPrecision>,
		lcorrelation2d_thread_6x6<flccPrecision>,     lcorrelation2d_thread_7x7<flccPrecision>,     lcorrelation2d_thread_8x8<flccPrecision>,     lcorrelation2d_thread_9x9<flccPrecision>,
		lcorrelation2d_thread_10x10<flccPrecision>,   lcorrelation2d_thread_11x11<flccPrecision>,   lcorrelation2d_thread_12x12<flccPrecision>,   lcorrelation2d_thread_13x13<flccPrecision>,
		lcorrelation2d_thread_14x14<flccPrecision>,   lcorrelation2d_thread_15x15<flccPrecision>,   lcorrelation2d_thread_16x16<flccPrecision>,   lcorrelation2d_thread_17x17<flccPrecision>,
		lcorrelation2d_thread_18x18<flccPrecision>,   lcorrelation2d_thread_19x19<flccPrecision>,   lcorrelation2d_thread_20x20<flccPrecision>,   lcorrelation2d_thread_21x21<flccPrecision>,
		lcorrelation2d_thread_22x22<flccPrecision>,   lcorrelation2d_thread_23x23<flccPrecision>,   lcorrelation2d_thread_24x24<flccPrecision>,   lcorrelation2d_thread_25x25<flccPrecision>,
		lcorrelation2d_thread_26x26<flccPrecision>,   lcorrelation2d_thread_27x27<flccPrecision>,   lcorrelation2d_thread_28x28<flccPrecision>,   lcorrelation2d_thread_29x29<flccPrecision>,
		lcorrelation2d_thread_30x30<flccPrecision>,   lcorrelation2d_thread_31x31<flccPrecision>,   lcorrelation2d_thread_32x32<flccPrecision>
	};
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Template normalization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template too small to parallelize normalization
	flccPrecision mean = findMean<flccPrecision> (templat, T_SIZE);
	flccPrecision std  = findStd<flccPrecision> (mean, templat, T_SIZE);
	if (std == 0)	return FLCC_INVALID_VALUE;
	
	normalize<flccPrecision> (mean, std, templat, T_SIZE);
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make arguments for threads
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	p_temp = 0;
	c_temp = 0;
	
	for (i=0; i<n; i++)
	{
		p[i].offset = B_SIZE/n;
		
		if (i < (B_SIZE%n))
			p[i].offset++;
		
	    p[i].initialPosition = p_temp;
	    p_temp += p[i].offset;
		
		p[i].padded   = paddedImage;
		
		p[i].th = TH;
		p[i].tw = TW;
		p[i].sh = SH;
		p[i].sw = SW;
		p[i].iw = IW;
		p[i].bw = BW;
		
		
		c[i].offset = S_SIZE/n;
		
		if (i < (S_SIZE%n))
			c[i].offset++;
			
	    c[i].initialPosition = c_temp;
	    c_temp += c[i].offset;
		
		c[i].image = paddedImage;
		c[i].templat = templat;
		
		c[i].th = TH;
		c[i].tw = TW;
		c[i].sw = SW;
		c[i].bw = BW;
    } 
	
	
	// Begin streaming of images
	for (k=0; k<imCount; k++)
	{
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Array padding
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
		// Start threads
		for (i=0; i<n; i++)
		{
			p[i].unpadded = image + k*I_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, padSignal2d_both<flccPrecision>, (void *)(p+i)) );
		}
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
		
	
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Correlation 
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
		void *(*lcorrelate2d)(void *);
		if (TH==TW && TH<=THREAD_MAX_SIZE_2D)	lcorrelate2d = lcorrelation2d_thread[TH-2];
		else					            	lcorrelate2d = lcorrelation2d_thread_generic<flccPrecision>;
		
		// Start threads
		for (i=0; i<n; i++)
		{
			c[i].lcc = LCC + k*S_SIZE;
			PTHREAD_SAFE_CALL ( pthread_create (&threads[i], &pthread_custom_attr, lcorrelate2d, (void *)(c+i)) );
		}
	
		// Synchronize the completion of each thread
		for (i=0; i<n; i++)
			PTHREAD_SAFE_CALL ( pthread_join (threads[i], &status) );
	}

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Destroy pthread attribute
	PTHREAD_SAFE_CALL ( pthread_attr_destroy (&pthread_custom_attr) );
	
	// Free threads' variables
	free (threads);
	free (p);
	free (c);
	
	// Free for Array padding
	free (paddedImage);
	
	
	return FLCC_SUCCESS;
}


// make instances
#ifndef NO_SINGLE_PREC
template flccResult lcorr2_host<float> (
		float *image,   flccSize imageDims, 
		float *templat, flccSize templateDims, 
		float *LCC, 
		int imCount, int n );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult lcorr2_host<double> (
		double *image,   flccSize imageDims, 
		double *templat, flccSize templateDims, 
		double *LCC, 
		int imCount, int n );
#endif
