/* 
 * FILE:
 * lcorr2_testStream_dev.cu
 * version 1.0
 *
 * DESCRIPTION:
 * This function executes only the transient stage of the LCC computation using the direct method and it is 
 * used as a benchmark for performance measurements of the 2D image streaming process. It runs on the GPU.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult lcorr2_testStream_dev (
		flccPrecision *tempImageDev, flccSize imSize, 
		flccPrecision *templateDev,  flccSize temSize )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int rh = (THREADS_2D - sh%THREADS_2D)%THREADS_2D;
	const int SH = sh + rh;
	const int sw = IW+TW-1;
	const int rw = (THREADS_2D - sw%THREADS_2D)%THREADS_2D;
	const int SW = sw + rw;
	const int s_size = sh*sw;
	const int S_SIZE = SH*SW;
	
	// Double padded arrays dimensions
	const int BH = SH+TH-1;
	const int BW = SW+TW-1;
	const int B_SIZE = BH*BW;
	
	
	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_2D, THREADS_2D);

	// Find number of blocks in grid
	dim3 dimGrid_S ( SW/dimBlock.x, SH/dimBlock.y );
	
	
	// Array with function pointers to kernels
	void (*lcorrelation2d_kernel[])(flccPrecision *, flccPrecision *, flccPrecision *, int, int) = {
		lcorrelation2d_kernel_2x2,     lcorrelation2d_kernel_3x3,     lcorrelation2d_kernel_4x4,     lcorrelation2d_kernel_5x5,
		lcorrelation2d_kernel_6x6,     lcorrelation2d_kernel_7x7,     lcorrelation2d_kernel_8x8,     lcorrelation2d_kernel_9x9,
		lcorrelation2d_kernel_10x10,   lcorrelation2d_kernel_11x11,   lcorrelation2d_kernel_12x12,   lcorrelation2d_kernel_13x13,
		lcorrelation2d_kernel_14x14,   lcorrelation2d_kernel_15x15,   lcorrelation2d_kernel_16x16,   lcorrelation2d_kernel_17x17,
		lcorrelation2d_kernel_18x18,   lcorrelation2d_kernel_19x19,   lcorrelation2d_kernel_20x20,   lcorrelation2d_kernel_21x21,
		lcorrelation2d_kernel_22x22,   lcorrelation2d_kernel_23x23,   lcorrelation2d_kernel_24x24,   lcorrelation2d_kernel_25x25,
		lcorrelation2d_kernel_26x26,   lcorrelation2d_kernel_27x27,   lcorrelation2d_kernel_28x28,   lcorrelation2d_kernel_29x29,
		lcorrelation2d_kernel_30x30,   lcorrelation2d_kernel_31x31,   lcorrelation2d_kernel_32x32
	};

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make requisite arrays
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Device variables' declaration
	flccPrecision *imageDev, *lccDev, *lccDevDepadded;

	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&imageDev,       B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDev,         S_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDevDepadded, s_size*sizeof(flccPrecision)) );
	
	// Set memory to zeros
	CUDA_SAFE_CALL ( cudaMemset (imageDev, 0, B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	// Pad image with zeros
	CUDA_SAFE_CALL ( cudaMemcpy2D (imageDev+(TW-1)+(TH-1)*BW, BW*sizeof(flccPrecision), tempImageDev, IW*sizeof(flccPrecision), IW*sizeof(flccPrecision), IH, cudaMemcpyDeviceToDevice) );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Correlation 
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Launch kernel function
	if (TH==TW && TH<=KERNEL_MAX_SIZE_2D)
	{
		lcorrelation2d_kernel[TH-2] <<<dimGrid_S, dimBlock>>> (
				imageDev, 
				templateDev,
				lccDev,
				SW, BW );
	}	
	else
	{
		lcorrelation2d_kernel_generic<flccPrecision> <<<dimGrid_S, dimBlock>>> (
				imageDev, 
				templateDev,
				lccDev,
				TH, TW,
				SW, BW );
	}

	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	// Depad 
	CUDA_SAFE_CALL ( cudaMemcpy2D (lccDevDepadded, sw*sizeof(flccPrecision), lccDev, SW*sizeof(flccPrecision), sw*sizeof(flccPrecision), sh, cudaMemcpyDeviceToDevice) );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (imageDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDevDepadded) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult lcorr2_testStream_dev<float> (
		float *tempImageDev, flccSize imSize, 
		float *templateDev,  flccSize temSize );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult lcorr2_testStream_dev<double> (
		double *tempImageDev, flccSize imSize, 
		double *templateDev,  flccSize temSize );
#endif
