/* 
 * FILE:
 * lcorr3_testStream_dev.cu
 * version 1.0
 *
 * DESCRIPTION:
 * This function executes only the transient stage of the LCC computation using the direct method and it is 
 * used as a benchmark for performance measurements of the 3D image streaming process. It runs on the GPU.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
// System include
#include <FLCC_general_defs.h>
#include <FLCC_cuda_defs.h>

	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Main function
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class flccPrecision>
flccResult lcorr3_testStream_dev (
		flccPrecision *tempImageDev, flccSize imSize, 
		flccPrecision *templateDev,  flccSize temSize )
{
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable initialization
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Template dimensions
	const int TH = temSize.h;
	const int TW = temSize.w;
	const int TD = temSize.d;
	
	// Image dimensions
	const int IH = imSize.h;
	const int IW = imSize.w;
	const int ID = imSize.d;

	// Single padded arrays dimensions
	const int sh = IH+TH-1;
	const int rh = (THREADS_3D - sh%THREADS_3D)%THREADS_3D;
	const int SH = sh + rh;
	const int sw = IW+TW-1;
	const int rw = (THREADS_3D - sw%THREADS_3D)%THREADS_3D;
	const int SW = sw + rw;
	const int sd = ID+TD-1;
	const int rd = (THREADS_3D - sd%THREADS_3D)%THREADS_3D;
	const int SD = sd + rd;
	const int s_size = sh*sw*sd;
	const int S_SIZE = SH*SW*SD;
	
	// Double padded arrays dimensions
	const int BH = SH+TH-1;
	const int BW = SW+TW-1;
	const int BD = SD+TD-1;
	const int B_SIZE = BH*BW*BD;
	
	
	// Number of threads in block is fixed
    dim3 dimBlock (THREADS_3D, THREADS_3D, THREADS_3D);

	// Find number of blocks in grid
	dim3 dimGrid_B ( ((BD+dimBlock.x-1)/dimBlock.x) * ((BW+dimBlock.y-1)/dimBlock.y), (BH+dimBlock.z-1)/dimBlock.z );
	dim3 dimGrid_S ( (SD/dimBlock.x) * (SW/dimBlock.y), SH/dimBlock.z );
	
	
	// Array with function pointers to kernels
	void (*lcorrelation3d_kernel[])(flccPrecision *, flccPrecision *, flccPrecision *, int, int, int, int) = {
		lcorrelation3d_kernel_2x2x2,   lcorrelation3d_kernel_3x3x3,   lcorrelation3d_kernel_4x4x4,   lcorrelation3d_kernel_5x5x5,
		lcorrelation3d_kernel_6x6x6,   lcorrelation3d_kernel_7x7x7,   lcorrelation3d_kernel_8x8x8
	};

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make requisite arrays
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Device variables' declaration
	flccPrecision *imageDev, *lccDev, *lccDevDepadded;

	// Allocate memory in device for requisite arrays
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&imageDev,       B_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDev,         S_SIZE*sizeof(flccPrecision)) );
	CUDA_SAFE_CALL ( cudaMalloc ((void **)&lccDevDepadded, s_size*sizeof(flccPrecision)) );

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Array padding 
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Launch kernel function
	padSignal3d_both<flccPrecision> <<<dimGrid_B, dimBlock>>> (
			imageDev, 
			tempImageDev,
			TH, TW, TD,
			IH, IW, ID,
			BH, BW, BD );

	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
		
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Correlation 
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Launch kernel function
	if (TH==TW && TH==TD && TH<=KERNEL_MAX_SIZE_3D)
	{
		lcorrelation3d_kernel[TH-2] <<<dimGrid_S, dimBlock>>> (
				imageDev, 
				templateDev,
				lccDev,
				SW, SD,
				BW, BD );
	}	
	else
	{
		lcorrelation3d_kernel_generic<flccPrecision> <<<dimGrid_S, dimBlock>>> (
				imageDev, 
				templateDev,
				lccDev,
				TH, TW, TD,
				SW, SD,
				BW, BD );
	}

	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	// Depad
	depadSignal3d<flccPrecision> <<<dimGrid_S, dimBlock>>> (
			lccDevDepadded, 
			lccDev,
			sh, sw, sd,
			SW, SD );

	CHECK_ERROR;
	CUDA_SAFE_CALL ( cudaThreadSynchronize() );
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	// Free memory on device
	CUDA_SAFE_CALL ( cudaFree (imageDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDev) );
	CUDA_SAFE_CALL ( cudaFree (lccDevDepadded) );

	
	return FLCC_SUCCESS;
} 


// make instances
#ifndef NO_SINGLE_PREC
template flccResult lcorr3_testStream_dev<float> (
		float *tempImageDev, flccSize imSize, 
		float *templateDev,  flccSize temSize );
#endif
#ifndef NO_DOUBLE_PREC
template flccResult lcorr3_testStream_dev<double> (
		double *tempImageDev, flccSize imSize, 
		double *templateDev,  flccSize temSize );
#endif
