/* 
 * FILE:
 * conv_get_times_2d.c
 * version 1.1
 *
 * DESCRIPTION:
 * This program computes the 2D convolution given an image and a template.
 * The convolution is computed by both the direct method and the FFT algorithm, 
 * on both the host and the device. It collects the timing results for all the above.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * April 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
#include <iostream>
#include <FLCC_general_defs.h>
#ifndef NO_CUDA
#include <FLCC_cuda_defs.h>
#endif
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif
using namespace std;

#define ERROR_LIMIT 1.0e-4


#include "../compare.c"


int main (int argc, char **argv)
{	
	if (argc!=2 && argc!=7)
	{
		cout << "USAGE:\t" << argv[0] << " IH IW TH TW imCount numThreads\n";
		cout << "   OR:\t" << argv[0] << " numThreads\n";
		cout << "\n";
		return 1;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	double *image; // the array containing the pixel values of the scene
	double *templat; // the array containing the pixel values of template that we try to locate within the scene
	
	flccSize imSize, temSize;
	int imCount;
	int numThreads;
	
	switch (argc)
	{
	case 2:
		srand ( time(NULL) );
	
		imSize.h = rand()%4999 + 2;
		imSize.w = rand()%4999 + 2;
	
		temSize.h = rand()%31 + 2;
		temSize.w = rand()%31 + 2;
		
		imCount = rand()%10 + 1;
		
		numThreads = atoi(argv[1]);
		
		break;
		
	case 7:
		imSize.h = atoi(argv[1]);
		imSize.w = atoi(argv[2]);
	
		temSize.h = atoi(argv[3]);
		temSize.w = atoi(argv[4]);
		
		imCount = atoi(argv[5]);
		
		numThreads = atoi(argv[6]);
		
		break;
	}
	
	
	int I_SIZE = imSize.h  * imSize.w;
	int T_SIZE = temSize.h * temSize.w;
	int s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1);
	int SH = alignSize2d (imSize.h+temSize.h-1);
	int SW = alignSize2d (imSize.w+temSize.w-1);
	int S_SIZE = SH*SW;
	int H_SIZE = SH*(SW/2 + 1);
	
	flccResult res;
	double err;
	
	struct timeval start, end;
	struct timezone tzp;
	double lapsed;
	
	#ifndef NO_CUDA
	
	cufftHandle c2r_plan_dev;
	cufftHandle r2c_plan_dev;

	CUFFT_SAFE_CALL ( cufftPlan2d (&c2r_plan_dev, SH, SW, CUFFT_Z2D) );
	CUFFT_SAFE_CALL ( cufftPlan2d (&r2c_plan_dev, SH, SW, CUFFT_D2Z) );
	
	#endif
	
	
	#ifndef NO_FFTW
	
	fftw_plan c2r_plan_host_diff;
	fftw_plan r2c_plan_host_diff;
	fftw_complex *planFFTW_fft;
	double *planFFTW_real;
	MALLOC_SAFE_CALL ( planFFTW_fft = (fftw_complex *)fftw_malloc (sizeof(fftw_complex)*H_SIZE) ); 
	MALLOC_SAFE_CALL ( planFFTW_real = (double *)fftw_malloc (sizeof(double)*S_SIZE) ); 
	
	if (fftw_init_threads() != 0) 
		fftw_plan_with_nthreads (numThreads);
	
	c2r_plan_host_diff = fftw_plan_dft_c2r_2d (SH, SW, planFFTW_fft, planFFTW_real, FFTW_MEASURE);
	if (c2r_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	r2c_plan_host_diff = fftw_plan_dft_r2c_2d (SH, SW, planFFTW_real, planFFTW_fft, FFTW_MEASURE);
	if (r2c_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	
	fftw_free (planFFTW_fft);
	fftw_free (planFFTW_real);
	
	#endif
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make image and template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	MALLOC_SAFE_CALL ( image   = (double *)flcc_malloc (sizeof(double)*imCount*I_SIZE) );
	MALLOC_SAFE_CALL ( templat = (double *)flcc_malloc (sizeof(double)*T_SIZE) );
	
	randMatrix (image, imCount*I_SIZE);
	randMatrix (templat, T_SIZE);

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute the convolution
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	
	// Direct method on host
	double *conv_host;
	MALLOC_SAFE_CALL ( conv_host = (double *)flcc_malloc (sizeof(double)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = conv2_host<double> (
				image, imSize, 
				templat, temSize, 
				conv_host, imCount, 
				numThreads );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << lapsed << "\t";
	}
	else
	{
		cout << "error" << "\t";
	}
	
	
	#ifndef NO_FFTW
	
	// FLCC algorithm on host
	double *fconv_host;
	MALLOC_SAFE_CALL ( fconv_host = (double *)flcc_malloc (sizeof(double)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = fconv2_host<double, fftw_plan, fftw_complex> (
				image, imSize, 
				templat, temSize, 
				fconv_host, imCount, 
				numThreads, 
				c2r_plan_host_diff,
				r2c_plan_host_diff );	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		// Compare between the two methods 
		err = compare2d<double> (conv_host, fconv_host, imSize, temSize, imCount);
		
		if (err > ERROR_LIMIT)
			cout << "wrong result" << "\t";
		else
			cout << lapsed << "\t";
	}
	else
	{
		cout << "error" << "\t";
	}
	
	flcc_free (fconv_host);
	
	#endif // NO_FFTW
	
	
	#ifndef NO_CUDA
	
	// Direct method on device 
	double *conv_dev;
	MALLOC_SAFE_CALL ( conv_dev = (double *)flcc_malloc (sizeof(double)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = conv2_dev<double> (
				image, imSize, 
				templat, temSize, 
				conv_dev, imCount );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		// Compare between the two methods 
		err = compare2d<double> (conv_host, conv_dev, imSize, temSize, imCount);
		
		if (err > ERROR_LIMIT)
			cout << "wrong result" << "\t";
		else
			cout << lapsed << "\t";
	}
	else
	{
		cout << "error" << "\t";
	}
	
	flcc_free (conv_dev);
	
	
	// FLCC algorithm on device
	double *fconv_dev;
	MALLOC_SAFE_CALL ( fconv_dev = (double *)flcc_malloc (sizeof(double)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = fconv2_dev<double, cufftDoubleReal, cufftDoubleComplex> (
				image, imSize, 
				templat, temSize, 
				fconv_dev, imCount,
				c2r_plan_dev,
				r2c_plan_dev);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		// Compare between the two methods 
		err = compare2d<double> (conv_host, fconv_dev, imSize, temSize, imCount);
		
		if (err > ERROR_LIMIT)
			cout << "wrong result" << "\t";
		else
			cout << lapsed << "\t";
	}
	else
	{
		cout << "error" << "\t";
	}
	
	flcc_free (fconv_dev);
	
	#endif // NO_CUDA
	
	
	// Now use FLCC interface
	flccPlan plan;
	flccType type;
	if (imCount == 1)	type = FLCC_SINGLE;
	else				type = FLCC_STREAM;
	
	res = conv_plan_d (&plan, 2, imSize, temSize, type, FLCC_ANY);

	if (res == FLCC_SUCCESS)
	{
		double *conv;
		MALLOC_SAFE_CALL ( conv = (double *)flcc_malloc (sizeof(double)*imCount*s_size) );
		gettimeofday (&start, &tzp);
		res = conv_exec_d (plan, image, templat, conv, imCount);
		gettimeofday (&end, &tzp);
		lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
		
		if (res == FLCC_SUCCESS)
		{
			// Compare between the two methods 
			err = compare2d<double> (conv_host, conv, imSize, temSize, imCount);
			
			if (err > ERROR_LIMIT)
				cout << "wrong result" << "\t";
			else
				cout << lapsed << "\t";
		}
		else
		{
			cout << "error" << "\t";
		}
		
		flcc_free (conv);
		
		res = conv_destroy_d (&plan);
		
		if (res != FLCC_SUCCESS)
		{
			cout << "error" << "\t";
		}
	}
	else
	{
		cout << "error" << "\t";
	}	
	
	
	cout << "\n";
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	#ifndef NO_CUDA
	CUFFT_SAFE_CALL ( cufftDestroy (c2r_plan_dev) );
	CUFFT_SAFE_CALL ( cufftDestroy (r2c_plan_dev) );
	#endif
	
	#ifndef NO_FFTW
	fftw_destroy_plan (c2r_plan_host_diff);
	fftw_destroy_plan (r2c_plan_host_diff);
	#endif
	
	flcc_free (image);
	flcc_free (templat);
	flcc_free (conv_host);
	
	return 0;
}
