/* 
 * FILE:
 * conv_compare_3d.c
 * version 1.1
 *
 * DESCRIPTION:
 * This program computes the 3D convolution given an image and a template.
 * The convolution is computed by both the direct method and the FLCC algorithm, 
 * on both the host and the device. Comparison between the above follows.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * April 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
#include <iostream>
#include <FLCC_general_defs.h>
#ifndef NO_CUDA
#include <FLCC_cuda_defs.h>
#endif
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif
using namespace std;


#include "../compare.c"


int main (int argc, char **argv)
{	
	if (argc!=2 && argc!=9)
	{
		cout << "USAGE:\t" << argv[0] << " IH IW ID TH TW TD imCount numThreads\n";
		cout << "   OR:\t" << argv[0] << " numThreads\n";
		cout << "\n";
		return 1;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	float *image; // the array containing the pixel values of the scene
	float *templat; // the array containing the pixel values of template that we try to locate within the scene
	
	flccSize imSize, temSize;
	int imCount;
	int numThreads;
	
	switch (argc)
	{
	case 2:
		srand ( time(NULL) );
	
		imSize.h = rand()%399 + 2;
		imSize.w = rand()%399 + 2;
		imSize.d = rand()%399 + 2;
	
		temSize.h = rand()%7 + 2;
		temSize.w = rand()%7 + 2;
		temSize.d = rand()%7 + 2;
		
		imCount = rand()%10 + 1;
		
		numThreads = atoi(argv[1]);
		
		break;
		
	case 9:
		imSize.h = atoi(argv[1]);
		imSize.w = atoi(argv[2]);
		imSize.d = atoi(argv[3]);
	
		temSize.h = atoi(argv[4]);
		temSize.w = atoi(argv[5]);
		temSize.d = atoi(argv[6]);
	
		imCount = atoi(argv[7]);
		
		numThreads = atoi(argv[8]);
		
		break;
	}
	
	cout << "   IMAGE: " << imSize.h  << " " << imSize.w  << " " << imSize.d  << "\n";
	cout << "TEMPLATE: " << temSize.h << " " << temSize.w << " " << temSize.d << "\n";
	cout << "NUM OF IMAGES: " << imCount << "\n";
	cout << "\n";
	
	
	int I_SIZE = imSize.h  * imSize.w  * imSize.d;
	int T_SIZE = temSize.h * temSize.w * temSize.d;
	int s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1) * (imSize.d+temSize.d-1);
	int SH = alignSize3d (imSize.h+temSize.h-1);
	int SW = alignSize3d (imSize.w+temSize.w-1);
	int SD = alignSize3d (imSize.d+temSize.d-1);
	int S_SIZE = SH*SW*SD;
	int H_SIZE = SH*SW*(SD/2 + 1);
	
	flccResult res;
	double err;
	
	struct timeval start, end;
	struct timezone tzp;
	double lapsed;
	
	#ifndef NO_CUDA
	
	cufftHandle c2r_plan_dev;
	cufftHandle r2c_plan_dev;
	
	gettimeofday (&start, &tzp);
	
	CUFFT_SAFE_CALL ( cufftPlan3d (&c2r_plan_dev, SH, SW, SD, CUFFT_C2R) );
	CUFFT_SAFE_CALL ( cufftPlan3d (&r2c_plan_dev, SH, SW, SD, CUFFT_R2C) );
	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	cout << "Making gpu plans OK\tTime: " << lapsed << "sec\n";
	
	#endif
	
	
	#ifndef NO_FFTW
	
	fftwf_plan c2r_plan_host_diff;
	fftwf_plan r2c_plan_host_diff;
	fftwf_complex *planFFTW_fft;
	float *planFFTW_real;
	MALLOC_SAFE_CALL ( planFFTW_fft = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex)*H_SIZE) ); 
	MALLOC_SAFE_CALL ( planFFTW_real = (float *)fftwf_malloc (sizeof(float)*S_SIZE) ); 
	
	if (fftwf_init_threads() != 0) 
		fftwf_plan_with_nthreads (numThreads);
	
	gettimeofday (&start, &tzp);
	
	c2r_plan_host_diff = fftwf_plan_dft_c2r_3d (SH, SW, SD, planFFTW_fft, planFFTW_real, FFTW_MEASURE);
	if (c2r_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	r2c_plan_host_diff = fftwf_plan_dft_r2c_3d (SH, SW, SD, planFFTW_real, planFFTW_fft, FFTW_MEASURE);
	if (r2c_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	cout << "Making host plans OK\tTime: " << lapsed << "sec\n";
	cout << "\n";
	
	fftwf_free (planFFTW_fft);
	fftwf_free (planFFTW_real);
	
	#endif
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make image and template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	MALLOC_SAFE_CALL ( image   = (float *)flcc_malloc (sizeof(float)*imCount*I_SIZE) );
	MALLOC_SAFE_CALL ( templat = (float *)flcc_malloc (sizeof(float)*T_SIZE) );

	randMatrix (image, imCount*I_SIZE);
	randMatrix (templat, T_SIZE);

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute the convolution
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	
	// Direct method on host
	float *conv_host;
	MALLOC_SAFE_CALL ( conv_host = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = conv3_host<float> (
				image, imSize, 
				templat, temSize, 
				conv_host, imCount, 
				numThreads );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "conv3_host OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in conv3_host.\n";
	}

	cout << "\n";
	
	
	#ifndef NO_FFTW
	
	// FLCC algorithm on host
	float *fconv_host;
	MALLOC_SAFE_CALL ( fconv_host = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = fconv3_host<float, fftwf_plan, fftwf_complex> (
				image, imSize, 
				templat, temSize, 
				fconv_host, imCount, 
				numThreads, 
				c2r_plan_host_diff,		
				r2c_plan_host_diff );	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "fconv3_host OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare3d<float> (conv_host, fconv_host, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in fconv3_host.\n";
	}

	flcc_free (fconv_host);
	cout << "\n";
	
	#endif // NO_FFTW
	
	
	#ifndef NO_CUDA
	
	// Direct method on device 
	float *conv_dev;
	MALLOC_SAFE_CALL ( conv_dev = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = conv3_dev<float> (
				image, imSize, 
				templat, temSize, 
				conv_dev, imCount );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "conv3_dev OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare3d<float> (conv_host, conv_dev, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in conv3_dev.\n";
	}

	flcc_free (conv_dev);
	cout << "\n";
	
	
	// FLCC algorithm on device
	float *fconv_dev;
	MALLOC_SAFE_CALL ( fconv_dev = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = fconv3_dev<float, cufftReal, cufftComplex> (
				image, imSize, 
				templat, temSize, 
				fconv_dev, imCount,
				c2r_plan_dev,
				r2c_plan_dev);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "fconv3_dev OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare3d<float> (conv_host, fconv_dev, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in fconv3_dev.\n";
	}

	flcc_free (fconv_dev);
	cout << "\n";
	
	#endif // NO_CUDA
	
	
	// Now use FLCC interface
	flccPlan plan;
	flccType type;
	if (imCount == 1)	type = FLCC_SINGLE;
	else				type = FLCC_STREAM;
	
	gettimeofday (&start, &tzp);
	res = conv_plan (&plan, 3, imSize, temSize, type, FLCC_ANY);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );

	if (res == FLCC_SUCCESS)
	{
		cout << "conv_plan OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in conv_plan.\n";
	}	
	
	cout << "\n";
	
	float *conv;
	MALLOC_SAFE_CALL ( conv = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = conv_exec (plan, image, templat, conv, imCount);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "conv_exec OK\tTime: " << lapsed << "sec\n";

		// Compare between the two methods 
		err = compare3d<float> (conv_host, conv, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in conv_exec.\n";
	}
	
	flcc_free (conv);
	cout << "\n";
	
	gettimeofday (&start, &tzp);
	res = conv_destroy (&plan);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "conv_destroy OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in conv_destroy.\n";
	}
	
	cout << "\n";
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	#ifndef NO_CUDA
	CUFFT_SAFE_CALL ( cufftDestroy (c2r_plan_dev) );
	CUFFT_SAFE_CALL ( cufftDestroy (r2c_plan_dev) );
	#endif
	
	#ifndef NO_FFTW
	fftwf_destroy_plan (c2r_plan_host_diff);
	fftwf_destroy_plan (r2c_plan_host_diff);
	#endif
	
	flcc_free (image);
	flcc_free (templat);
	flcc_free (conv_host);
	
	return 0;
}
