/* 
 * FILE:
 * lcorr_compare_2d.c
 * version 1.1
 *
 * DESCRIPTION:
 * This program computes the 2D LCC distribution given an image and a template.
 * The LCC distribution is computed by both the direct method and the FLCC algorithm, 
 * on both the host and the device. Comparison between the above follows.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * April 2011
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
#include <iostream>
#include <FLCC_general_defs.h>
#ifndef NO_CUDA
#include <FLCC_cuda_defs.h>
#endif
#ifndef NO_FFTW
#include <FLCC_fftw_defs.h>
#endif
using namespace std;


#include "../compare.c"


int main (int argc, char **argv)
{	
	if (argc!=2 && argc!=7)
	{
		cout << "USAGE:\t" << argv[0] << " IH IW TH TW imCount numThreads\n";
		cout << "   OR:\t" << argv[0] << " numThreads\n";
		cout << "\n";
		return 1;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	float *image; // the array containing the pixel values of the scene
	float *templat; // the array containing the pixel values of template that we try to locate within the scene
	
	flccSize imSize, temSize;
	int imCount;
	int numThreads;
	
	switch (argc)
	{
	case 2:
		srand ( time(NULL) );
	
		imSize.h = rand()%4999 + 2;
		imSize.w = rand()%4999 + 2;
	
		temSize.h = rand()%31 + 2;
		temSize.w = rand()%31 + 2;
		
		imCount = rand()%10 + 1;
		
		numThreads = atoi(argv[1]);
		
		break;
		
	case 7:
		imSize.h = atoi(argv[1]);
		imSize.w = atoi(argv[2]);
	
		temSize.h = atoi(argv[3]);
		temSize.w = atoi(argv[4]);
		
		imCount = atoi(argv[5]);
		
		numThreads = atoi(argv[6]);
		
		break;
	}
	
	cout << "   IMAGE: " << imSize.h  << " " << imSize.w  << "\n";
	cout << "TEMPLATE: " << temSize.h << " " << temSize.w << "\n";
	cout << "NUM OF IMAGES: " << imCount << "\n";
	cout << "\n";
	
	
	int I_SIZE = imSize.h  * imSize.w;
	int T_SIZE = temSize.h * temSize.w;
	int s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1);
	int SH = alignSize2d (imSize.h+temSize.h-1);
	int SW = alignSize2d (imSize.w+temSize.w-1);
	int S_SIZE = SH*SW;
	int H_SIZE = SH*(SW/2 + 1);
	
	flccResult res;
	double err;
	
	struct timeval start, end;
	struct timezone tzp;
	double lapsed;
	
	#ifndef NO_CUDA
	
	cufftHandle c2r_plan_dev;
	cufftHandle r2c_plan_dev;
	
	gettimeofday (&start, &tzp);

	CUFFT_SAFE_CALL ( cufftPlan2d (&c2r_plan_dev, SH, SW, CUFFT_C2R) );
	CUFFT_SAFE_CALL ( cufftPlan2d (&r2c_plan_dev, SH, SW, CUFFT_R2C) );
	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	cout << "Making gpu plans OK\tTime: " << lapsed << "sec\n";
	
	#endif
	
	
	#ifndef NO_FFTW
	
	fftwf_plan c2r_plan_host_diff;
	fftwf_plan r2c_plan_host_diff;
	fftwf_complex *planFFTW_fft;
	float *planFFTW_real;
	MALLOC_SAFE_CALL ( planFFTW_fft = (fftwf_complex *)fftwf_malloc (sizeof(fftwf_complex)*H_SIZE) ); 
	MALLOC_SAFE_CALL ( planFFTW_real = (float *)fftwf_malloc (sizeof(float)*S_SIZE) ); 
	
	if (fftwf_init_threads() != 0) 
		fftwf_plan_with_nthreads (numThreads);
	
	gettimeofday (&start, &tzp);
	
	c2r_plan_host_diff = fftwf_plan_dft_c2r_2d (SH, SW, planFFTW_fft, planFFTW_real, FFTW_MEASURE);
	if (c2r_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	r2c_plan_host_diff = fftwf_plan_dft_r2c_2d (SH, SW, planFFTW_real, planFFTW_fft, FFTW_MEASURE);
	if (r2c_plan_host_diff == NULL)	return FLCC_EXEC_FAILED;
	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	cout << "Making host plans OK\tTime: " << lapsed << "sec\n";
	cout << "\n";
	
	fftwf_free (planFFTW_fft);
	fftwf_free (planFFTW_real);
	
	#endif
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make image and template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	
	MALLOC_SAFE_CALL ( image   = (float *)flcc_malloc (sizeof(float)*imCount*I_SIZE) );
	MALLOC_SAFE_CALL ( templat = (float *)flcc_malloc (sizeof(float)*T_SIZE) );

	randMatrix (image, imCount*I_SIZE);
	randMatrix (templat, T_SIZE);

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute the LCC distribution
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	
	// Direct method on host
	float *lcc_host;
	MALLOC_SAFE_CALL ( lcc_host = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = lcorr2_host<float> (
				image, imSize, 
				templat, temSize, 
				lcc_host, imCount, 
				numThreads );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr2_host OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in lcorr2_host.\n";
	}

	cout << "\n";
	
	
	#ifndef NO_FFTW
	
	// FLCC algorithm on host
	float *flcc_host;
	MALLOC_SAFE_CALL ( flcc_host = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = flcorr2_host<float, fftwf_plan, fftwf_complex> (
				image, imSize, 
				templat, temSize, 
				flcc_host, imCount, 
				numThreads, 
				c2r_plan_host_diff,
				r2c_plan_host_diff );	
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "flcorr2_host OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare2d<float> (lcc_host, flcc_host, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in flcorr2_host.\n";
	}

	flcc_free (flcc_host);
	cout << "\n";
	
	#endif // NO_FFTW
	
	
	#ifndef NO_CUDA
	
	// Direct method on device 
	float *lcc_dev;
	MALLOC_SAFE_CALL ( lcc_dev = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = lcorr2_dev<float> (
				image, imSize, 
				templat, temSize, 
				lcc_dev, imCount );
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr2_dev OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare2d<float> (lcc_host, lcc_dev, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in lcorr2_dev.\n";
	}
	
	flcc_free (lcc_dev);
	cout << "\n";
	
	
	// FLCC algorithm on device
	float *flcc_dev;
	MALLOC_SAFE_CALL ( flcc_dev = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = flcorr2_dev<float, cufftReal, cufftComplex> (
				image, imSize, 
				templat, temSize, 
				flcc_dev, imCount,
				c2r_plan_dev,
				r2c_plan_dev);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "flcorr2_dev OK\tTime: " << lapsed << "sec\n";
			
		// Compare between the two methods 
		err = compare2d<float> (lcc_host, flcc_dev, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in flcorr2_dev.\n";
	}

	flcc_free (flcc_dev);
	cout << "\n";
	
	#endif // NO_CUDA
	

	// Now use FLCC interface
	flccPlan plan;
	flccType type;
	if (imCount == 1)	type = FLCC_SINGLE;
	else				type = FLCC_STREAM;
	
	gettimeofday (&start, &tzp);
	res = lcorr_plan (&plan, 2, imSize, temSize, type, FLCC_ANY);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );

	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_plan OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in lcorr_plan.\n";
	}	
	
	cout << "\n";
	
	float *flcc;
	MALLOC_SAFE_CALL ( flcc = (float *)flcc_malloc (sizeof(float)*imCount*s_size) );
	gettimeofday (&start, &tzp);
	res = lcorr_exec (plan, image, templat, flcc, imCount);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_exec OK\tTime: " << lapsed << "sec\n";

		// Compare between the two methods 
		err = compare2d<float> (lcc_host, flcc, imSize, temSize, imCount);
		cout << "Error: " << err << "\n";
	}
	else
	{
		cout << "Error in lcorr_exec.\n";
	}
	
	flcc_free (flcc);
	cout << "\n";
	
	gettimeofday (&start, &tzp);
	res = lcorr_destroy (&plan);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_destroy OK\tTime: " << lapsed << "sec\n";
	}
	else
	{
		cout << "Error in lcorr_destroy.\n";
	}
	
	cout << "\n";
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	#ifndef NO_CUDA
	CUFFT_SAFE_CALL ( cufftDestroy (c2r_plan_dev) );
	CUFFT_SAFE_CALL ( cufftDestroy (r2c_plan_dev) );
	#endif
	
	#ifndef NO_FFTW
	fftwf_destroy_plan (c2r_plan_host_diff);
	fftwf_destroy_plan (r2c_plan_host_diff);
	#endif
	
	flcc_free (image);
	flcc_free (templat);
	flcc_free (lcc_host);
	
	return 0;
}
