/* 
 * FILE:
 * lcorr_test.c
 * version 1.0
 *
 * DESCRIPTION:
 * This program tests the FLCC library user interface routines that deal with LCC.
 *
 * AUTHORS:
 * Georgios Papamakarios 
 * Georgios Rizos
 *
 * Aristotle University of Thessaloniki
 * Facutly of Engineering
 * Department of Electrical and Computer Engineering
 *
 * DATE:
 * November 2009
 *
 * CONTACT INFO:
 * e-mail: gpapamak@auth.gr
 *         georgios.rizos12@imperial.ac.uk
 */

 
#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <flcc.h>
using namespace std;


// Fill a float matrix with random values
void fillMatrix (float *matrix, int size)
{
	srand ( time(NULL) );
	for (int i=0; i<size; i++)
		matrix[i] = (float)rand()/RAND_MAX;
} 


int main (int argc, char **argv)
{	
	if (argc!=8 && argc!=10)
	{
		cout << "USAGE:\t" << argv[0] << " IH IW TH TW imCount flccType flccPlatform\n";
		cout << "   OR:\t" << argv[0] << " IH IW ID TH TW TD imCount flccType flccPlatform\n";
		cout << "flccType:\t0 for FLCC_SINGLE, 1 for FLCC_STREAM\n";
		cout << "flccPlatform:\t0 for FLCC_DEVICE, 1 for FLCC_HOST, 2 for FLCC_ANY\n";
		cout << "\n";
		return 1;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variable declarations
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	float *image; // the array containing the pixel values of the scene
	float *templat; // the array containing the pixel values of template that we try to locate within the scene
	float *lcc;
	
	int dim;
	flccSize imSize, temSize;
	int imCount;
	int I_SIZE, T_SIZE, s_size;
	flccType type;
	flccPlatform platform;
	int t;

	
	switch (argc)
	{
	case 8:
		dim = 2;
	
		imSize.h = atoi(argv[1]);
		imSize.w = atoi(argv[2]);
	
		temSize.h = atoi(argv[3]);
		temSize.w = atoi(argv[4]);
		
		imCount = atoi(argv[5]);
		
		t = atoi(argv[6]);
		if      (t==0)  type = FLCC_SINGLE;
		else if (t==1)	type = FLCC_STREAM;
		
		t = atoi(argv[7]);
		if 	    (t==0)	platform = FLCC_DEVICE;
		else if (t==1)	platform = FLCC_HOST;
		else if (t==2)	platform = FLCC_ANY;
		
		T_SIZE = temSize.h * temSize.w;
		I_SIZE = imSize.h  * imSize.w;
		s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1);
		
		cout << "   IMAGE: " << imSize.h  << " " << imSize.w  << "\n";
		cout << "TEMPLATE: " << temSize.h << " " << temSize.w << "\n";
		cout << "NUM OF IMAGES: " << imCount << "\n";
		if      (type==FLCC_SINGLE)			cout << "TYPE: FLCC_SINGLE\n";
		else if (type==FLCC_STREAM)			cout << "TYPE: FLCC_STREAM\n";
		if      (platform==FLCC_HOST)		cout << "PLATFORM: FLCC_HOST\n";
		else if (platform==FLCC_DEVICE)		cout << "PLATFORM: FLCC_DEVICE\n";
		else if (platform==FLCC_ANY)		cout << "PLATFORM: FLCC_ANY\n";
		cout << "\n";
		
		break;
		
	case 10:
		dim = 3;
	
		imSize.h = atoi(argv[1]);
		imSize.w = atoi(argv[2]);
		imSize.d = atoi(argv[3]);
	
		temSize.h = atoi(argv[4]);
		temSize.w = atoi(argv[5]);
		temSize.d = atoi(argv[6]);
		
		imCount = atoi(argv[7]);
		
		t = atoi(argv[8]);
		if      (t==0)	type = FLCC_SINGLE;
		else if (t==1)	type = FLCC_STREAM;
		
		t = atoi(argv[9]);
		if 	    (t==0)	platform = FLCC_DEVICE;
		else if (t==1)	platform = FLCC_HOST;
		else if (t==2)	platform = FLCC_ANY;
		
		T_SIZE = temSize.h * temSize.w * temSize.d;
		I_SIZE = imSize.h  * imSize.w  * imSize.d;
		s_size = (imSize.h+temSize.h-1) * (imSize.w+temSize.w-1) * (imSize.d+temSize.d-1);
		
		cout << "   IMAGE: " << imSize.h  << " " << imSize.w  << " " << imSize.d  << "\n";
		cout << "TEMPLATE: " << temSize.h << " " << temSize.w << " " << temSize.d << "\n";
		cout << "NUM OF IMAGES: " << imCount << "\n";
		if      (type==FLCC_SINGLE)			cout << "TYPE: FLCC_SINGLE\n";
		else if (type==FLCC_STREAM)			cout << "TYPE: FLCC_STREAM\n";
		if      (platform==FLCC_HOST)		cout << "PLATFORM: FLCC_HOST\n";
		else if (platform==FLCC_DEVICE)		cout << "PLATFORM: FLCC_DEVICE\n";
		else if (platform==FLCC_ANY)		cout << "PLATFORM: FLCC_ANY\n";
		cout << "\n";
		
		break;
	}
	
	
	flccResult res;
	
	struct timeval start, end;
	struct timezone tzp;
	double lapsed;
	
	flccPlan plan;
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Make image and template
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	image   = (float *)flcc_malloc (sizeof(float)*imCount*I_SIZE);
	templat = (float *)flcc_malloc (sizeof(float)*T_SIZE);
	lcc     = (float *)flcc_malloc (sizeof(float)*imCount*s_size);
	
	if (image==NULL || templat==NULL || lcc==NULL)	
	{
		cout << "Memory allocation failure.\n";
		exit(1);
	}

	fillMatrix (image,   imCount*I_SIZE);
	fillMatrix (templat, T_SIZE);

	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Compute the LCC distribution
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	gettimeofday (&start, &tzp);
	res = lcorr_plan (&plan, dim, imSize, temSize, type, platform);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_plan OK\tTime: " << lapsed << "sec\n";
		cout << "\n";
	}
	else
	{
		cout << "Error in lcorr_plan.\n";
		exit(1);
	}
	
	switch (plan.method)
	{
	case FLCC_DIRECT2_DEV:	
		cout << "METHOD SELECTED: FLCC_DIRECT2_DEV\n";
		cout << "\n";
					
		break;
		
	case FLCC_DIRECT3_DEV:
		cout << "METHOD SELECTED: FLCC_DIRECT3_DEV\n";
		cout << "\n";
					
		break;

	case FLCC_FAST2_DEV:
		cout << "METHOD SELECTED: FLCC_FAST2_DEV\n";
		cout << "\n";
					
		break;
		
	case FLCC_FAST3_DEV:
		cout << "METHOD SELECTED: FLCC_FAST3_DEV\n";
		cout << "\n";
					
		break;
		
	case FLCC_DIRECT2_HOST:
		cout << "METHOD SELECTED: FLCC_DIRECT2_HOST\n";
		cout << "\n";
					
		break;
		
	case FLCC_DIRECT3_HOST:
		cout << "METHOD SELECTED: FLCC_DIRECT3_HOST\n";
		cout << "\n";
					
		break;
		
	case FLCC_FAST2_HOST:
		cout << "METHOD SELECTED: FLCC_FAST2_HOST\n";
		cout << "\n";
					
		break;
		
	case FLCC_FAST3_HOST:
		cout << "METHOD SELECTED: FLCC_FAST3_HOST\n";
		cout << "\n";
					
		break;
		
	default:
		cout << "METHOD SELECTED: ERROR\n";
		cout << "\n";
	}

	
	gettimeofday (&start, &tzp);
	res = lcorr_exec (plan, image, templat, lcc, imCount);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_exec OK\tTime: " << lapsed << "sec\n";
		cout << "\n";
	}
	else
	{
		cout << "Error in lcorr_exec.\n";
		exit(1);
	}
	
	
	gettimeofday (&start, &tzp);
	res = lcorr_destroy (&plan);
	gettimeofday (&end, &tzp);
	lapsed = (double)( (end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec)/1.0e6 );
	
	if (res == FLCC_SUCCESS)
	{
		cout << "lcorr_destroy OK\tTime: " << lapsed << "sec\n";
		cout << "\n";
	}
	else
	{
		cout << "Error in lcorr_destroy.\n";
		exit(1);
	}
	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Free memory
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	flcc_free (image);
	flcc_free (templat);
	flcc_free (lcc);
	
	return 0;
}
